/*
<:copyright-broadcom 
 
 Copyright (c) 2002 Broadcom Corporation 
 All Rights Reserved 
 No portions of this material may be reproduced in any form without the 
 written permission of: 
          Broadcom Corporation 
          16215 Alton Parkway 
          Irvine, California 92619 
 All information contained in this document is Broadcom Corporation 
 company private, proprietary, and trade secret. 
 
:>
*/
/****************************************************************************
 *
 * AdslFile.c -- Adsl driver file I/O functions
 *
 * Authors: Ilya Stomakhin
 *
 * $Revision: 1.1 $
 *
 * $Id: AdslFile.c,v 1.1 2004/04/14 21:11:59 ilyas Exp $
 *
 * $Log: AdslFile.c,v $
 * Revision 1.1  2004/04/14 21:11:59  ilyas
 * Inial CVS checkin
 *
 ****************************************************************************/

#include <linux/version.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/kernel.h>
#include <linux/fcntl.h>
#include <linux/syscalls.h>
#else
#define __KERNEL_SYSCALLS__
#endif

#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/unistd.h>
#include <asm/uaccess.h>

#include "AdslFile.h"

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#define	adslf_open		sys_open
#define	adslf_read		sys_read
#define	adslf_close		sys_close
#define	adslf_lseek		sys_lseek
#else
#define	adslf_open		open
#define	adslf_read		read
#define	adslf_close		close
#define	adslf_lseek		lseek

static int errno;
#endif

void * AdslCoreSetSdramImageAddr(ulong lmem2, ulong sdramSize);

int AdslFileLoadImage(char * fname, void *pAdslLMem, void *pAdslSDRAM)
{
	adslPhyImageHdr		phyHdr;
	int			fd;
	mm_segment_t		fs = get_fs();
	set_fs(get_ds());

	fd = adslf_open(fname, 0, 0);
	if (fd == -1) {
		printk("Unable to load '%s'.\n", fname);
		return 0;
	}

	if (adslf_read(fd, (void *)&phyHdr, sizeof(phyHdr)) != sizeof(phyHdr)) {
		printk("Failed to read image header from '%s'.\n", fname);
		adslf_close(fd);
		return 0;
	}

	adslf_lseek(fd, phyHdr.lmemOffset, 0);
	if (adslf_read(fd, pAdslLMem, phyHdr.lmemSize) != phyHdr.lmemSize) {
		printk("Failed to read LMEM image from '%s'.\n", fname);
		adslf_close(fd);
		return 0;
	}

	pAdslSDRAM = AdslCoreSetSdramImageAddr(((ulong*)pAdslLMem)[2], phyHdr.sdramSize);
	adslf_lseek(fd, phyHdr.sdramOffset, 0);
	if (adslf_read(fd, pAdslSDRAM, phyHdr.sdramSize) != phyHdr.sdramSize) {
		printk("Failed to read SDRAM image from '%s'.\n", fname);
		adslf_close(fd);
		return 0;
	}

#if 0
	printk ("Image Load: LMEM=(0x%lX, %ld,%ld) SDRAM=(0x%lX, %ld,%ld)\n", 
		pAdslLMem, phyHdr.lmemOffset, phyHdr.lmemSize, 
		pAdslSDRAM, phyHdr.sdramOffset, phyHdr.sdramSize);
#endif
	adslf_close(fd);
	set_fs(fs);
	return phyHdr.lmemSize + phyHdr.sdramSize;
}

