/*
<:copyright-broadcom 
 
 Copyright (c) 2002 Broadcom Corporation 
 All Rights Reserved 
 No portions of this material may be reproduced in any form without the 
 written permission of: 
          Broadcom Corporation 
          16215 Alton Parkway 
          Irvine, California 92619 
 All information contained in this document is Broadcom Corporation 
 company private, proprietary, and trade secret. 
 
:>
*/
/***************************************************************************
 * File Name  : AdslDrv.c
 *
 * Description: This file contains Linux character device driver entry points
 *              for the ATM API driver.
 *
 * Updates    : 08/24/2001  lat.  Created.
 ***************************************************************************/


/* Includes. */
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/capability.h>
#include <linux/slab.h>
#include <linux/if_arp.h>
#include <asm/uaccess.h>

#include <bcmtypes.h>
#include <adsldrv.h>
#include <bcmadsl.h>
#include <bcmatmapi.h>
#include <board.h>

#include <DiagDef.h>
#include "BcmOs.h"

#ifndef DSL_IFNAME
#define	DSL_IFNAME "dsl0"
#endif

/* Ron */
#include "led/led.h"

#if defined(CONFIG_BCM_ATM_BONDING_ETH) || defined(CONFIG_BCM_ATM_BONDING_ETH_MODULE)
extern int atmbonding_enabled;
extern BCMADSL_STATUS BcmAdsl_SetGfc2VcMapping(int bOn) ;
#endif

/* typedefs. */
typedef void (*ADSL_FN_IOCTL) (unsigned long arg);

/* Externs. */
BCMADSL_STATUS BcmAdsl_SetVcEntry (int gfc, int port, int vpi, int vci);
BCMADSL_STATUS BcmAdsl_SetVcEntryEx (int gfc, int port, int vpi, int vci, int pti);
void BcmAdsl_AtmSetPortId(int path, int portId);
void BcmAdsl_AtmClearVcTable(void);
void BcmAdsl_AtmAddVc(int vpi, int vci);
void BcmAdsl_AtmDeleteVc(int vpi, int vci);
void BcmAdsl_AtmSetMaxSdu(unsigned short maxsdu);

extern void *g_pfnAdslSetVcEntry;
extern void *g_pfnAdslSetAtmLoopbackMode;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
extern void *g_pfnAdslSetVcEntryEx;
#ifdef CONFIG_ATM_EOP_MONITORING
extern void *g_pfnAdslAtmClearVcTable;
extern void *g_pfnAdslAtmAddVc;
extern void *g_pfnAdslAtmDeleteVc;
extern void *g_pfnAdslAtmSetMaxSdu;
#endif
#ifdef BCM_ATM_BONDING
extern void *g_pfnAdslAtmbSetPortId;
#endif
#if defined(BCM_ATM_BONDING) || defined(BCM_ATM_BONDING_ETH)
extern void *g_pfnAdslAtmbGetConnInfo;
#endif
#endif

/* Prototypes. */
static int __init adsl_init( void );
static void __exit adsl_cleanup( void );
static int adsl_open( struct inode *inode, struct file *filp );
static int adsl_ioctl( struct inode *inode, struct file *flip,
    unsigned int command, unsigned long arg );
static void DoCheck( unsigned long arg );
static void DoInitialize( unsigned long arg );
static void DoUninitialize( unsigned long arg );
static void DoConnectionStart( unsigned long arg );
static void DoConnectionStop( unsigned long arg );
static void DoGetPhyAddresses( unsigned long arg );
static void DoSetPhyAddresses( unsigned long arg );
static void DoMapATMPortIDs( unsigned long arg );
static void DoGetConnectionInfo( unsigned long arg );
static void DoDiagCommand( unsigned long arg );
static void DoGetObjValue( unsigned long arg );
static void DoSetObjValue( unsigned long arg );
static void DoStartBERT( unsigned long arg );
static void DoStopBERT( unsigned long arg );
static void DoConfigure( unsigned long arg );
static void DoTest( unsigned long arg );
static void DoGetConstelPoints( unsigned long arg );
static void DoGetVersion( unsigned long arg );
static void DoSetSdramBase( unsigned long arg );
static void DoResetStatCounters( unsigned long arg );
static void DoSetOemParameter( unsigned long arg );
static void DoBertStartEx( unsigned long arg );
static void DoBertStopEx( unsigned long arg );
static void AdslConnectCb( ADSL_LINK_STATE AdslLinkState, UINT32 ulParm );

#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
static ssize_t adsl_read(struct file *file, char __user *buf,
                         size_t count, loff_t *ppos);
static ssize_t adsl_write(struct file *file, const char __user *buf,
                          size_t count, loff_t *ppos);
static unsigned int adsl_poll(struct file *file, poll_table *wait);
extern void dumpaddr( unsigned char *pAddr, int nLen );
#endif /* defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO) */

typedef struct
{
  int readEventMask;    
  int writeEventMask;    
  char *buffer;
} ADSL_IO, *pADSL_IO;

/* Globals. */
static struct file_operations adsl_fops =
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
    .ioctl      = adsl_ioctl,
    .open       = adsl_open,
#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
    .read       = adsl_read,
    .write      = adsl_write,
    .poll       = adsl_poll,
#endif 
#else
    ioctl:      adsl_ioctl,
    open:       adsl_open,
#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
    read:       adsl_read,
    write:      adsl_write,
    poll:       adsl_poll,
#endif 
#endif
};

static UINT16 g_usAtmFastPortId;
static UINT16 g_usAtmInterleavedPortId;
static struct net_device_stats g_dsl_if_stats;
static struct net_device *g_dslNetDev;
static int g_eoc_hdr_offset = 0;
static int g_eoc_hdr_len = 0;
static char *outBuffer=NULL;

#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
#if 0
static int g_eocEvent = (ADSL_LINK_DROPPED | ADSL_EOC_FRAME_SENT);
#else
/* no link up message is received from DSLAM, we assume it's up */
static int g_eocEvent = (ADSL_EOC_FRAME_SENT);
#endif
#endif 

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
#if defined(MODULE)
/***************************************************************************
 * Function Name: init_module
 * Description  : Initial function that is called if this driver is compiled
 *                as a module.  If it is not, adsl_init is called in
 *                chr_dev_init() in drivers/char/mem.c.
 * Returns      : None.
 ***************************************************************************/
int init_module(void)
{
    return( adsl_init() );              
}


/***************************************************************************
 * Function Name: cleanup_module
 * Description  : Final function that is called if this driver is compiled
 *                as a module.
 * Returns      : None.
 ***************************************************************************/
void cleanup_module(void)
{
    if (MOD_IN_USE)
        printk("adsl: cleanup_module failed because module is in use\n");
    else
        adsl_cleanup();
}
#endif //MODULE 
#endif //LINUX_VERSION_CODE

/***************************************************************************
 * Function Name: dsl_config_netdev
 * Description  : Configure a network device for DSL interface
 * Returns      : None.
 ***************************************************************************/
struct net_device_stats *dsl_get_stats(struct net_device *dev)
{
/*
        UINT32 interfaceId;
        ATM_INTERFACE_CFG Cfg;
        ATM_INTERFACE_STATS KStats;
        PATM_INTF_ATM_STATS pAtmStats = &KStats.AtmIntfStats;

        BcmAtm_GetInterfaceId(0, &interfaceId);
        Cfg.ulStructureId = ID_ATM_INTERFACE_CFG;
        KStats.ulStructureId = ID_ATM_INTERFACE_STATS;
        if ( BcmAtm_GetInterfaceCfg(interfaceId, &Cfg)==STS_SUCCESS &&
            BcmAtm_GetInterfaceStatistics(interfaceId, &KStats, 0)==STS_SUCCESS)
        {
            atm_if_stats.rx_bytes = (unsigned long) pAtmStats->ulIfInOctets;
            atm_if_stats.tx_bytes = (unsigned long) pAtmStats->ulIfOutOctets;
            atm_if_stats.rx_errors = (unsigned long) pAtmStats->ulIfInErrors;
            atm_if_stats.tx_errors = (unsigned long) pAtmStats->ulIfOutErrors;  
        }
*/
        return &g_dsl_if_stats;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static void dsl_setup(struct net_device *dev) {
        
    dev->type = ARPHRD_DSL;
    dev->mtu = 0;
    dev->get_stats = dsl_get_stats;        
}

static int dsl_config_netdev(void) {

    g_dslNetDev = alloc_netdev(0, DSL_IFNAME, dsl_setup);
    if ( g_dslNetDev )
         return register_netdev(g_dslNetDev);
    else    
         return -ENOMEM;
}
#else
static int dsl_config_netdev(void) {
    int status = 0;
    char name[16]="";

    sprintf(name, "dsl0");
    g_dslNetDev = dev_alloc(name,  &status);
    g_dslNetDev->type = ARPHRD_DSL;
    g_dslNetDev->mtu = 0;
    g_dslNetDev->get_stats = dsl_get_stats;
    register_netdev(g_dslNetDev);    
    return status;
}
#endif

/***************************************************************************
 * Function Name: adsl_init
 * Description  : Initial function that is called at system startup that
 *                registers this device.
 * Returns      : None.
 ***************************************************************************/
static int __init adsl_init( void )
{
    printk( "adsl: adsl_init entry\n" );
    if ( register_chrdev( ADSLDRV_MAJOR, "adsl", &adsl_fops ) )
    {
         printk("adsldd: failed to create adsl\n");
         return( -1 );
    }            

    /* add by Ron */
    my_led_init();

    if ( dsl_config_netdev() )
    {
         printk("adsldd: failed to create %s\n", DSL_IFNAME);
         return( -1 );
    }
    else
         return( 0 );
} /* adsl_init */


/***************************************************************************
 * Function Name: adsl_cleanup
 * Description  : Final function that is called when the module is unloaded.
 * Returns      : None.
 ***************************************************************************/
static void __exit adsl_cleanup( void )
{
    printk( "adsl: adsl_cleanup entry\n" );

    /* Add by Ron */
    my_led_cleanup();

    unregister_chrdev( ADSLDRV_MAJOR, "adsl" );
} /* adsl_cleanup */


/***************************************************************************
 * Function Name: adsl_open
 * Description  : Called when an application opens this device.
 * Returns      : 0 - success
 ***************************************************************************/
static int adsl_open( struct inode *inode, struct file *filp )
{
    return( 0 );
} /* adsl_open */


/***************************************************************************
 * Function Name: adsl_ioctl
 * Description  : Main entry point for an application send issue ATM API
 *                requests.
 * Returns      : 0 - success or error
 ***************************************************************************/
static int adsl_ioctl( struct inode *inode, struct file *flip,
    unsigned int command, unsigned long arg )
{
    int ret = 0;
    unsigned int cmdnr = _IOC_NR(command);

    ADSL_FN_IOCTL IoctlFuncs[] = {DoCheck, DoInitialize, DoUninitialize,
        DoConnectionStart, DoConnectionStop, DoGetPhyAddresses, DoSetPhyAddresses,
                DoMapATMPortIDs,DoGetConnectionInfo, DoDiagCommand, DoGetObjValue, 
                DoStartBERT, DoStopBERT, DoConfigure, DoTest, DoGetConstelPoints,
                DoGetVersion, DoSetSdramBase, DoResetStatCounters, DoSetOemParameter, 
                DoBertStartEx, DoBertStopEx , DoSetObjValue };

    if( cmdnr >= 0 && cmdnr < MAX_ADSLDRV_IOCTL_COMMANDS )
        (*IoctlFuncs[cmdnr]) (arg);
    else
        ret = -EINVAL;

    return( ret );
} /* adsl_ioctl */


/***************************************************************************
 * Function Name: DoCheck
 * Description  : Calls BcmAdsl_Check on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoCheck( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_Check();
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
} /* DoCheck */


/***************************************************************************
 * Function Name: DoInitialize
 * Description  : Calls BcmAdsl_Initialize on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoInitialize( unsigned long arg )
{
        ADSLDRV_INITIALIZE      KArg;
        adslCfgProfile          adslCfg, *pAdslCfg;

    /* For now, the user callback function is not used.  Rather,
     * this module handles the connection up/down callback.  The
     * user application will need to call BcmAdsl_GetConnectionStatus
     * in order to determine the state of the connection.
     */

        pAdslCfg = NULL;
        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;

                if (NULL != KArg.pAdslCfg) {
                        if( copy_from_user( &adslCfg, KArg.pAdslCfg, sizeof(adslCfg)) != 0 )
                                break;
                        pAdslCfg = &adslCfg;
                }
        } while (0);

    KArg.bvStatus = BcmAdsl_Initialize( AdslConnectCb, 0, pAdslCfg);
#ifdef DYING_GASP_API
    kerSysRegisterDyingGaspHandler(DSL_IFNAME, &BcmAdsl_DyingGaspHandler, 0);
#endif
    put_user( KArg.bvStatus, &((PADSLDRV_INITIALIZE) arg)->bvStatus );

    /* This ADSL module is loaded after the ATM API module so the ATM API
     * cannot call the functions below directly.  This modules sets the
     * function address to a global variable which is used by the ATM API
     * module.
     */
    g_pfnAdslSetVcEntry = (void *) BcmAdsl_SetVcEntry;
    g_pfnAdslSetAtmLoopbackMode = (void *) BcmAdsl_SetAtmLoopbackMode;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
    g_pfnAdslSetVcEntryEx = (void *) BcmAdsl_SetVcEntryEx;
#ifdef CONFIG_ATM_EOP_MONITORING
	
    g_pfnAdslAtmClearVcTable = BcmAdsl_AtmClearVcTable;
    g_pfnAdslAtmAddVc = BcmAdsl_AtmAddVc;
    g_pfnAdslAtmDeleteVc = BcmAdsl_AtmDeleteVc;
    g_pfnAdslAtmSetMaxSdu = BcmAdsl_AtmSetMaxSdu;
#endif
#ifdef BCM_ATM_BONDING
	 g_pfnAdslAtmbSetPortId = BcmAdsl_AtmSetPortId;
#endif
#if defined(BCM_ATM_BONDING) || defined(BCM_ATM_BONDING_ETH)
    g_pfnAdslAtmbGetConnInfo = BcmAdsl_GetConnectionInfo;
#endif

#if defined(CONFIG_BCM_ATM_BONDING_ETH) || defined(CONFIG_BCM_ATM_BONDING_ETH_MODULE)
    if (atmbonding_enabled == 1)
    {
      printk("Disabling Gfc2VcMapping\n");
      BcmAdsl_SetGfc2VcMapping(0);
    }
#endif
#endif

    /* Set LED into mode that is looking for carrier. */
    kerSysLedCtrl(kLedAdsl, kLedStateFail);
} /* DoInitialize */


/***************************************************************************
 * Function Name: DoUninitialize
 * Description  : Calls BcmAdsl_Uninitialize on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoUninitialize( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_Uninitialize();
#ifdef DYING_GASP_API
    kerSysDeregisterDyingGaspHandler(DSL_IFNAME);
#endif
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
} /* DoUninitialize */


/***************************************************************************
 * Function Name: DoConnectionStart
 * Description  : Calls BcmAdsl_ConnectionStart on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoConnectionStart( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_ConnectionStart();
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
} /* DoConnectionStart */


/***************************************************************************
 * Function Name: DoConnectionStop
 * Description  : Calls BcmAdsl_ConnectionStop on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoConnectionStop( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_ConnectionStop();
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
} /* DoConnectionStop */


/***************************************************************************
 * Function Name: DoGetPhyAddresses
 * Description  : Calls BcmAdsl_GetPhyAddresses on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoGetPhyAddresses( unsigned long arg )
{
    ADSLDRV_PHY_ADDR KArg;

    KArg.bvStatus = BcmAdsl_GetPhyAddresses( &KArg.ChannelAddr );
    copy_to_user( (void *) arg, &KArg, sizeof(KArg) );
} /* DoGetPhyAddresses */


/***************************************************************************
 * Function Name: DoSetPhyAddresses
 * Description  : Calls BcmAdsl_SetPhyAddresses on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoSetPhyAddresses( unsigned long arg )
{
    ADSLDRV_PHY_ADDR KArg;

    if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) == 0 )
    {
        KArg.bvStatus = BcmAdsl_SetPhyAddresses( &KArg.ChannelAddr );
        put_user( KArg.bvStatus, &((PADSLDRV_PHY_ADDR) arg)->bvStatus );
    }
} /* DoSetPhyAddresses */


/***************************************************************************
 * Function Name: DoMapATMPortIDs
 * Description  : Maps ATM Port IDs to DSL PHY Utopia Addresses.
 * Returns      : None.
 ***************************************************************************/
static void DoMapATMPortIDs( unsigned long arg )
{
    ADSLDRV_MAP_ATM_PORT KArg;

    if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) == 0 ) {
                g_usAtmFastPortId = KArg.usAtmFastPortId;
                g_usAtmInterleavedPortId = KArg.usAtmInterleavedPortId;
            put_user( BCMADSL_STATUS_SUCCESS, &((PADSLDRV_MAP_ATM_PORT) arg)->bvStatus );
        }
} /* DoMapATMPortIDs */


/***************************************************************************
 * Function Name: DoGetConnectionInfo
 * Description  : Calls BcmAdsl_GetConnectionInfo on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoGetConnectionInfo( unsigned long arg )
{
    ADSLDRV_CONNECTION_INFO KArg;

    KArg.bvStatus = BcmAdsl_GetConnectionInfo( &KArg.ConnectionInfo );
    copy_to_user( (void *) arg, &KArg, sizeof(KArg) );
} /* DoGetConnectionInfo */


/***************************************************************************
 * Function Name: DoDiagCommand
 * Description  : Calls BcmAdsl_DiagCommand on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoDiagCommand ( unsigned long arg )
{
    ADSLDRV_DIAG        KArg;
        char                    cmdData[LOG_MAX_DATA_SIZE];

        KArg.bvStatus = BCMADSL_STATUS_ERROR;
        do {
                if (copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0)
                        break;

                if ((LOG_CMD_CONNECT != KArg.diagCmd) && (0 != KArg.diagMap)) {
                        if (KArg.logTime > sizeof(cmdData))
                                break;
                        if (copy_from_user(cmdData, (char*)KArg.diagMap, KArg.logTime) != 0)
                                break;

                        KArg.diagMap = (int) cmdData;
                }

        KArg.bvStatus = BcmAdsl_DiagCommand ( (PADSL_DIAG) &KArg );
        } while (0);
        put_user( KArg.bvStatus, &((PADSLDRV_DIAG) arg)->bvStatus );
} /* DoGetConnectionInfo */

/***************************************************************************
 * Function Name: DoGetObjValue
 * Description  : Calls BcmAdsl_GetObjectValue on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoGetObjValue( unsigned long arg )
{
        ADSLDRV_GET_OBJ KArg;
        char                    objId[kOidMaxObjLen];
        int                             retObj;
		BcmAdslCoreDiagWriteStatusString("*******DoGetObjValue function, g_ecoEvent \n");
        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;

                if ((NULL != KArg.objId) && (KArg.objIdLen)) {
                        if (KArg.objIdLen > kOidMaxObjLen)
                                break;
                        if( copy_from_user( objId, KArg.objId, KArg.objIdLen ) != 0 )
                                break;
                        retObj = BcmAdsl_GetObjectValue (objId, KArg.objIdLen, NULL, &KArg.dataBufLen);
                }
                else
                        retObj = BcmAdsl_GetObjectValue (NULL, 0, NULL, &KArg.dataBufLen);

                if ((retObj >= kAdslMibStatusLastError) && (retObj < 0))
                        break;

                copy_to_user( KArg.dataBuf, (void *) retObj, KArg.dataBufLen );
                KArg.bvStatus = BCMADSL_STATUS_SUCCESS;
        } while (0);

        if (BCMADSL_STATUS_ERROR == KArg.bvStatus)
                KArg.dataBufLen = 0;
        copy_to_user( (void *) arg, &KArg, sizeof(KArg) );
}


/***************************************************************************
 * Function Name: DoSetObjValue
 * Description  : Calls BcmAdsl_SetObjectValue on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoSetObjValue( unsigned long arg )
{
        ADSLDRV_GET_OBJ KArg;
        char                    objId[kOidMaxObjLen];
        char					dataBuf[100];
		int                     retObj;

		BcmAdslCoreDiagWriteStatusString("***********DoSetObjValue function, g_ecoEvent \n");
        do {
				
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;
				if ((NULL != KArg.objId) && (KArg.objIdLen)) {
                        if (KArg.objIdLen > kOidMaxObjLen)
                                break;
                        if( copy_from_user( objId, KArg.objId, KArg.objIdLen ) != 0 )
                                break;
						else if (copy_from_user( dataBuf, KArg.dataBuf, KArg.dataBufLen ) != 0 )
								break;
						retObj = BcmAdsl_SetObjectValue (objId, KArg.objIdLen, dataBuf, &KArg.dataBufLen);
                }
                
                if ((retObj >= kAdslMibStatusLastError) && (retObj < 0))
                        break;
                KArg.bvStatus = BCMADSL_STATUS_SUCCESS;
        } while (0);

        if (BCMADSL_STATUS_ERROR == KArg.bvStatus)
                KArg.dataBufLen = 0;
        //copy_to_user( (void *) arg, &KArg, sizeof(KArg) );
}

/***************************************************************************
 * Function Name: DoStartBERT
 * Description  : Calls BcmAdsl_StartBERT on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoStartBERT( unsigned long arg )
{
        ADSLDRV_BERT    KArg;

    if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) == 0 ) {
        KArg.bvStatus = BcmAdsl_StartBERT (KArg.totalBits);
        put_user( KArg.bvStatus, &((PADSLDRV_BERT) arg)->bvStatus );
    }
}

/***************************************************************************
 * Function Name: DoStopBERT
 * Description  : Calls BcmAdsl_StopBERT on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoStopBERT( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_StopBERT();
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
}

/***************************************************************************
 * Function Name: DoConfigure
 * Description  : Calls BcmAdsl_Configureon behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoConfigure( unsigned long arg )
{
        ADSLDRV_CONFIGURE       KArg;
        adslCfgProfile          adslCfg;

        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;

                if (NULL != KArg.pAdslCfg) {
                        if( copy_from_user( &adslCfg, KArg.pAdslCfg, sizeof(adslCfg)) != 0 )
                                break;
                        KArg.bvStatus = BcmAdsl_Configure(&adslCfg);
                }
                else
                        KArg.bvStatus = BCMADSL_STATUS_SUCCESS;
        } while (0);

        put_user( KArg.bvStatus, &((PADSLDRV_CONFIGURE) arg)->bvStatus );
}

/***************************************************************************
 * Function Name: DoTest
 * Description  : Controls ADSl driver/PHY special test modes
 * Returns      : None.
 ***************************************************************************/
static void DoTest( unsigned long arg )
{
        ADSLDRV_TEST            KArg;
        char                            rcvToneMap[512/8];
        char                            xmtToneMap[512/8];
        int                                     nToneBytes;

        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;

                if ((NULL != KArg.xmtToneMap) && (0 != KArg.xmtNumTones)) {
                        nToneBytes = (KArg.xmtNumTones + 7) >> 3;
                        if (nToneBytes > sizeof(xmtToneMap)) {
                                nToneBytes = sizeof(xmtToneMap);
                                KArg.xmtNumTones = nToneBytes << 3;
                        }
                        if( copy_from_user( xmtToneMap, KArg.xmtToneMap, nToneBytes) != 0 ) {
                                KArg.xmtNumTones = 0;
                                break;
                        }
                        KArg.xmtToneMap = xmtToneMap;
                }

                if ((NULL != KArg.rcvToneMap) && (0 != KArg.rcvNumTones)) {
                        nToneBytes = (KArg.rcvNumTones + 7) >> 3;
                        if (nToneBytes > sizeof(rcvToneMap)) {
                                nToneBytes = sizeof(rcvToneMap);
                                KArg.rcvNumTones = nToneBytes << 3;
                        }
                        if( copy_from_user( rcvToneMap, KArg.rcvToneMap, nToneBytes) != 0 ) {
                                KArg.rcvNumTones = 0;
                                break;
                        }
                        KArg.rcvToneMap = rcvToneMap;
                }
                
                if (ADSL_TEST_SELECT_TONES != KArg.testCmd)
                        KArg.bvStatus = BcmAdsl_SetTestMode(KArg.testCmd);
                else
                        KArg.bvStatus = BcmAdsl_SelectTones(
                                KArg.xmtStartTone, KArg.xmtNumTones,
                                KArg.rcvStartTone, KArg.rcvNumTones,
                                xmtToneMap, rcvToneMap);
        } while (0);

        put_user( KArg.bvStatus, &((PADSLDRV_TEST) arg)->bvStatus );
}

/***************************************************************************
 * Function Name: DoGetConstelPoints
 * Description  : Calls BcmAdsl_GetObjectValue on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoGetConstelPoints( unsigned long arg )
{
        ADSLDRV_GET_CONSTEL_POINTS      KArg;
        ADSL_CONSTELLATION_POINT        pointBuf[64];
        int                                                     numPoints;

        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;

                numPoints = (KArg.numPoints > (sizeof(pointBuf)/sizeof(pointBuf[0])) ? sizeof(pointBuf)/sizeof(pointBuf[0]) : KArg.numPoints);
                numPoints = BcmAdsl_GetConstellationPoints (KArg.toneId, pointBuf, KArg.numPoints);
                if (numPoints > 0)
                        copy_to_user( KArg.pointBuf, (void *) pointBuf, numPoints * sizeof(ADSL_CONSTELLATION_POINT) );
                KArg.numPoints = numPoints;
                KArg.bvStatus = BCMADSL_STATUS_SUCCESS;
        } while (0);

        if (BCMADSL_STATUS_ERROR == KArg.bvStatus)
                KArg.numPoints = 0;
        copy_to_user( (void *) arg, &KArg, sizeof(KArg) );
}

static void DoGetVersion( unsigned long arg )
{
        ADSLDRV_GET_VERSION             KArg;
        adslVersionInfo                 adslVer;

        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }

                BcmAdsl_GetVersion(&adslVer);
                copy_to_user( KArg.pAdslVer, (void *) &adslVer, sizeof(adslVersionInfo) );
                KArg.bvStatus = BCMADSL_STATUS_SUCCESS;
        } while (0);

        put_user( KArg.bvStatus , &((PADSLDRV_GET_VERSION) arg)->bvStatus );
}

static void DoSetSdramBase( unsigned long arg )
{
        ADSLDRV_SET_SDRAM_BASE          KArg;

    if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) == 0 ) {
        KArg.bvStatus = BcmAdsl_SetSDRAMBaseAddr((void *)KArg.sdramBaseAddr);
        put_user( KArg.bvStatus, &((PADSLDRV_SET_SDRAM_BASE) arg)->bvStatus );
    }
}

static void DoResetStatCounters( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_ResetStatCounters();
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
}

static void DoSetOemParameter( unsigned long arg )
{
        ADSLDRV_SET_OEM_PARAM   KArg;
        char                                    dataBuf[256];

        do {
                if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) != 0 ) {
                        KArg.bvStatus = BCMADSL_STATUS_ERROR;
                        break;
                }
                KArg.bvStatus = BCMADSL_STATUS_ERROR;

                if ((NULL != KArg.buf) && (KArg.len > 0)) {
                        if (KArg.len > sizeof(dataBuf))
                                KArg.len = sizeof(dataBuf);
                        if( copy_from_user( dataBuf, (void *)KArg.buf, KArg.len) != 0 )
                                break;
                        KArg.bvStatus = BcmAdsl_SetOemParameter (KArg.paramId, dataBuf, KArg.len);
                }
                else
                        KArg.bvStatus = BCMADSL_STATUS_SUCCESS;
        } while (0);

        put_user( KArg.bvStatus, &((PADSLDRV_SET_OEM_PARAM) arg)->bvStatus );
}

/***************************************************************************
 * Function Name: DoBertStartEx
 * Description  : Calls BcmAdsl_BertStartEx on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoBertStartEx( unsigned long arg )
{
        ADSLDRV_BERT_EX         KArg;

    if( copy_from_user( &KArg, (void *) arg, sizeof(KArg) ) == 0 ) {
        KArg.bvStatus = BcmAdsl_BertStartEx (KArg.totalSec);
        put_user( KArg.bvStatus, &((PADSLDRV_BERT_EX) arg)->bvStatus );
    }
}

/***************************************************************************
 * Function Name: DoBertStopEx
 * Description  : Calls BcmAdsl_BertStopEx on behalf of a user program.
 * Returns      : None.
 ***************************************************************************/
static void DoBertStopEx( unsigned long arg )
{
    BCMADSL_STATUS bvStatus = BcmAdsl_BertStopEx();
    put_user( bvStatus, &((PADSLDRV_STATUS_ONLY) arg)->bvStatus );
}

/***************************************************************************
 * Function Name: AdslConnectCb
 * Description  : Callback function that is called when by the ADSL driver
 *                when there is a change in status.
 * Returns      : None.
 ***************************************************************************/
static void AdslConnectCb( ADSL_LINK_STATE AdslLinkState, UINT32 ulParm )
{
    UINT32 ulInterfaceId;
    ADSL_CONNECTION_INFO AdslConnInfo;
    ATM_INTERFACE_LINK_INFO AtmLinkInfo;

#ifdef DYING_GASP_API
    /* Wake up the user mode application that monitors link status. */
    kerSysWakeupMonitorTask();
#endif

    AtmLinkInfo.ulLinkState = LINK_DOWN;
    switch (AdslLinkState)
      {
      case ADSL_LINK_UP:
        AtmLinkInfo.ulLinkState = LINK_UP;
        //kerSysLedCtrl(kLedAdsl, kLedStateOn);
	//Add by Ron
	led_on(GPIO_ADSL_ACT, 1);
	led_blink(GPIO_ADSL_DOWN,0,LED_SLOW,0);
	led_off(GPIO_ADSL_DOWN, 1);
#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
        snmp_adsl_eoc_event();
#endif
        break;
      case ADSL_LINK_DOWN:
      case BCM_ADSL_ATM_IDLE:
        if (AdslLinkState == ADSL_LINK_DOWN) {
            //kerSysLedCtrl(kLedAdsl, kLedStateFail);
	    //Add by Ron
	    led_off(GPIO_ADSL_ACT, 1);
	    led_blink(GPIO_ADSL_DOWN,0,LED_SLOW,0);
	    led_off(GPIO_ADSL_DOWN, 1);
#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
            g_eocEvent |= ADSL_LINK_DROPPED;
#endif 
        }
        break;
      case BCM_ADSL_TRAINING_G992_EXCHANGE:
      case BCM_ADSL_TRAINING_G992_CHANNEL_ANALYSIS:
      /* Ron change for led blink */
      //case BCM_ADSL_TRAINING_G992_STARTED:
      //  kerSysLedCtrl(kLedAdsl, kLedStateFastBlinkContinues);
	led_blink(GPIO_ADSL_DOWN,1000,LED_SLOW,0);
        break;
#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
      case BCM_ADSL_G997_FRAME_RECEIVED:
        g_eocEvent |= ADSL_EOC_FRAME_RCVD;
 #if 0 /* no link up message is checked right now */
        if ((g_eocEvent & ADSL_LINK_DROPPED) == ADSL_LINK_DROPPED)
          AdslCheckLinkupMsg();
#endif
        return;
      case BCM_ADSL_G997_FRAME_SENT:
        g_eocEvent |= ADSL_EOC_FRAME_SENT;
	if (outBuffer) {
	  kfree(outBuffer);
	  outBuffer=NULL;
	}
#ifdef BUILD_SNMP_TRANSPORT_DEBUG
        printk("BCM_ADSL_G997_FRAME_SENT event, g_ecoEvent %x\n",g_eocEvent);
#endif
        return;
#endif /* defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO) */
      default:
        return;
      }
    /* record time stamp of link state change (in 1/100ths of a second unit per MIB spec.) */
    if (g_dslNetDev != NULL) 
      g_dslNetDev->trans_start = jiffies * 100 / HZ;

    BcmAdsl_GetConnectionInfo( &AdslConnInfo );
    AtmLinkInfo.ulStructureId = ID_ATM_INTERFACE_LINK_INFO;

    /* For now, only ATM port 0 is supported */
    BcmAtm_GetInterfaceId( 0, &ulInterfaceId );
    AtmLinkInfo.ulLineRate = (AdslConnInfo.ulFastUpStreamRate) ? AdslConnInfo.ulFastUpStreamRate
        : AdslConnInfo.ulInterleavedUpStreamRate;
    BcmAtm_SetInterfaceLinkInfo( ulInterfaceId, &AtmLinkInfo );

    if (AdslLinkState == BCM_ADSL_ATM_IDLE) {
        AtmLinkInfo.ulLinkState = LINK_UP;
                BcmAtm_SetInterfaceLinkInfo( ulInterfaceId, &AtmLinkInfo );
    }
} /* AdslConnectCb */


#if defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO)
/* routine copied from tiny bridge written by lt */
void snmp_adsl_eoc_event(void)
{
  adslMibInfo adslMib;
  long size = sizeof(adslMib);
  
  if( BcmAdsl_GetObjectValue(NULL, 0, (char *)&adslMib, &size) ==
      BCMADSL_STATUS_SUCCESS )
    {
      if( kAdslModAdsl2p == adslMib.adslConnection.modType ||
          kAdslModAdsl2  == adslMib.adslConnection.modType )
        {
          printk("ADSL2/ADSL2+ connection\n");
          g_eoc_hdr_offset = ADSL_2P_HDR_OFFSET;
          g_eoc_hdr_len    = ADSL_2P_EOC_HDR_LEN; 
        }
      else
        {
          printk("ADSL1 connection\n");
          g_eoc_hdr_offset = ADSL_HDR_OFFSET;
          g_eoc_hdr_len    = ADSL_EOC_HDR_LEN; 
          
        }
    }
}

void AdslCheckLinkupMsg(void) {
  char *p;
  int len;
  unsigned char adsl_eoc_enable[] = ADSL_EOC_ENABLE;

  p = BcmAdsl_G997FrameGet( &len );
  if( p && len > 0 )
    {

#ifdef BUILD_SNMP_TRANSPORT_DEBUG
      printk("AdslCheckLinkupMsg(void), G997FrameGet returns len %d, p %x, expects ff 03 00 57 \n",len,(int)p);
      dumpaddr(p,len);
#endif

      if( memcmp(p, adsl_eoc_enable, sizeof(adsl_eoc_enable)) == 0 ) {
#if 0
        g_eocEvent &= ~(ADSL_LINK_DROPPED | ADSL_EOC_FRAME_RCVD);
#else
        g_eocEvent &= ~(ADSL_LINK_DROPPED);
#endif
      }

#if 1
        p = BcmAdsl_G997FrameGetNext( &len );
        while ( p && len > 0 ) {
          printk("*** AdslCheckLinkupMsg: Got next SNMP request segment on ADSL EOC "
                 "connection ***\n");
          dumpaddr( p, len );
        }
#endif
        BcmAdsl_G997FrameFinished();
    } /* p & len > 0 */
}

/***************************************************************************
 * Function Name: adsl_read
 * The EOC read code (slightly modified) is copied from cfeBridge written by lat
 ***************************************************************************/
static ssize_t adsl_read(struct file *file, char __user *buf,
                         size_t count, loff_t *ppos)
{
    int ret = 0, len, request_length=0;
    char *p;
    unsigned char adsl_eoc_hdr[] = ADSL_EOC_HDR;
    unsigned char adsl_eoc_enable[] = ADSL_EOC_ENABLE;
    char *inputBuffer=NULL;

#ifdef BUILD_SNMP_TRANSPORT_DEBUG
    printk("adsl_read(entry): g_eocEvent %x\n",g_eocEvent);
#endif

    if ((g_eocEvent & ADSL_EOC_FRAME_RCVD) != ADSL_EOC_FRAME_RCVD)
      return ret;
    
    /* Receive SNMP request. */
    p = BcmAdsl_G997FrameGet( &len );
    if( p && len > 0 )
      {
        if( memcmp(p, adsl_eoc_enable, sizeof(adsl_eoc_enable)) == 0 ) {
          g_eocEvent &= ~(ADSL_LINK_DROPPED | ADSL_EOC_FRAME_RCVD);
          /* return now and let the upper layer know that link up is received */
          /* don't clear the FRAME_RCVD bit, just in case there are things to read */
#ifdef BUILD_SNMP_TRANSPORT_DEBUG
	        printk("*** Got message up on ADSL EOC connection, len %d ***\n",len);
      		 dumpaddr(p,len);
#endif
          BcmAdsl_G997FrameFinished();
          return len;
        }

        if (memcmp(p, adsl_eoc_hdr+g_eoc_hdr_offset, g_eoc_hdr_len) == 0) {
          p += g_eoc_hdr_len;
          len -= g_eoc_hdr_len;
        }
        else {
          printk("something is wrong, len %d, no adsl_eoc_hdr found\n",len);
          dumpaddr(p,len);
        }


        inputBuffer = kmalloc(count,GFP_KERNEL);
        do
          {
#ifdef BUILD_SNMP_TRANSPORT_DEBUG
        	printk("***malloc input buffer size of %d, processing rx from EOC link, len %d/requestLen %d ***\n",
        		count,len,request_length);
        	dumpaddr(p,len);
#endif
            if( len + request_length < count )
              {
                memcpy(inputBuffer + request_length, p, len);
                request_length += len;
              }
            p = BcmAdsl_G997FrameGetNext( &len );
          } while( p && len > 0 );
#if 0
        g_eocEvent &= ~(ADSL_EOC_FRAME_RCVD);
#endif
        BcmAdsl_G997FrameFinished();
        if (!(copy_to_user(buf, inputBuffer, request_length)))
          ret = request_length;

#ifdef BUILD_SNMP_TRANSPORT_DEBUG
        printk("adsl_read(): end, entire g_eocEvent %x, request_length %d\n",
               g_eocEvent,request_length);
        printk("inputBuffer:\n");
        dumpaddr(inputBuffer,request_length);
#endif
        kfree(inputBuffer);
      } /* (p  & len > 0) */
    g_eocEvent &= ~(ADSL_EOC_FRAME_RCVD);
    return( ret );
} /* adsl_read */

/***************************************************************************
 * Function Name: adsl_write
 * The EOC write code is copied from cfeBridge written by lat
 ***************************************************************************/
static ssize_t adsl_write(struct file *file, const char __user *buf,
                          size_t count, loff_t *ppos)
{
  unsigned char adsl_eoc_hdr[] = ADSL_EOC_HDR;
  int ret = 0, len;
  int retry = 0;

#ifdef BUILD_SNMP_TRANSPORT_DEBUG
  printk("adsl_write(entry): count %d\n",count);
#endif    

  /* first of all, check to see if ADSL driver finished sending */
  while ( ((g_eocEvent & ADSL_EOC_FRAME_SENT) != ADSL_EOC_FRAME_SENT) && 
	  (retry < 10) ) {
#ifdef BUILD_SNMP_TRANSPORT_DEBUG
      printk("g_eocEvent %x (ADSL_LINK_DROPPED %x, ADSL_EOC_FRAME_SENT %x), retry %d\n",
             g_eocEvent, ADSL_LINK_DROPPED, ADSL_EOC_FRAME_SENT, retry);
#endif
      /* will wait for maximum of 5 second for link to come up */
      bcmOsSleep(HZ/2);
      retry++;
  }
  if (retry == 10) {
    printk("ADSL not ready to send: g_eocEvent %x; ADSL_LINK_DROPPED %x or ADSL_EOC_FRAME_SENT %x\n",g_eocEvent,
	   ADSL_LINK_DROPPED,ADSL_EOC_FRAME_SENT);
    return ret;
  }

  /* need to add 4 bytes adsl_eoc_hdr */
  len = count + g_eoc_hdr_len;
  if ((outBuffer = kmalloc(len,GFP_KERNEL)) == NULL)
    return ret;

  memcpy(outBuffer,adsl_eoc_hdr+g_eoc_hdr_offset,g_eoc_hdr_len);
  if (copy_from_user((outBuffer+g_eoc_hdr_len), buf, count) == 0) {
    g_eocEvent &= ~ADSL_EOC_FRAME_SENT;

#ifdef BUILD_SNMP_TRANSPORT_DEBUG
    printk("adsl_write(ready to send): sending frame over EOC, len %d (gocEvent %x):\n",len,g_eocEvent);
    dumpaddr(outBuffer,len);
#endif

    if(BcmAdsl_G997SendData(outBuffer,len) != BCMADSL_STATUS_SUCCESS) {
      g_eocEvent |= ADSL_EOC_FRAME_SENT;
#ifdef BUILD_SNMP_TRANSPORT_DEBUG
      printk("BcmAdsl_G997SendData(): returns error\n");
#endif
      kfree(outBuffer);
      outBuffer=NULL;
    }
    else 
      ret = len;
  }
  return ret;
} 

static unsigned int adsl_poll(struct file *file, poll_table *wait)
{
    int ret = 0;
    
    /* is it necessary to do write event (for retransmit?),
       and exception event (link down) */
    if ((g_eocEvent & ADSL_EOC_FRAME_RCVD) == ADSL_EOC_FRAME_RCVD) {
      ret = 1;
    }
    return( ret );  
}
#endif /* defined(BUILD_SNMP_EOC) || defined(BUILD_SNMP_AUTO) */

/***************************************************************************
 * MACRO to call driver initialization and cleanup functions.
 ***************************************************************************/
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)) || !defined(MODULE)
module_init( adsl_init );
module_exit( adsl_cleanup );
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
MODULE_LICENSE("Proprietary");
#endif

