/*
 * Common code for wl command-line swiss-army-knife utility
 *
 * Policy notes:
 * This source file generates both the "wl" command and the
 * more restricted "wl-tool" manufacturing test command (-DWL_TOOL).
 * - wl driver wlc_ioctl() commands which are enclosed within #ifdef BCMINTERNAL
 * should be enclosed within #ifdef BCMINTERNAL in wlu.c .
 * - exposing any new commands in wl-tool requires an "ok" from romanb.
 *
 * $Id$
 */

#if	defined(_CFE_)
#include <lib_types.h>
#include <lib_string.h>
#include <lib_printf.h>
#include <lib_malloc.h>
#include <cfe_error.h>
#elif	defined(_HNDRTE_)
#include <hndrte.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#endif

#include <typedefs.h>
#include <epivers.h>
#include <proto/ethernet.h>
#include <proto/802.11.h>
#include <proto/eapol.h>
#include <wlioctl.h>
#include <bcmutils.h>
#include <bcmendian.h>
#include "wlu.h"

#if	defined(linux)
#define stricmp strcasecmp
#define strnicmp strncasecmp
#elif	defined(UNDER_CE)
#define stricmp _stricmp
#define strnicmp _strnicmp
#elif	defined(_HNDRTE_)
#include <bcmutils.h>
#define fprintf(stream, fmt, args...) printf(fmt, ## args)
#define fputs(s, stream) printf("%s", s)
#define malloc(size) hndrte_malloc(size)
#define free(ptr) hndrte_free(ptr)
#define isalnum(c) bcm_isalnum(c)
#define isalpha(c) bcm_isalpha(c)
#define iscntrl(c) bcm_iscntrl(c)
#define isdigit(c) bcm_isdigit(c)
#define isgraph(c) bcm_isgraph(c)
#define islower(c) bcm_islower(c)
#define isprint(c) bcm_isprint(c)
#define ispunct(c) bcm_ispunct(c)
#define isspace(c) bcm_isspace(c)
#define isupper(c) bcm_isupper(c)
#define isxdigit(c) bcm_isxdigit(c)
#define strspn(s, accept) (0)
#define stricmp strcmp
#define strnicmp strncmp
#define atoi(nptr) ((int)bcm_strtoul((nptr), NULL, 10))
#define strtoul(nptr, endptr, base) bcm_strtoul((nptr), (endptr), (base))
#define strtol(nptr, endptr, base) ((long)bcm_strtoul((nptr), (endptr), (base)))
#define tolower(c) (bcm_isupper((c)) ? ((c) + 'a' - 'A') : (c))
#define toupper(c) (bcm_toupper((c)))
#elif	defined(_CFE_)
#include <bcmutils.h>
#include <osl.h>
#define isalnum(c) bcm_isalnum(c)
#define isalpha(c) bcm_isalpha(c)
#define iscntrl(c) bcm_iscntrl(c)
#define isdigit(c) bcm_isdigit(c)
#define isgraph(c) bcm_isgraph(c)
#define islower(c) bcm_islower(c)
#define isprint(c) bcm_isprint(c)
#define ispunct(c) bcm_ispunct(c)
#define isspace(c) bcm_isspace(c)
#define isupper(c) bcm_isupper(c)
#define isxdigit(c) bcm_isxdigit(c)
#define stricmp(s1, s2) lib_strcmpi((s1), (s2))
#define strtoul(nptr, endptr, base) bcm_strtoul((nptr), (endptr), (base))
#define tolower(c) (bcm_isupper((c)) ? ((c) + 'a' - 'A') : (c))
#define fprintf(stream, fmt, args...) xprintf(fmt, ## args)
#define fputs(s, stream) puts(s)
#define malloc(size) KMALLOC((size), 0)
#define free(ptr) KFREE(ptr)
#define strnicmp(s1, s2, len) strncmp((s1), (s2), (len))
#define strspn(s, accept) (0)
#define strtol(nptr, endptr, base) bcm_strtoul((nptr), (endptr), (base))
#endif

#define MIN(a, b) (((a)<(b))?(a):(b))

cmd_func_t wl_int;
static cmd_func_t wl_print_deprecate;
static cmd_func_t wl_void, wl_rssi, wl_locale, wl_gmode;
static cmd_func_t wlu_dump, wlu_srdump, wl_rate_mrate, wl_phy_rate;
static cmd_func_t wl_channel, wl_radio, wl_version, wl_list;
static cmd_func_t wl_band, wl_bandlist, wl_phylist, wl_out;
static cmd_func_t wl_join, wl_tssi, wl_txpwr, wl_atten, wl_evm, wl_country;
static cmd_func_t wl_txpwr1;
static cmd_func_t wl_maclist, wl_get_pktcnt, wl_upgrade;
static cmd_func_t wl_maclist_1;
static cmd_func_t wl_rateset, wl_interfere;
#ifdef DSLCPE
static cmd_func_t wl_aci_args;
#endif
static cmd_func_t wl_radar_args, wl_dfs_status;
#ifdef DSLCPE
static cmd_func_t wl_get_txpwr_limit, wl_get_current_power, wl_get_instant_power;
#else
static cmd_func_t wl_get_txpwr_limit, wl_get_current_power;
#endif
static cmd_func_t wl_var_get, wl_var_getint, wl_var_getandprintstr;
static cmd_func_t wl_var_setint, wl_varint, wl_addwep, wl_rmwep;
static cmd_func_t wl_nvdump, wl_nvget, wl_nvset, wl_sta_info, wl_chan_info;
static cmd_func_t wl_wme_ac_req, wl_add_ie, wl_del_ie, wl_list_ie;
static cmd_func_t wl_wme_apsd_sta, wl_wme_dp;
static cmd_func_t wl_rand, wl_otpw, wl_counters, wl_assoc_info, wl_wme_counters;
static cmd_func_t wl_bitvec128, wl_diag, wl_var_void;
#ifdef BCMWPA2
static cmd_func_t wl_pmkid_info;
#endif

static int wl_parse_rateset(wl_rateset_t* rs, char **argv);

static cmd_func_t wl_dump_networks, wl_mac, wl_revinfo;
static void dump_networks(char *buf);
static void dump_bss_info(wl_bss_info_t *bi);



static cmd_func_t wl_auto_channel_sel;

static cmd_func_t wl_bsscfg_int, wl_bsscfg_enable;
static cmd_func_t wl_msglevel, wl_plcphdr, wl_reg, wl_macreg, wl_band_elm;
static cmd_func_t wl_rateparam, wl_wepstatus, wl_status, wl_spect;
static cmd_func_t wl_sup_rateset, wl_scan, wl_send_csa;
static cmd_func_t wl_dump_chanlist, wl_primary_key, wl_measure_req, wl_send_quiet;
static cmd_func_t wl_wsec, wl_keys,  wl_wsec_test; 
static cmd_func_t wl_channels_in_country;
static cmd_func_t wl_wpa_auth, wl_tsc, wl_deauth_rc, wl_ssid;
static cmd_func_t wl_wds_wpa_role_old, wl_wds_wpa_role, wl_set_pmk;
static cmd_func_t wl_rm_request, wl_rm_report;
static cmd_func_t wl_join_pref, wl_assoc_pref;

void dump_rateset(uint8 *rates, uint count);
int wl_parse_channel_list(char* list_str, uint16* channel_list, int channel_num);

static uint16 wl_qdbm_to_mw(uint8 qdbm);
static uint8 wl_mw_to_qdbm(uint16 mw);


static int wl_cfg_option(char **argv, char *fn_name, int *bsscfg_idx, int *consumed);
static int get_oui_bytes (uchar *oui_str, uchar *oui);
static int get_ie_data (uchar *data_str, uchar *ie_data, int len);
static void wl_printrate(int val);
static int rate_string2int(char *s);
static char *rate_int2string(char *buf, int val);

static int wl_get_scan(void *wl, int opc, char *buf, uint buf_len);
static int wl_var_getbuf(void *wl, char *iovar, void *param, int param_len, void **bufptr);
static int wl_var_setbuf(void *wl, char *iovar, void *param, int param_len);

static uint wl_iovar_mkbuf(char *name, char *data, uint datalen, char *buf, uint buflen, int *perr);
static int wl_iovar_getint(void *wl, char *name, int *var);
static int wl_iovar_setint(void *wl, char *name, int var);

static int wl_bssiovar_mkbuf(char *iovar, int bssidx, void *param, int paramlen, void *bufptr, int buflen, int *perr);
static int wl_bssiovar_setbuf(void* wl, char *iovar, int bssidx, void *param, int paramlen, void *bufptr, int buflen);
static int wl_bssiovar_getbuf(void* wl, char *iovar, int bssidx, void *param, int paramlen, void *bufptr, int buflen);
static int wl_bssiovar_set(void *wl, char *iovar, int bssidx, void *param, int paramlen);
static int wl_bssiovar_get(void *wl, char *iovar, int bssidx, void *outbuf, int len);
static int wl_bssiovar_setint(void *wl, char *iovar, int bssidx, int val);
static int wl_bssiovar_getint(void *wl, char *iovar, int bssidx, int *pval);

static int wl_vndr_ie(void *wl, char *command, char **argv);
static int hexstrtobitvec(const uchar *cp, uchar *bitvec, int veclen);

#define MINIOPT_MAXKEY	128
typedef struct miniopt {
	char* name;		/* name for prompt in error strings */
	char* flags;		/* set of chars that are option flags
				 * with no args */
	int consumed;		/* number of argv entries cosumed in
				 * the most recent call to miniopt() */
	bool opt_end;		/* at end of options since a "--" flag
				 * was encountered */
	bool positional;	
	bool good_int;		/* 'val' member is the result of a sucessful
				 * strtol conversion of the option value */
	char opt;		
	char key[MINIOPT_MAXKEY];	
	char* valstr;		/* positional param, or value for the option,
				 * or null if the option had
				 * no accompanying value */
	int val;		/* strtol translation of valstr */
} miniopt_t;

void miniopt_init(miniopt_t *t, char* name, char* flags);
int miniopt(miniopt_t *t, char **argv);

#define WL_DUMP_BUF_LEN (64 * 1024)
#define OUI_STR_SIZE	8
#define MAX_OUI_SIZE	3
#define MAX_BYTE_CHARS	2
#define MAX_DATA_COLS	16
#ifdef DSLCPE
#define DIV_QUO(num, div) ((num)/div)  /* Return the quotient of division to avoid floats */
#define DIV_REM(num, div) (((num%div) * 100)/div) /* Return the remainder of division */
#endif

static char buf[WLC_IOCTL_MAXLEN];

static char *wlc_ww[] = { "Worldwide", NULL };
static char *wlc_thailand[] = { "Thailand", NULL };
static char *wlc_israel[] = { "Israel", NULL };
static char *wlc_jordan[] = { "Jordan", NULL };
static char *wlc_china[] = { "P.R. China", "China", NULL };
static char *wlc_japan[] = { "Japan", NULL };
static char *wlc_fcc[] = { "USA", "Canada", "ANZ", "New Zealand", "FCC", NULL };
static char *wlc_eur[] = { "Europe", NULL };
static char *wlc_usl[] = { "USAlow", NULL };
static char *wlc_jph[] = { "JapanHigh", NULL };
static char *wlc_all[] = { "AllTheChannels", NULL };
static char *wlc_11d[] = { "802.11d", NULL };
static struct {
	int val;
	char **names;
} wl_locales[] = {
	{ WLC_WW, wlc_ww },
	{ WLC_THA, wlc_thailand },
	{ WLC_ISR, wlc_israel },
	{ WLC_JDN, wlc_jordan },
	{ WLC_PRC, wlc_china },
	{ WLC_JPN, wlc_japan },
	{ WLC_FCC, wlc_fcc },
	{ WLC_EUR, wlc_eur },
	{ WLC_USL, wlc_usl },
	{ WLC_JPH, wlc_jph },
	{ WLC_ALL, wlc_all },
	{ WLC_11D, wlc_11d }
};
/* integer output format, default to signed integer */
static uint8 int_fmt = INT_FMT_DEC;

/*
 * Country names and abbreviations from ISO 3166
 */
typedef struct {
	char *name;	/* Long name */
	char *abbrev;	/* Abbreviation */
} cntry_name_t;
cntry_name_t cntry_names[];	/* At end of this file */ 


cmd_t wl_cmds[] = {
	{ "ver", wl_version, -1, -1,
	  "get version information" },
	{ "cmds", wl_list, -1, -1,
	  "generate a short list of available commands"},
	{ "up",	wl_void, -1, WLC_UP,
	  "reinitialize and mark adapter up (operational)" },
	{ "down", wl_void, -1, WLC_DOWN,
	  "reset and mark adapter down (disabled)" },
#ifndef WME_ONLY
	{ "out", wl_out, -1, WLC_OUT,
	  "mark adapter down but do not reset hardware(disabled)\n"
	  "\tOn dualband cards, cards must be bandlocked before use."},
	{ "clk", wl_int, WLC_GET_CLK, WLC_SET_CLK,
	  "set board clock state. return error for set_clk attempt if the driver is not down\n"
	  "\t0: clock off\n"
	  "\t1: clock on" },
	{ "restart", wl_void, -1, WLC_RESTART,
	  "Restart driver.  Driver must already be down."},
	{ "reboot", wl_void, -1, WLC_REBOOT,
	  "Reboot platform"},
	{ "ucflags", wl_int, WLC_GET_UCFLAGS, WLC_SET_UCFLAGS,
	  "Get/Set ucode flags"},
	{ "radio", wl_radio, WLC_GET_RADIO, WLC_SET_RADIO,
	  "Set the radio on or off.\n"
	  "\t\"on\" or \"off\"" },
	{ "dump", wlu_dump, WLC_DUMP, -1,
	  "print driver software state and chip registers to stdout" },
	{ "srdump", wlu_srdump, WLC_GET_SROM, -1,
	  "print contents of SPROM to stdout" },
	{ "nvdump", wl_nvdump, WLC_NVRAM_DUMP, -1,
	  "print nvram variables to stdout" },
	{ "nvset", wl_nvset, -1, WLC_NVRAM_SET,
	  "set an nvram variable\n"
	  "\tname=value (no spaces around \'=\')\n" },
	{ "nvget", wl_nvget, WLC_NVRAM_GET, -1,
	  "get the value of an nvram variable\n" },
	{ "revinfo", wl_revinfo, WLC_GET_REVINFO, -1,
	  "get hardware revision information\n" },
	{ "msglevel", wl_msglevel, WLC_GET_MSGLEVEL, WLC_SET_MSGLEVEL,
	  "set driver console debugging message bitvector\n"
	  "\ttype \'wl msglevel ?\' for values" },
	{ "PM", wl_int, WLC_GET_PM, WLC_SET_PM,
	  "set driver power management mode:\n"
	  "\t0: CAM (constantly awake)\n"
	  "\t1: PS  (power-save)\n" 
	  "\t2: FAST PS mode" },
	{ "wake", wl_int, WLC_GET_WAKE, WLC_SET_WAKE,
	  "set driver power-save mode sleep state:\n"
	  "\t0: core-managed\n"
	  "\t1: awake" },
	{ "promisc", wl_int, WLC_GET_PROMISC, WLC_SET_PROMISC,
	  "set promiscuous mode ethernet address reception\n"
	  "\t0 - disable\n"
	  "\t1 - enable" },
	{ "monitor", wl_int, WLC_GET_MONITOR, WLC_SET_MONITOR,
	  "set monitor mode\n"
	  "\t0 - disable\n"
	  "\t1 - enable active monitor mode (interface still operates)" },
	{ "frag", wl_print_deprecate, -1, -1,
	  "Deprecated. Use fragthresh." },
	{ "rts", wl_print_deprecate, -1, -1,
	  "Deprecated. Use rtsthresh." },
	{ "cwmin", wl_int, WLC_GET_CWMIN, WLC_SET_CWMIN,
	  "Set the cwmin.  (integer [1, 255])" },
	{ "cwmax", wl_int, WLC_GET_CWMAX, WLC_SET_CWMAX,
	  "Set the cwmax.  (integer [256, 2047])" },
	{ "srl", wl_int, WLC_GET_SRL, WLC_SET_SRL,
	  "Set the short retry limit.  (integer [1, 255])" },
	{ "lrl", wl_int, WLC_GET_LRL, WLC_SET_LRL,
	  "Set the long retry limit.  (integer [1, 255])" },
	{ "rate", wl_rate_mrate, WLC_GET_RATE, -1, 
	  "force a fixed rate:\n"
	  "\tvalid values for 802.11a are (6, 9, 12, 18, 24, 36, 48, 54)\n"
	  "\tvalid values for 802.11b are (1, 2, 5.5, 11)\n"
	  "\tvalid values for 802.11g are (1, 2, 5.5, 6, 9, 11, 12, 18, 24, 36, 48, 54)\n"
	  "\t-1 (default) means automatically determine the best rate" },
	{ "mrate", wl_rate_mrate, -1, -1, /* Deprecated. Use "bg_mrate" or "a_mrate" */
	  "force a fixed multicast rate:\n"
	  "\tvalid values for 802.11a are (6, 9, 12, 18, 24, 36, 48, 54)\n"
	  "\tvalid values for 802.11b are (1, 2, 5.5, 11)\n"
	  "\tvalid values for 802.11g are (1, 2, 5.5, 6, 9, 11, 12, 18, 24, 36, 48, 54)\n"
	  "\t-1 (default) means automatically determine the best rate" },
	{ "a_rate", wl_phy_rate, WLC_GET_VAR, WLC_SET_VAR,
	  "force a fixed rate for the A PHY:\n"
	  "\tvalid values for 802.11a are (6, 9, 12, 18, 24, 36, 48, 54)\n"
	  "\t-1 (default) means automatically determine the best rate" },
	{ "a_mrate", wl_phy_rate, WLC_GET_VAR, WLC_SET_VAR,
	  "force a fixed multicast rate for the A PHY:\n"
	  "\tvalid values for 802.11a are (6, 9, 12, 18, 24, 36, 48, 54)\n"
	  "\t-1 (default) means automatically determine the best rate" },
	{ "bg_rate", wl_phy_rate, WLC_GET_VAR, WLC_SET_VAR,
	  "force a fixed rate for the B/G PHY:\n"
	  "\tvalid values for 802.11b are (1, 2, 5.5, 11)\n"
	  "\tvalid values for 802.11g are (1, 2, 5.5, 6, 9, 11, 12, 18, 24, 36, 48, 54)\n"
	  "\t-1 (default) means automatically determine the best rate" },
	{ "bg_mrate", wl_phy_rate, WLC_GET_VAR, WLC_SET_VAR,
	  "force a fixed multicast rate for the B/G PHY:\n"
	  "\tvalid values for 802.11b are (1, 2, 5.5, 11)\n"
	  "\tvalid values for 802.11g are (1, 2, 5.5, 6, 9, 11, 12, 18, 24, 36, 48, 54)\n"
	  "\t-1 (default) means automatically determine the best rate" },
	{ "infra", wl_int, WLC_GET_INFRA, WLC_SET_INFRA,
	  "Set Infrastructure mode: 0 (IBSS) or 1 (Infra BSS)" },
	{ "ap", wl_int, WLC_GET_AP, WLC_SET_AP,
	  "Set AP mode: 0 (STA) or 1 (AP)" },
	{ "bssid", wl_mac, WLC_GET_BSSID, -1,
	  "Get the BSSID value, error if STA and not associated"},
	{ "channel", wl_channel, WLC_GET_CHANNEL, WLC_SET_CHANNEL,
	  "Set the channel:\n"
	  "\tvalid channels for 802.11b/g (2.4GHz band) are 1 through 14\n"
	  "\tvalid channels for 802.11a  (5 GHz band) are:\n"
	  "\t\t36, 40, 44, 48, 52, 56, 60, 64,\n"
	  "\t\t100, 104, 108, 112, 116,120, 124, 128, 132, 136, 140,\n"
	  "\t\t149, 153, 157, 161,\n"
	  "\t\t184, 188, 192, 196, 200, 204, 208, 212, 216"},
	{ "tssi", wl_tssi, WLC_GET_TSSI, -1,
	  "Get the tssi value from radio" },
	{ "txpwr", wl_txpwr, -1, -1, /* Deprecated. Use "txpwr1" */
	  "Set tx power in milliwatts.  Range [1, 84]." },
	{ "txpwr1", wl_txpwr1, WLC_GET_VAR, WLC_SET_VAR,
	  "Set tx power in in various units. Choose one of (default: dbm): \n"
	  "\t-d dbm units\n"
	  "\t-q quarter dbm units\n"
	  "\t-m milliwatt units\n"
	  "Can be combined with:\n"
	  "\t-o turn on override to disable regulatory and other limitations\n"
	  "Use wl txpwr -1 to restore defaults\n"},
	{ "txpathpwr", wl_int, WLC_GET_TX_PATH_PWR, WLC_SET_TX_PATH_PWR,
	  "Turn the tx path power on or off on 2050 radios" },
	{ "txpwrlimit", wl_get_txpwr_limit, WLC_CURRENT_PWR, -1,
	  "Return current tx power limit" },
	{ "powerindex", wl_int, WLC_GET_PWRIDX, WLC_SET_PWRIDX,
	  "Set the transmit power for A band(0-63).\n"
	  "\t-1 - default value" },
#ifdef DSLCPE
	{ "atten", wl_atten, WLC_GET_ATTEN, WLC_SET_ATTEN,
	"Set the transmit attenuation for B band. Args: bb radio txctl1.\n"
	"\tauto to revert to automatic control\n"
	"\tmanual to supspend automatic control" },
#else
	{ "atten", wl_atten, WLC_GET_ATTEN, WLC_SET_ATTEN,
	  "Set the transmit attenuation for B band. Args: bb radio txctl1.\n"
	  "\tauto to revert to automatic control" },
#endif
	{ "phyreg", wl_reg, WLC_GET_PHYREG, WLC_SET_PHYREG,
	  "Get/Set a phy register:\n"
	  "\toffset [ value ] [ band ]" },
	{ "radioreg", wl_reg, WLC_GET_RADIOREG, WLC_SET_RADIOREG,
	  "Get/Set a radio register:\n"
	  "\toffset [ value ] [ band ]" },
	{ "shmem", wl_reg, WLC_GET_SHMEM, WLC_SET_SHMEM,
	  "Get/Set a shared memory location:\n"
	  "\toffset [ value ] [ band ]" },
	{ "macreg", wl_macreg, WLC_R_REG, WLC_W_REG,
	  "Get/Set any mac registers(include IHR and SB):\n"
	  "\tmacreg offset size[2,4] [value] [ band ]" },
	{ "ucantdiv", wl_int, WLC_GET_UCANTDIV, WLC_SET_UCANTDIV,
	  "Enable/disable ucode antenna diversity (1/0 or on/off)" },
	{ "antdiv", wl_int, WLC_GET_ANTDIV, WLC_SET_ANTDIV,
	  "Set antenna diversity for rx\n"
	  "\t0 - force use of antenna 0\n"
	  "\t1 - force use of antenna 1\n"
	  "\t3 - automatic selection of antenna diversity" },
	{ "txant", wl_int, WLC_GET_TXANT, WLC_SET_TXANT,
	  "Set the transmit antenna\n"
	  "\t0 - force use of antenna 0\n"
	  "\t1 - force use of antenna 1\n"
	  "\t3 - use the RX antenna selection that was in force during\n"
	  "\t    the most recently received good PLCP header" },
	{ "plcphdr", wl_plcphdr, WLC_GET_PLCPHDR, WLC_SET_PLCPHDR,
	  "Set the plcp header.\n"
	  "\t\"long\" or \"auto\" or \"debug\"" },
	{ "phytype", wl_int, WLC_GET_PHYTYPE, -1,
	  "Get phy type" },
	{ "scbdump", wlu_dump, WLC_DUMP_SCB, -1,
	  "print driver scb state to stdout" },
	{ "rateparam", wl_rateparam, -1, WLC_SET_RATE_PARAMS,
	  "set driver rate selection tunables\n"
	  "\targ 1: tunable id\n"
	  "\targ 2: tunable value" },
	{ "wepstatus", wl_wepstatus, -1, -1, /* Deprecated. Use "wsec" */
	  "Set or Get WEP status\n"
	  "\twepstatus [on|off]" },
	{ "primary_key", wl_primary_key, WLC_GET_KEY_PRIMARY, WLC_SET_KEY_PRIMARY,
	  "Set or get index of primary key" },
	{ "addwep", wl_addwep, -1, WLC_SET_KEY,
	  "Set an encryption key.  The key must be 5, 13 or 16 bytes long, or\n"
	  "\t10, 26, 32, or 64 hex digits long.  The encryption algorithm is\n"
	  "\tautomatically selected based on the key size. keytype is accepted\n"
	  "\tonly when key length is 16 bytes/32 hex digits and specifies\n"
	  "\twhether AES-OCB or AES-CCM encryption is used. Default is ccm.\n"
	  "\taddwep <keyindex> <keydata> [ocb | ccm] [notx] [xx:xx:xx:xx:xx:xx]" },
	{ "rmwep", wl_rmwep, -1, WLC_SET_KEY,
	  "Remove the encryption key at the specified key index." },
	{ "keys", wl_keys, WLC_GET_KEY, -1,
	  "Prints a list of the current WEP keys" },
	{ "tsc", wl_tsc, WLC_GET_KEY_SEQ, -1,
	  "Print Tx Sequence Couter for key at specified key index." },
	{ "wsec_test", wl_wsec_test, -1, WLC_SET_WSEC_TEST,
	  "Generate wsec errors\n"
	  "\twsec_test <test_type> <keyindex|xx:xx:xx:xx:xx:xx>\n"
	  "\ttype \'wl wsec_test ?\' for test_types" },
	{ "tkip_countermeasures", wl_int, -1, WLC_TKIP_COUNTERMEASURES,
	  "Enable or disable TKIP countermeasures (TKIP-enabled AP only)\n"
	  "\t0 - disable\n"
	  "\t1 - enable" },
	{ "wsec_restrict", wl_bsscfg_int, WLC_GET_WEP_RESTRICT, WLC_SET_WEP_RESTRICT,
	  "Drop unencrypted packets if WSEC is enabled\n"
	  "\t0 - disable\n"
	  "\t1 - enable" },
	{ "eap", wl_int, WLC_GET_EAP_RESTRICT, WLC_SET_EAP_RESTRICT,
	  "restrict traffic to 802.1X packets until 802.1X authorization succeeds\n"
	  "\t0 - disable\n"
	  "\t1 - enable" },
	{ "authorize", wl_mac, -1, WLC_SCB_AUTHORIZE,
	  "restrict traffic to 802.1X packets until 802.1X authorization succeeds" },
	{ "deauthorize", wl_mac, -1, WLC_SCB_DEAUTHORIZE,
	  "do not restrict traffic to 802.1X packets until 802.1X authorization succeeds" },
	{ "deauthenticate", wl_deauth_rc, -1, WLC_SCB_DEAUTHENTICATE_FOR_REASON,
	  "deauthenticate a STA from the AP with optional reason code (AP ONLY)" },
	{ "wsec", wl_wsec, WLC_GET_WSEC, WLC_SET_WSEC,
	  "wireless security bit vector\n"
	  "\t1 - WEP enabled\n"
	  "\t2 - TKIP enabled\n"
	  "\t4 - AES enabled\n"
	  "\t8 - WSEC in software" },
	{ "auth", wl_bsscfg_int, WLC_GET_AUTH, WLC_SET_AUTH,
	  "set/get 802.11 authentication type. 0 = OpenSystem, 1= SharedKey" },
	{ "wpa_auth", wl_wpa_auth, WLC_GET_WPA_AUTH, WLC_SET_WPA_AUTH,
	  "Bitvector of WPA authorization modes:\n"
	  "\t1	WPA-NONE\n"
	  "\t2	WPA-802.1X/WPA-Professional\n"
	  "\t4	WPA-PSK/WPA-Personal\n"
#ifdef BCMWPA2
	  "\t64	WPA2-802.1X/WPA2-Professional\n"
	  "\t128	WPA2-PSK/WPA2-Personal\n"
#endif /* BCMWPA2 */
	  "\t0	disable WPA"
	},
	{ "wpa_cap", wl_bsscfg_int, WLC_GET_VAR, WLC_SET_VAR,
	  "set/get 802.11i RSN capabilities" },
	{ "set_pmk", wl_set_pmk, -1, WLC_SET_WSEC_PMK,
	  "Set passphrase for PMK in driver-resident supplicant." },
	{ "scan", wl_scan, -1, WLC_SCAN,
	  "Initiate a scan.\n"
	  "\tDefault an active scan across all channels for any SSID.\n"
	  "\tOptional arg: SSID, the SSID to scan.\n"
	  "\tOptions:\n"
	  "\t-s S, --ssid=S\t\tSSID to scan\n"
	  "\t-t ST, --scan_type=ST\t[active|passive] scan type\n"
	  "\t--bss_type=BT\t\t[bss/infra|ibss/adhoc] bss type to scan\n"
	  "\t-b MAC, --bssid=MAC\tparticular BSSID MAC address to scan, xx:xx:xx:xx:xx:xx\n"
	  "\t-n N, --nprobes=N\tnumber of probes per scanned channel\n"
	  "\t-a N, --active=N\tdwell time per channel for active scanning\n"
	  "\t-p N, --passive=N\tdwell time per channel for passive scanning\n"
	  "\t-h N, --home=N\t\tdwell time for the home channel between channel scans\n"
	  "\t-c L, --channels=L\tcomma or space separated list of channels to scan"},
	{ "passive", wl_int, WLC_GET_PASSIVE_SCAN, WLC_SET_PASSIVE_SCAN,
	  "Puts scan engine into passive mode" },
	{ "regulatory", wl_int, WLC_GET_REGULATORY, WLC_SET_REGULATORY,
	  "Get/Set regulatory domain mode (802.11d). Driver must be down." },
	{ "spect", wl_spect, WLC_GET_SPECT_MANAGMENT, WLC_SET_SPECT_MANAGMENT,
	  "Get/Set 802.11h Spectrum Management mode.\n" 
	  "\t0 - Off\n"
	  "\t1 - Loose interpretation of spec - may join non-11h APs\n"
	  "\t2 - Strict interpretation of spec - may not join non-11h APs"
	  "\t3 - Disable 11H and enable 11D"
	  },
	{ "scanresults", wl_dump_networks, WLC_SCAN_RESULTS, -1,
	  "Return results from last scan." },
	{ "assoc",  wl_status, -1, -1,
	  "Print information about current network association.\n"
	  "\t(also known as \"status\")" },
	{ "status", wl_status, -1, -1,
	  "Print information about current network association.\n"
	  "\t(also known as \"assoc\")" },
	{ "disassoc", wl_void, -1, WLC_DISASSOC,
	  "Disassociate from the current BSS/IBSS." },
	{ "chanlist", wl_print_deprecate, WLC_GET_VALID_CHANNELS, -1,
	  "Deprecated. Use channels." },
	{ "channels", wl_dump_chanlist, WLC_GET_VALID_CHANNELS, -1,
	  "Return valid channels for the current settings." },
	{ "channels_in_country", wl_channels_in_country, WLC_GET_CHANNELS_IN_COUNTRY, -1,
	  "Return valid channels for the country specified.\n" 
	  "\tArg 1 is the country abbreviation\n"
	  "\tArg 2 is the band(a or b)"},
	{ "curpower", wl_get_current_power, WLC_CURRENT_PWR, -1, 
	  "Return current tx power settings.\n" 
	  "\t-q (quiet): estimated power only." },
#ifdef DSLCPE
	{ "txinstpwr", wl_get_instant_power, WLC_GET_VAR, -1,
	"Return tx power based on instant TSSI "},
#endif
	{ "scansuppress", wl_int, WLC_GET_SCANSUPPRESS, WLC_SET_SCANSUPPRESS,
	  "Suppress all scans for testing.\n"
	  "\t0 - allow scans\n"
	  "\t1 - suppress scans" },
	{ "evm", wl_evm, -1, WLC_EVM,
	  "Start an EVM test on the given channel, or stop EVM test.\n"
	  "\tArg 1 is channel number 1-14, or \"off\" or 0 to stop the test.\n"
	  "\tArg 2 is optional rate (1, 2, 5.5 or 11)"},
	{ "rateset", wl_rateset, WLC_GET_RATESET, WLC_SET_RATESET,
	  "Returns or sets the supported and basic rateset, (b) indicates basic\n"
	  "\tWith no args, returns the rateset. Args are\n"
	  "\trateset \"default\" | \"all\" | <arbitrary rateset>\n"
	  "\t\tdefault - driver defaults\n"
	  "\t\tall - all rates are basic rates\n"
	  "\t\tarbitrary rateset - list of rates\n"
	  "\tList of rates are in Mbps and each rate is optionally followed\n"
	  "\tby \"(b)\" or \"b\" for a Basic rate. Example: 1(b) 2b 5.5 11\n"
	  "\tAt least one rate must be Basic for a legal rateset."},
	{ "roam_trigger", wl_band_elm, WLC_GET_ROAM_TRIGGER, WLC_SET_ROAM_TRIGGER,
	  "Set the roam trigger RSSI threshold: roam_trigger [integer [, a/b]] " },
	{ "roam_delta",	wl_band_elm, WLC_GET_ROAM_DELTA, WLC_SET_ROAM_DELTA,
	  "Set the roam candidate qualification delta. roam_delta [integer [, a/b]]" },
	{ "roam_scan_period", wl_int, WLC_GET_ROAM_SCAN_PERIOD, WLC_SET_ROAM_SCAN_PERIOD,
	  "Set the roam candidate qualification delta.  (integer)" },
	{ "suprates", wl_sup_rateset, WLC_GET_SUP_RATESET_OVERRIDE, WLC_SET_SUP_RATESET_OVERRIDE,
	  "Returns or sets the 11g override for the supported rateset\n"
	  "\tWith no args, returns the rateset. Args are a list of rates,\n"
	  "\tor 0 or -1 to specify an empty rateset to clear the override.\n"
	  "\tList of rates are in Mbps, example: 1 2 5.5 11"},
	{ "scan_channel_time", wl_int, WLC_GET_SCAN_CHANNEL_TIME, WLC_SET_SCAN_CHANNEL_TIME,
	  "Get/Set scan channel time"},
	{ "scan_unassoc_time", wl_int, WLC_GET_SCAN_UNASSOC_TIME, WLC_SET_SCAN_UNASSOC_TIME,
	  "Get/Set unassociated scan channel dwell time"},
	{ "scan_home_time", wl_int, WLC_GET_SCAN_HOME_TIME, WLC_SET_SCAN_HOME_TIME,
	  "Get/Set scan home channel dwell time"},
	{ "scan_passive_time", wl_int, WLC_GET_SCAN_PASSIVE_TIME, WLC_SET_SCAN_PASSIVE_TIME,
	  "Get/Set passive scan channel dwell time"},
	{ "scan_nprobes", wl_int, WLC_GET_SCAN_NPROBES, WLC_SET_SCAN_NPROBES,
	  "Get/Set scan parameter for number of probes to use per channel scanned"},
	{ "prb_resp_timeout", wl_int, WLC_GET_PRB_RESP_TIMEOUT, WLC_SET_PRB_RESP_TIMEOUT,
	  "Get/Set probe response timeout"},
	{ "channel_qa", wl_int, WLC_GET_CHANNEL_QA, -1,
	  "Get last channel quality measurment"},
	{ "channel_qa_start", wl_void, -1, WLC_START_CHANNEL_QA,
	  "Start a channel quality measurment"},
	{ "country", wl_country, WLC_GET_COUNTRY, WLC_SET_COUNTRY,
	  "Select Country code for use with 802.11d\n"
	  "\tUse either long name or abbreviation from ISO 3166.\n"
	  "\tUse 'wl country list [band(a or b)]' for the list of supported countries"},
	{ "locale", wl_locale, WLC_GET_LOCALE, WLC_SET_LOCALE,
	  "OBSOLETE: use \"wl country\"\n"
	  "\tSelect the country:\n"
	  "\tWorldwide\n"
	  "\tThailand\n"
	  "\tIsrael\n"
	  "\tJordan\n"
	  "\tChina\n"
	  "\tJapan\n"
	  "\tUSA/Canada/ANZ\n"
	  "\tEurope\n"
	  "\tUSAlow\n"
	  "\tJapanHigh\n"
	  "\tAll" },
	{ "join", wl_join, -1, -1,
	  "Join a specified network SSID.\n"
	  "\tJoin syntax is: join <ssid> [key xxxxx] [imode bss|ibss] [amode open|shared|wpa|wpapsk|wpa2|wpa2psk|wpanone]"},
	{ "ssid", wl_ssid, WLC_GET_SSID, WLC_SET_SSID,
	  "Set or get a configuration's SSID.\n"
	  "\twl ssid [-C num]|[--cfg=num] [<ssid>]\n"
	  "\tIf the configuration index 'num' is not given, configuraion #0 is assumed and\n"
	  "\tsetting will initiate an assoication attempt if in infrastructure mode,\n"
	  "\tor join/creation of an IBSS if in IBSS mode,\n"
	  "\tor creation of a BSS if in AP mode.\n"},
	{ "mac", wl_maclist, WLC_GET_MACLIST, WLC_SET_MACLIST,
	  "Set or get the list of source MAC address matches.\n"
	  "\twl mac xx:xx:xx:xx:xx:xx [xx:xx:xx:xx:xx:xx ...]\n"
	  "\tTo Clear the list: wl mac none" },
	{ "macmode", wl_int, WLC_GET_MACMODE, WLC_SET_MACMODE,
	  "Set the mode of the MAC list.\n"
	  "\t0 - Disable MAC address matching.\n"
	  "\t1 - Deny association to stations on the MAC list.\n"
	  "\t2 - Allow association to stations on the MAC list."},
	{ "wds", wl_maclist, WLC_GET_WDSLIST, WLC_SET_WDSLIST,
	  "Set or get the list of WDS member MAC addresses.\n"
	  "\tSet using a space separated list of MAC addresses.\n"
	  "\t wl wds xx:xx:xx:xx:xx:xx [xx:xx:xx:xx:xx:xx ...]" },
	{ "lazywds", wl_int, WLC_GET_LAZYWDS, WLC_SET_LAZYWDS,
	  "Set or get \"lazy\" WDS mode (dynamically grant WDS membership to anyone)."},
	{ "noise", wl_int, WLC_GET_PHY_NOISE, -1,
	  "Get noise (moving average) right after tx in dBm" },
	{ "fqacurcy", wl_int, -1, WLC_FREQ_ACCURACY,
	  "Manufacturing test: set frequency accuracy mode.\n"
	  "\tfreqacuracy syntax is: fqacurcy <channel>\n"
	  "\tArg is channel number 1-14, or 0 to stop the test." },
	{ "crsuprs", wl_int, -1, WLC_CARRIER_SUPPRESS,
	  "Manufacturing test: set carrier suppression mode.\n"
	  "\tcarriersuprs syntax is: crsuprs <channel>\n"
	  "\tArg is channel number 1-14, or 0 to stop the test." },
	{ "longtrain", wl_int, -1, WLC_LONGTRAIN,
	  "Manufacturing test: set longtraining mode.\n"
	  "\tlongtrain syntax is: longtrain <channel>\n"
	  "\tArg is A band channel number or 0 to stop the test." },
	{ "band", wl_band, WLC_GET_BAND, WLC_SET_BAND,
	  "Returns or sets the current band\n"
	  "\tauto - auto switch between available bands (default)\n"
	  "\ta - force use of 802.11a band\n"
	  "\tb - force use of 802.11b band" },
	{ "bands", wl_bandlist, WLC_GET_BANDLIST, -1,
	  "Return the list of available 802.11 bands" },
	{ "phylist", wl_phylist, WLC_GET_PHYLIST, -1,
	  "Return the list of available phytypes" },
	{ "shortslot", wl_int, WLC_GET_SHORTSLOT, -1,
	  "Get current 11g Short Slot Timing mode. (0=long, 1=short)" },
	{ "shortslot_override", wl_int, WLC_GET_SHORTSLOT_OVERRIDE, WLC_SET_SHORTSLOT_OVERRIDE,
	  "Get/Set 11g Short Slot Timing mode override. (-1=auto, 0=long, 1=short)" },
	{ "shortslot_restrict", wl_int, WLC_GET_SHORTSLOT_RESTRICT, WLC_SET_SHORTSLOT_RESTRICT,
	  "Get/Set AP Restriction on associations for 11g Short Slot Timing capable STAs.\n"
	  "\t0 - Do not restrict association based on ShortSlot capability\n"
	  "\t1 - Restrict association to STAs with ShortSlot capability" },
	{ "ignore_bcns", wl_int, WLC_GET_IGNORE_BCNS, WLC_SET_IGNORE_BCNS,
	  "AP only (G mode): Check for beacons without NONERP element (0=Examine beacons, 1=Ignore beacons)" },
	{ "pktcnt", wl_get_pktcnt, WLC_GET_PKTCNTS, -1,
	  "Get the summary of good and bad packets." },
	{ "upgrade", wl_upgrade, -1, WLC_UPGRADE,
	  "Upgrade the firmware on an embedded device" },
#endif /* WME_ONLY */
	{ "gmode", wl_gmode, WLC_GET_GMODE, WLC_SET_GMODE,
	  "Set the 54g Mode (LegacyB|Auto||GOnly|BDeferred|Performance|LRS)" },
#ifndef WME_ONLY
	{ "gmode_protection", wl_int, WLC_GET_GMODE_PROTECTION, -1,
	  "Get G protection mode. (0=disabled, 1=enabled)" },
	{ "gmode_protection_control", wl_int, WLC_GET_GMODE_PROTECTION_CONTROL, WLC_SET_GMODE_PROTECTION_CONTROL,
	  "Get/Set 11g protection mode control alg. (0=always off, 1=monitor local association, 2=monitor overlapping BSS)" },
	{ "gmode_protection_cts", wl_int, WLC_GET_GMODE_PROTECTION_CTS, WLC_SET_GMODE_PROTECTION_CTS,
	  "Get/Set 11g protection type to CTS (0=disable, 1=enable)" },
	{ "gmode_protection_override", wl_int, WLC_GET_GMODE_PROTECTION_OVERRIDE, WLC_SET_GMODE_PROTECTION_OVERRIDE,
	  "Get/Set 11g protection mode override. (-1=auto, 0=disable, 1=enable)" },
	{ "legacy_erp", wl_int, WLC_GET_LEGACY_ERP, WLC_SET_LEGACY_ERP,
	  "Get/Set 11g legacy ERP inclusion (0=disable, 1=enable)" },
	{ "scb_timeout", wl_int, WLC_GET_SCB_TIMEOUT, WLC_SET_SCB_TIMEOUT,
	  "AP only: inactivity timeout value for authenticated stas" },
	{ "assoclist", wl_maclist, WLC_GET_ASSOCLIST, -1,
	  "AP only: Get the list of associated MAC addresses."},
	{ "rssi", wl_rssi, WLC_GET_RSSI, -1,
	  "Get the current RSSI val, for an AP you must specify the mac addr of the STA" },
	{ "isup", wl_int, WLC_GET_UP, -1,
	  "Get driver operational state (0=down, 1=up)"},
	{ "fasttimer", wl_print_deprecate, -1, -1,
	  "Deprecated. Use fast_timer."},
	{ "slowtimer", wl_print_deprecate, -1, -1,
	  "Deprecated. Use slow_timer."},
	{ "glacialtimer", wl_print_deprecate, -1, -1,
	  "Deprecated. Use glacial_timer."},
	{ "radar", wl_int, WLC_GET_RADAR, WLC_SET_RADAR,
	  "Enable/Disable radar"},
	{ "radarargs", wl_radar_args, WLC_GET_VAR, WLC_SET_VAR, 
	  "Get/Set Radar parameters in \n"
	  "\torder as npulses, ncontig, min_pw , max_pw, thresh0, thresh1"},
	{ "dfs_status", wl_dfs_status, WLC_GET_VAR, -1, 
	  "Get dfs status\n"},
	{ "interference", wl_interfere, WLC_GET_INTERFERENCE_MODE, WLC_SET_INTERFERENCE_MODE,
	  "Get/Set interference mitigation mode. Choices are:\n"
	  "\t0 = none\n"
	  "\t1 = non wlan\n"
	  "\t2 = wlan manual\n"
	  "\t3 = wlan automatic"},
#ifdef DSLCPE
	{ "aciargs", wl_aci_args, WLC_GET_ACI_ARGS, WLC_SET_ACI_ARGS,
	  "Get/Set various aci tuning parameters.  Choices are:\n"
	  "\tenter:\tCRS glitch trigger level to start detecting ACI\n"
	  "\texit:\tCRS glitch trigger level to exit ACI mode\n"
	  "\tglitch\tSeconds interval between ACI scans when glitchcount is continuously high\n"
	  "\tspin:\tNum microsecs to delay between rssi samples\n"
	  "\tUsage: wl aciargs [enter x][exit x][spin x][glitch x]"},
#endif
	{ "frameburst", wl_int, WLC_GET_FAKEFRAG, WLC_SET_FAKEFRAG,
	  "Disable/Enable frameburst mode" },
	{ "pwr_percent", wl_int, WLC_GET_PWROUT_PERCENTAGE, WLC_SET_PWROUT_PERCENTAGE,
	  "Get/Set power output percentage"},
	{ "wet", wl_int, WLC_GET_WET, WLC_SET_WET,
	  "Get/Set wireless ethernet bridging mode"},
	{ "bi", wl_int, WLC_GET_BCNPRD, WLC_SET_BCNPRD,
	  "Get/Set the beacon period (bi=beacon interval)"},
	{ "dtim", wl_int, WLC_GET_DTIMPRD, WLC_SET_DTIMPRD,
	  "Get/Set DTIM"},
	{ "wds_remote_mac", wl_mac, WLC_WDS_GET_REMOTE_HWADDR, -1,
	  "Get WDS link remote endpoint's MAC address"},
	{ "wds_wpa_role_old", wl_wds_wpa_role_old, WLC_WDS_GET_WPA_SUP, -1,
	  "Get WDS link local endpoint's WPA role (old)"},
	{ "wds_wpa_role", wl_wds_wpa_role, WLC_GET_VAR, WLC_SET_VAR,
	  "Get/Set WDS link local endpoint's WPA role"},
	{ "authe_sta_list", wl_maclist_1, WLC_GET_VAR, -1,
	  "Get authenticated sta mac address list"},
	{ "autho_sta_list", wl_maclist_1, WLC_GET_VAR, -1,
	  "Get authorized sta mac address list"},
	{ "measure_req", wl_measure_req, -1, WLC_MEASURE_REQUEST,
	  "Send an 802.11h measurement request.\n"
	  "\tUsage: wl measure_req <type> <target MAC addr>\n"
	  "\tMeasurement types are: TPC, Basic, CCA, RPI\n"
	  "\tTarget MAC addr format is xx:xx:xx:xx:xx:xx"},
	{ "quiet", wl_send_quiet, -1, WLC_SEND_QUIET,
	   "Send an 802.11h quiet command.\n"
	   "\tUsage: wl quiet <TBTTs until start>, <duration (in TUs)>, <offset (in TUs)>"},
	{ "csa", wl_send_csa, -1, WLC_CHANNEL_SWITCH,
	   "Send an 802.11h channel switch anouncement\n"
	   "\tUsage wl csa <mode> <when (in TBTTs)> <channel>"},
	{ "constraint", wl_int, -1, WLC_SEND_PWR_CONSTRAINT,
	   "Send an 802.11h Power Constraint IE\n"
	   "\tUsage: wl constraint 1-255 db"},
	{ "rm_req", wl_rm_request, -1, WLC_SET_VAR,
	  "Request a radio measurement of type basic, cca, or rpi\n"
	  "\tspecify a series of measurement types each followed by options.\n"
	  "\texample: wl rm_req cca -c 1 -d 50 cca -c 6 cca -c 11\n"
	  "\tOptions:\n"
	  "\t-t n  numeric token id for measurement set or measurement\n"
	  "\t-c n  channel\n"
	  "\t-d n  duration in TUs (1024 us)\n"
	  "\t-p    parallel flag, measurement starts at the same time as previous\n"
	  "\n"
	  "\tEach measurement specified uses the same channel and duration as the\n"
	  "\tprevious unless a new channel or duration is specified."},
	{ "rm_rep", wl_rm_report, WLC_GET_VAR, -1,
	  "Get current radio measurement report"},
	{ "join_pref", wl_join_pref, WLC_GET_VAR, WLC_SET_VAR,
	  "Set/Get join target preferences.\n"},
	{ "assoc_pref", wl_assoc_pref, WLC_GET_ASSOC_PREFER, WLC_SET_ASSOC_PREFER,
	  "Set/Get association preference.\n"
	  "Usage: wl assoc_pref [auto|a|b|g]\n"},
#endif /* WME_ONLY */
	{ "wme", wl_varint, WLC_GET_VAR, WLC_SET_VAR,
	  "Set WME (Wireless Multimedia Extensions) mode (0=off, 1=on, -1=auto)" },
	{ "wme_ac", wl_wme_ac_req, WLC_GET_VAR, WLC_SET_VAR,
	  "wl wme_ac sta/ap [be, bk, vi, vo] [ecwmax, ecwmin, txop, aifsn, acm] value"},
	{ "wme_apsd", wl_varint, WLC_GET_VAR, WLC_SET_VAR,
	  "Set APSD (Automatic Power Save Delivery) mode on AP (0=off, 1=on)" },
	{ "wme_apsd_sta", wl_wme_apsd_sta, WLC_GET_VAR, WLC_SET_VAR,
	  "Set APSD parameters on STA. Driver must be down.\n"
	  "Usage: wl wme_apsd_sta <max_sp_len> <be> <bk> <vi> <vo>\n"
	  "   <max_sp_len>: number of frames per USP: 0 (all), 2, 4, or 6\n"
	  "   <xx>: value 0 to disable, 1 to enable U-APSD per AC\n" },
	{ "wme_dp", wl_wme_dp, WLC_GET_VAR, WLC_SET_VAR,
	  "Set AC queue discard policy.\n"
	  "Usage: wl wme_dp <be> <bk> <vi> <vo>\n"
	  "   <xx>: value 0 for newest-first, 1 for oldest-first\n" },
#ifndef WME_ONLY
	{ "wme_counters", wl_wme_counters, WLC_GET_VAR, -1,
	"print WMM stats" },
	{ "reinit", wl_void, -1, WLC_INIT,
	  "Reinitialize device"},
	{ "sta_info", wl_sta_info, WLC_GET_VAR, -1,
	  "wl sta_info <xx:xx:xx:xx:xx:xx>"},
	{ "cap", wl_var_getandprintstr, WLC_GET_VAR, -1, "driver capabilities"},
	{ "malloc_dump", wl_var_getandprintstr, WLC_GET_VAR, -1, "debug malloc info"},
	{ "chan_info", wl_chan_info, WLC_GET_VAR, -1, "channel info"},
	{ "add_ie", wl_add_ie, -1, WLC_SET_VAR,
	  "Add a vendor proprietary IE to 802.11 management packets\n"
	  "Usage: wl add_ie <pktflag> length OUI hexdata\n"
	  "<pktflag>: Bit 0 - Beacons\n"
	  "           Bit 1 - Probe Rsp\n"
	  "           Bit 2 - Assoc/Reassoc Rsp\n"
	  "           Bit 3 - Auth Rsp\n"
	  "Example: wl add_ie 3 10 00:90:4C 0101050c121a03\n"
	  "         to add this IE to beacons and probe responses\n" },
	{ "del_ie", wl_del_ie, -1, WLC_SET_VAR,
	  "Delete a vendor proprietary IE from 802.11 management packets\n"
	  "Usage: wl del_ie <pktflag> length OUI hexdata\n"
	  "<pktflag>: Bit 0 - Beacons\n"
	  "           Bit 1 - Probe Rsp\n"
	  "           Bit 2 - Assoc/Reassoc Rsp\n"
	  "           Bit 3 - Auth Rsp\n"
	  "Example: wl del_ie 3 10 00:90:4C 0101050c121a03\n" },
	{ "list_ie", wl_list_ie, WLC_GET_VAR, -1,
	  "Dump the list of vendor proprietary IEs\n" },
	{ "rand", wl_rand, WLC_GET_VAR, -1,
	  "Get a 2-byte Random Number from the MAC's PRNG\n"
	  "Usage: wl rand\n"},
	{ "nvotpw", wl_otpw, -1, WLC_NVOTPW,
	  "Write nvram to on-chip otp\n"
	  "Usage: wl nvotpw file\n"},
	{ "bcmerrorstr", wl_var_getandprintstr, WLC_GET_VAR, -1, "errorstring"},
	{ "freqtrack", wl_varint, WLC_GET_VAR, WLC_SET_VAR,
	  "Set Frequency Tracking Mode (0=Auto, 1=On, 2=OFF)\n"},
	{ "eventing", wl_bitvec128, WLC_GET_VAR, WLC_SET_VAR,
	  "set/get 128-bit hex filter bitmask for MAC event reporting up to application layer\n"},
	{ "event_msgs", wl_bitvec128, WLC_GET_VAR, WLC_SET_VAR,
	  "set/get 128-bit hex filter bitmask for MAC event reporting via packet indications\n"},
	{ "counters", wl_counters, WLC_GET_VAR, -1,
	  "Return driver counter values\n" },
	{ "assoc_info", wl_assoc_info, WLC_GET_VAR, -1,
	  "Returns the assoc req and resp information [STA only]\n" },
	{ "autochannel", wl_auto_channel_sel, WLC_GET_CHANNEL_SEL, WLC_START_CHANNEL_SEL,
	  "auto channel selection: \n"
	  "\t1 to issue a channel scanning;\n"
	  "\t2 to set channel based on the channel scanning result;\n"
	  "\twithout argument to only show the channel selected; \n"
	  "\tssid must set to null before this process, RF must be up"},	  	  
#ifdef DSLCPE	  
	{ "csscantimer", wl_int, WLC_GET_CS_SCAN_TIMER, WLC_SET_CS_SCAN_TIMER,
	  "auto channel scan timer in minutes (0 to disable)" },
#endif		  	  	  
	{ "closed", wl_int, WLC_GET_CLOSED, WLC_SET_CLOSED,
	  "hides the network from active scans, 0 or 1.\n"
	  "\t0 is open, 1 is hide" },
#ifdef BCMWPA2
	{ "pmkid_info", wl_pmkid_info, WLC_GET_VAR, WLC_SET_VAR,
	  "Returns the pmkid table\n" },
#endif
	{ "abminrate", wl_phy_rate, WLC_GET_VAR, WLC_SET_VAR,
	  "get/set afterburner minimum rate threshold\n" },
	{ "bss", wl_bsscfg_enable, WLC_GET_VAR, WLC_SET_VAR,
	  "set/get BSS enabled status: up/down\n"},
	{ "closednet", wl_bsscfg_int, WLC_GET_VAR, WLC_SET_VAR,
	  "set/get BSS closed network attribute\n"},
	{ "diag", wl_diag, WLC_GET_VAR, -1,
	  "diag testindex(1-interrupt, 2-loopback, 3-memory, 4-led); precede by 'wl down' and follow by 'wl up'" },
	{ "reset_d11cnts", wl_var_void, -1, WLC_SET_VAR,
	  "reset 802.11 MIB counters" },
#endif /* WME_ONLY */

#if defined(DSLCPE) && defined(RADIUS_RESTRICTION)
	{ "8021x", wl_bsscfg_int, WLC_GET_VAR, WLC_SET_VAR,
	  "Set or get 802.1x in use.\n"
	  "\t0 - disable\n"
	  "\t1 - enable\n" },
	{ "idle_timeout", wl_varint, WLC_GET_VAR, WLC_SET_VAR,
	  "Set or get the session idle timeout in seconds.\n" },
	{ "auth_blockout_time", wl_varint, WLC_GET_VAR, WLC_SET_VAR,
	  "Set or get the barring time (seconds) for log in retry is accepted.\n" },
	{ "auth_blockout_list", wl_maclist_1, WLC_GET_VAR, -1,
	  "Get the list of authentication block out MAC addresses."},
#endif
	{ NULL, NULL, 0, 0, NULL }
};

cmd_t wl_varcmd = { "var", wl_varint, -1, -1, "unrecognized name, type -h for help" } ;
char *av0 = "";

/* initialize stuff needed before processing the command */
void
wl_cmd_init(void)
{
	int_fmt = INT_FMT_DEC;
}

/* parse/validate the command line arguments */
/*
 * pargv is updated upon return if the first argument is an option.
 * It remains intact otherwise.
 */
int
wl_option(char ***pargv, char **pifname, int *phelp)
{
	char *ifname = NULL;
	int help = FALSE;
	int status = CMD_OPT;
	char **argv = *pargv;

	while (*argv) {
		/* select different adapter */
		if (!strcmp(*argv, "-a") || !strcmp(*argv, "-i")) {
			char *opt = *argv++;
			ifname = *argv;
			if (!ifname) {
				fprintf(stderr, "error: expected interface name after option %s\n", opt);
				status = CMD_ERR;
				break;
			}
		}

		/* integer output format */
		else if (!strcmp(*argv, "-d"))
			int_fmt = INT_FMT_DEC;
		else if (!strcmp(*argv, "-u"))
			int_fmt = INT_FMT_UINT;
		else if (!strcmp(*argv, "-x"))
			int_fmt = INT_FMT_HEX;

		/* command usage */
		else if (!strcmp(*argv, "-h"))
			help = TRUE;

		/* start of non wl options */
		else {
			status = CMD_WL;
			break;
		}

		/* consume the argument */
		argv ++;
		break;
	}
	
	*phelp = help;
	*pifname = ifname;
	*pargv = argv;
	
	return status;
}

void
cmd_usage(cmd_t *cmd)
{
	if (strlen(cmd->name) >= 8)
		fprintf(stderr, "%s\n\t%s\n\n", cmd->name, cmd->help);
	else
		fprintf(stderr, "%s\t%s\n\n", cmd->name, cmd->help);
}	

static int
wl_print_deprecate(void *wl, cmd_t *cmd, char **argv)
{
	cmd_usage(cmd); /* warning string is in command table */
	return 0;
}

/* Dump out short list of commands */
static int
wl_list(void *wl, cmd_t *garb, char **argv)
{
	cmd_t *cmd;
	int nrows, i, len;
	char *buf;
	int letter, col, row, pad;

	for (cmd = wl_cmds, nrows = 0; cmd->name; cmd++, nrows++);
	nrows /= 4;
	nrows++;

	len = nrows * 80 + 2;
	buf = malloc(len);
	if (buf == NULL) {
		fprintf(stderr, "Failed to allocate buffer of %d bytes\n",len );
		return -1;
	}
	for (i=0; i < len; i++)
		*(buf+i) = 0;

	row = col = 0;
	for (letter = 'a'; letter < 'z'; letter++){
		for (cmd = wl_cmds; cmd->name; cmd++){
			if (cmd->name[0] == letter || cmd->name[0] == letter - 0x20){
				strcat(buf+row*80, cmd->name);
				pad = 18 * (col + 1) - strlen(buf+row*80);
				if (pad < 1)
					pad = 1;
				for (; pad; pad--)
					strcat(buf+row*80, " ");
				row++;
				if (row == nrows){
					col++; row = 0;
				}
			}
		}
	}
	for (row = 0; row < nrows; row++)
		printf("%s\n", buf+row*80);

	printf("\n");
	return(0);
}

void
wl_cmds_usage(cmd_t *port_cmds)
{
	cmd_t *port_cmd;
	cmd_t *cmd;

	/* print usage of port commands */
	for (port_cmd = port_cmds; port_cmd && port_cmd->name; port_cmd++)
		cmd_usage(port_cmd);

	/* print usage of common commands without port counterparts */
	for (cmd = wl_cmds; cmd->name; cmd++) {
		/* search if port counterpart exists */
		for (port_cmd = port_cmds; port_cmd && port_cmd->name; port_cmd++)
			if (!strcmp(port_cmd->name, cmd->name))
				break;
		if (!port_cmd || !port_cmd->name)
			cmd_usage(cmd);
	}
}

void
wl_usage(cmd_t *port_cmds)
{
	fprintf(stderr, "Usage: %s [-a|i <adapter>] [-h] [-d|u|x] <command> [arguments]\n", av0);
	
	fprintf(stderr, "\n");
	fprintf(stderr, "  -h		this message\n");
	fprintf(stderr, "  -a, -i	adapter name or number\n");
	fprintf(stderr, "  -d		signed integer\n");
	fprintf(stderr, "  -u		unsigned integer\n");
	fprintf(stderr, "  -x		hexdecimal\n");
	fprintf(stderr, "\n");

	wl_cmds_usage(port_cmds);
}

int
wl_check(void *wl)
{
	int ret;
	int val;

	if ((ret = wl_get(wl, WLC_GET_MAGIC, &val, sizeof(int)) < 0))
		return ret;
	if (val != WLC_IOCTL_MAGIC)
		return -1;
	if ((ret = wl_get(wl, WLC_GET_VERSION, &val, sizeof(int)) < 0))
		return ret;
	if (val > WLC_IOCTL_VERSION) {
		fprintf(stderr, "Version mismatch, please upgrade\n");
		return -1;
	}
	return 0;
}

void
wl_printint(int val)
{
	switch (int_fmt) {
	case INT_FMT_UINT:
		printf("%u\n", val);
		break;
	case INT_FMT_HEX:
		printf("0x%x\n", val);
		break;
	case INT_FMT_DEC:
	default:
		printf("%d\n", val);
		break;
	}
}


/* Common routine to check for an option arg specifying the configuration index.
 * Takes the syntax -C num, --cfg=num, --config=num, or --configuration=num
 * Returns -1 if there is a command line parsing error.
 * Returns 0 if no error, and sets *consumed to the number of argv strings
 * used. Sets *bsscfg_idx to the index to use. Will set *bsscfg_idx to zero if there
 * was no config arg.
 */
static int
wl_cfg_option(char **argv, char *fn_name, int *bsscfg_idx, int *consumed)
{
	miniopt_t mo;
	int opt_err;
	
	*bsscfg_idx = 0;
	*consumed = 0;

	miniopt_init(&mo, fn_name, NULL);

	/* process the first option */
	opt_err = miniopt(&mo, argv);

	/* check for no args or end of options */
	if (opt_err == -1)
		return 0;
	
	/* check for no options, just a positional arg encountered */
	if (mo.positional)
		return 0;

	/* check for error parsing options */
	if (opt_err == 1)
		return -1;
	
	/* check for -C, --cfg=X, --config=X, --configuration=X */
	if (mo.opt == 'C' ||
	    !strcmp(mo.key, "cfg") ||
	    !strcmp(mo.key, "config") ||
	    !strcmp(mo.key, "configuration")) {
		if (!mo.good_int) {
			fprintf(stderr, "%s: could not parse \"%s\" as an integer for the configuartion index\n",
				fn_name, mo.valstr);
			return -1;
		}
		*bsscfg_idx = mo.val;
		*consumed = mo.consumed;
	}
	
	return 0;
}

static int
wl_void(void *wl, cmd_t *cmd, char **argv)
{
	if (cmd->set < 0)
		return -1;
	return wl_set(wl, cmd->set, NULL, 0);
}	

int
wl_int(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	int val;
	char *endptr;

	if (!*++argv) {
		if (cmd->get == -1)
			return -1;			
		if ((ret = wl_get(wl, cmd->get, &val, sizeof(int))) < 0)
			return ret;

		wl_printint(val);
	} else {
		if (cmd->set == -1)
			return -1;
		if (!stricmp(*argv, "on"))
			val = 1;
		else if (!stricmp(*argv, "off"))
			val = 0;
		else {
			val = strtol(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}
		}

		ret = wl_set(wl, cmd->set, &val, sizeof(int));
	}

	return ret;
}

static int
wl_bsscfg_int(void *wl, cmd_t *cmd, char **argv)
{
	char *endptr;
	char *val_name;
	int bsscfg_idx = 0;
	int val;
	int consumed;
	int ret;

	val_name = *argv++;
	
	/* parse a bsscfg_idx option if present */
	if ((ret = wl_cfg_option(argv, val_name, &bsscfg_idx, &consumed)) != 0)
		return ret;

	/* handle a bsscfg int with a legacy ioctl */
	if (consumed == 0 && cmd->set != WLC_SET_VAR) {
		/* back up to the orig command and run as an ioctl int */
		argv--;
		return wl_int(wl, cmd, argv);
	}

	argv += consumed;

	if (!*argv) {
		/* This is a GET */
		if (cmd->get == -1)
			return -1;

		if (consumed == 0)
			ret = wl_iovar_getint(wl, val_name, &val);
		else
			ret = wl_bssiovar_getint(wl, val_name, bsscfg_idx, &val);
		
		if (ret < 0)
			return ret;

#ifdef DSLCPE
#ifdef IL_BIGENDIAN
                if(!strcmp(val_name,"wpa_cap"))
                        val = bcmswap32(val);
#endif  
#endif

		wl_printint(val);
	} else {
		/* This is a SET */
		if (cmd->set == -1)
			return -1;

		if (!stricmp(*argv, "on"))
			val = 1;
		else if (!stricmp(*argv, "off"))
			val = 0;
		else {
			val = strtol(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}
		}
		
#ifdef DSLCPE
#ifdef IL_BIGENDIAN
                if(!strcmp(val_name,"wpa_cap"))
                        val = bcmswap32(val);
#endif
#endif 

		if (consumed == 0)
			ret = wl_iovar_setint(wl, val_name, val);
		else
			ret = wl_bssiovar_setint(wl, val_name, bsscfg_idx, val);
	}

	return ret;
}

static int
wl_bsscfg_enable(void *wl, cmd_t *cmd, char **argv)
{
	char *endptr;
	char *val_name = "bss";
	int bsscfg_idx = 0;
	int val;
	int consumed;
	int ret;

	/* skip the command name */
	argv++;
	
	/* parse a bsscfg_idx option if present */
	if ((ret = wl_cfg_option(argv, val_name, &bsscfg_idx, &consumed)) != 0)
		return ret;

	argv += consumed;

	if (!*argv) {
		int *val_ptr;
		if ((ret = wl_var_getbuf(wl, val_name, &bsscfg_idx, sizeof(bsscfg_idx), (void**)&val_ptr)) < 0)
			return ret;
		memcpy(&val, val_ptr, sizeof(int));
		if (val)
			printf("up\n");
		else
			printf("down\n");
		return 0;
	} else {
		struct {
			int cfg;
			int val;
		} setbuf;
		if (!stricmp(*argv, "up"))
			val = 1;
		else if (!stricmp(*argv, "down"))
			val = 0;
		else {
			val = strtol(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}
		}
		setbuf.cfg = bsscfg_idx;
		setbuf.val = val;

		return wl_var_setbuf(wl, val_name, &setbuf, sizeof(setbuf));
	}
}

/* Get/Set the gmode config */
static int
wl_gmode(void *wl, cmd_t *cmd, char **argv)
{
	char *endptr;
	int ret=0, val;

	if (!*++argv) {
		char *gconfig;

		/* Get the current G mode */
		if ((ret = wl_get(wl, cmd->get, &val, sizeof (val))))
			return -1;

		switch (val) {
		case GMODE_LEGACY_B:
			gconfig = "54g Legacy B";
			break;
		case GMODE_AUTO:
			gconfig = "54g Auto";
			break;
		case GMODE_ONLY:
			gconfig = "54g Only";
			break;
		case GMODE_PERFORMANCE:
			gconfig = "54g Performance";
			break;
		case GMODE_LRS:
			gconfig = "54g LRS";
			break;
		default:
			gconfig = "unknown";
			break;
		}

		printf("%s (%d)\n", gconfig, val);

	} else {
		/* Set the new G mode */
#ifndef DSLCPE		
		wlc_ssid_t ssid;
#endif
		if (!strnicmp(*argv, "legacy", 6))
			val = GMODE_LEGACY_B;
		else if (!strnicmp(*argv, "auto", 4))
			val = GMODE_AUTO;
		else if (!strnicmp(*argv, "gonly", 5))
			val = GMODE_ONLY;
		else if (!strnicmp(*argv, "perf", 4))
			val = GMODE_PERFORMANCE;
		else if (!strnicmp(*argv, "lrs", 3))
			val = GMODE_LRS;
		else {
			val = strtol(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}
		}

		/* Set the gmode configration */
		if ((ret = wl_set(wl, cmd->set, &val, sizeof (val))))
			goto done;

#ifndef DSLCPE
		/* check if we are in AP mode */
		if ((ret = wl_get(wl, WLC_GET_AP, &val, sizeof (val))))
			goto done;
		if (val) {
			/* Get and reset the ssid (in the AP case this restarts the AP) */
			if ((ret = wl_get(wl, WLC_GET_SSID, &ssid, sizeof (wlc_ssid_t))))
				goto done;
			if ((ret = wl_set(wl, WLC_SET_SSID, &ssid, sizeof (wlc_ssid_t))))
				goto done;
		}
#endif		
	}

done:
 	return(ret);
}

static int
wl_reg(void *wl, cmd_t *cmd, char **argv)
{
	int reg;
	int ret;
	struct {
		int val;
		int band;
	} x;
	char *endptr;
	uint argc;

	/* eat command name */
	argv++;

	/* arg count */
	for (argc = 0; argv[argc]; argc++)
		;

	/* required arg: reg offset */
	if (argc < 1)
		return -1;

	reg = strtol(argv[0], &endptr, 0);
	if (*endptr != '\0')
		return -1;

	x.band = WLC_BAND_AUTO;

	/* Second arg: value or band */
	if (argc >= 2) {
		if (!stricmp(argv[1], "a"))
			x.band = WLC_BAND_A;
		else if (!stricmp(argv[1], "b"))
			x.band = WLC_BAND_B;
		else {
			x.val = strtol(argv[1], &endptr, 0);
			if (*endptr != '\0')
				return (-1);
		}
	}

	/* Third arg: band */
	if (argc >= 3) {
		if (!stricmp(argv[2], "a"))
			x.band = WLC_BAND_A;
		else if (!stricmp(argv[2], "b"))
			x.band = WLC_BAND_B;
		else
			return (-1);
	}

	x.val = (x.val << 16) | (reg & 0xffff);

	/* issue the get or set ioctl */
	if ((argc == 1) || ((argc == 2) && (x.band != WLC_BAND_AUTO))) {
		if ((ret = wl_get(wl, cmd->get, &x, sizeof (x))) < 0)
			return (ret);
		printf("0x%04x\n", (uint16)x.val);
	}
	else
		ret = wl_set(wl, cmd->set, &x, sizeof (x));

	return (ret);
}

static int
wl_macreg(void *wl, cmd_t *cmd, char **argv)
{
	int reg;
	int size;
	uint32 val;
	int ret;
	char *endptr;
        rw_reg_t rwt;
	uint argc;

	val = 0;

	/* eat command name */
	argv++;

	/* arg count */
	for (argc = 0; argv[argc]; argc++)
		;

	/* required arg: reg offset */
	if (argc < 1)
		return -1;

	reg = strtol(argv[0], &endptr, 0);
	if (*endptr != '\0')
		return -1;

	/* required arg: reg size */
	if (argc < 2)
		return (-1);

	size = strtol(argv[1], &endptr, 0);
	if (*endptr != '\0')
		return -1;

	rwt.band = WLC_BAND_AUTO;

	/* Third arg: new value or band */
	if (argc >= 3) {
		if (!stricmp(argv[2], "a"))
			rwt.band = WLC_BAND_A;
		else if (!stricmp(argv[2], "b"))
			rwt.band = WLC_BAND_B;
		else {
			val = strtoul(argv[2], &endptr, 0);
			if (*endptr != '\0')
				return -1;
		}

	}

	/* Fourth arg: band */
	if (argc >= 4) {
		if (!stricmp(argv[3], "a"))
			rwt.band = WLC_BAND_A;
		else if (!stricmp(argv[3], "b"))
			rwt.band = WLC_BAND_B;
		else
			return (-1);
	}

	if ((argc == 2) || ((argc == 3) && (rwt.band != WLC_BAND_AUTO))) {
		rwt.byteoff = reg;
		rwt.size = size;
		if ((ret = wl_get(wl, cmd->get, &rwt, sizeof (rw_reg_t))) < 0)
			return (ret);
		printf("0x%04x\n", rwt.val);
	}
	else {
		rwt.byteoff = reg;
		rwt.size = size;
		rwt.val = val;
		ret = wl_set(wl, cmd->set, &rwt, sizeof (rw_reg_t));
	}

	return (ret);
}

/* 
 * get or get a band specific variable
 * the band can be a/b/all or omitted. "all"(set only) means all supported bands. blank means current band
 */
static int
wl_band_elm(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	struct {
		int val;
		int band;
	} x;
	char *endptr;
	uint argc;

	/* eat command name */
	argv++;

	/* arg count */
	for (argc = 0; argv[argc]; argc++)
		;

	x.val = 0;
	x.band = WLC_BAND_AUTO;

	/* First arg: value or band */
	if (argc >= 1) {
		if (!stricmp(argv[0], "a"))
			x.band = WLC_BAND_A;
		else if (!stricmp(argv[0], "b"))
			x.band = WLC_BAND_B;
		else if (!stricmp(argv[0], "all"))
			x.band = WLC_BAND_ALL;
		else {
			x.val = strtol(argv[0], &endptr, 0);
			if (*endptr != '\0')
				return (-1);
		}
	}

	/* Second arg: band */
	if (argc >= 2) {
		if (!stricmp(argv[1], "a"))
			x.band = WLC_BAND_A;
		else if (!stricmp(argv[1], "b"))
			x.band = WLC_BAND_B;
		else if (!stricmp(argv[1], "all"))
			x.band = WLC_BAND_ALL;
		else
			return (-1);
	}

	/* issue the get or set ioctl */
	if ((argc == 0) || ((argc == 1) && (x.band != WLC_BAND_AUTO))) {
		if (x.band == WLC_BAND_ALL) {
			printf("band option \"all\" is for set only, not get\n");
			return (-1);
		}

		if ((ret = wl_get(wl, cmd->get, &x, sizeof (x))) < 0)
			return (ret);

		printf("%s is 0x%04x(%d)\n", cmd->name, (uint16)x.val, x.val);
	}
	else
		ret = wl_set(wl, cmd->set, &x, sizeof (x));

	return (ret);
}

/* Command may or may not take a MAC address */
static int
wl_rssi(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	scb_val_t scb_val;

	if (!*++argv) {
		if ((ret = wl_get(wl, cmd->get, &scb_val.val, sizeof(scb_val))) < 0)
			return ret;
		printf("%d\n", scb_val.val);
		return 0;
	} else {
		if (!wl_ether_atoe(*argv, (char*)&scb_val.ea))
			return -1;
		if ((ret = wl_get(wl, cmd->get, &scb_val.val, sizeof(scb_val))) < 0)
			return ret;
		printf("%d\n", scb_val.val);
		return 0;
	}
}	

/* Commands that take a MAC address */
static int
wl_mac(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	struct ether_addr ea;

	if (!*++argv) {
		if ((ret = wl_get(wl, cmd->get, &ea, ETHER_ADDR_LEN)) < 0)
			return ret;
		printf("%s\n", wl_ether_etoa(ea.octet));
		return 0;
	} else {
		if (!wl_ether_atoe(*argv, ea.octet))
			return -1;
		return wl_set(wl, cmd->set, &ea, ETHER_ADDR_LEN);
	}
}


static int
wlu_dump(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	char *dump_buf;
	
	if (cmd->get < 0)
		return -1;
	
	dump_buf = malloc(WL_DUMP_BUF_LEN);
	if (dump_buf == NULL) {
		fprintf(stderr, "Failed to allocate dump buffer of %d bytes\n", WL_DUMP_BUF_LEN);
		return -1;
	}

	if ((ret = wl_get(wl, cmd->get, dump_buf, WL_DUMP_BUF_LEN)) < 0)
		return ret;
	fputs(dump_buf, stdout);

	return 0;
}


static int
wlu_srdump(void *wl, cmd_t *cmd, char **argv)
{
	int ret, x, y;
	uint16 *words = (uint16 *)&buf[8];

        srom_rw_t   *srt;

        srt = (srom_rw_t *)buf;
        srt->byteoff = 0;
#if defined(DSLCPE) && defined(BCMSDIO)
        srt->nbytes = 176;
#else        
        srt->nbytes = 128;
#endif        

	if (cmd->get < 0)
		return -1;
	if ((ret = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
		return ret;
#if defined(DSLCPE) && defined(BCMSDIO)		
	for ( x = 0; x < 11; x++ ) {
#else
	for ( x = 0; x < 8; x++ ) {
#endif		
		printf("  srom[%03d]:  ", (x*8) );
		for (y = 0; y < 8; y++ )
			printf("0x%04x  ", words[(x*8)+y] );
		printf("\n");
	}

	return 0;
}


static int
wl_locale(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	int val;
	char *endptr;
	int i;
	char **name;

	if (!*++argv) {
		if ((ret = wl_get(wl, cmd->get, &val, sizeof(int))) < 0)
			return ret;
		for (i = 0; i < (int)ARRAYSIZE(wl_locales); i++)
			if (wl_locales[i].val == val) {
				printf("%s\n", wl_locales[i].names[0]);
				return 0;
			}
		fprintf(stderr, "unknown: %d\n", val);
		return 1;
	} else {
		/* Try by name */
		for (i = 0; i < (int)ARRAYSIZE(wl_locales); i++)
			for (name = wl_locales[i].names; *name; name++)
				if (!strnicmp(*name, *argv, strlen(*argv)))
					return wl_set(wl, cmd->set, &wl_locales[i].val, sizeof(int));
		/* Try by value for backward compatibilty */
		val = strtol(*argv, &endptr, 0);
		if (*endptr != '\0') {
			fprintf(stderr, "unknown locale: %s\n", *argv);
			return 0;
		}

		for (i = 0; i < (int)ARRAYSIZE(wl_locales); i++)
			if (wl_locales[i].val == val)
				return wl_set(wl, cmd->set, &wl_locales[i].val, sizeof(int));
		return 1;
	}
}

static struct {
	uint value;
	char *string;
} dbg_msg[] = {
	{WL_ERROR_VAL, 	"error"},
	{WL_ERROR_VAL, 	"err"},
	{WL_TRACE_VAL, 	"trace"},
	{WL_PRHDRS_VAL,	"prhdrs"},
	{WL_PRPKT_VAL,	"prpkt"},
	{WL_INFORM_VAL,	"inform"},
	{WL_INFORM_VAL,	"info"},
	{WL_INFORM_VAL,	"inf"},
	{WL_TMP_VAL,	"tmp"},
	{WL_OID_VAL,	"oid"},
	{WL_RATE_VAL,	"rate"},
	{WL_ASSOC_VAL,	"assoc"},
	{WL_ASSOC_VAL,	"as"},
	{WL_PRUSR_VAL,	"prusr"},
	{WL_PS_VAL,	"ps"},
	{WL_TXPWR_VAL,	"txpwr"},
	{WL_TXPWR_VAL,	"pwr"},
	{WL_GMODE_VAL,	"gmode"},
	{WL_DUAL_VAL,	"dual"},
	{WL_WSEC_VAL,	"wsec"},
	{WL_WSEC_DUMP_VAL,"wsec_dump"},
	{WL_LOG_VAL,	"log"},
	{WL_NRSSI_VAL,	"nrssi"},
	{WL_LOFT_VAL,	"loft"},
	{WL_REGULATORY_VAL,	"regulatory"},
	{WL_ACI_VAL,	"aci"},
	{WL_RADAR_VAL,	"radar"},
	{WL_MPC_VAL,	"mpc"},
	{WL_APSTA_VAL,	"apsta"},
	{WL_DFS_VAL,	"dfs"},
	{0,		NULL}
};

static int
wl_msglevel(void *wl, cmd_t *cmd, char **argv)
{
	int ret, val, msglevel, i;
	char *endptr;
	int last_val;
	
	if (!*++argv) {
		if ((ret = wl_get(wl, cmd->get, &val, sizeof (int))) < 0)
			return (ret);
		printf("0x%x\n", val);
		return (0);
	}

	msglevel = 0;
	while (*argv) {
		val = strtol(*argv, &endptr, 0);
		if (*endptr == '\0') {
			msglevel |= val;
		} else {
			/* not all the value string was parsed by strtol */
			for (i = 0; dbg_msg[i].value; i++)
				if (stricmp(dbg_msg[i].string, *argv) == 0) {
					msglevel |= dbg_msg[i].value;
					break;
				}
			if (dbg_msg[i].value == 0)
				goto usage;
		}
		++argv;
	} 

	return (wl_set(wl, cmd->set, &msglevel, sizeof (int)));

usage:
	fprintf(stderr, "msg values may be a list of numbers or names from the following set:");
	last_val = -1;
	for (i = 0; (val = dbg_msg[i].value); i++) {
		if (val != last_val)
			fprintf(stderr, "\n0x%04x %s", val, dbg_msg[i].string);
		else
			fprintf(stderr, ", %s", dbg_msg[i].string);
		last_val = val;
	}
	fprintf(stderr, "\n");

	return 0;
}

/* take rate arg in units of 500Kbits/s and print it in units of Mbit/s */
static void
wl_printrate(int val)
{
	char buf[32];

	printf("%s\n", rate_int2string(buf, val));
}

/* convert rate string in Mbit/s format, like "11", "5.5", to internal 500 Kbit/s units */
static int
rate_string2int(char *s)
{
	if (!stricmp(s, "-1"))
		return (0);
	if (!stricmp(s, "5.5"))
		return (11);
	return (atoi(s) * 2);
}

/* convert rate internal 500 Kbits/s units to string in Mbits/s format, like "11", "5.5" */
static char*
rate_int2string(char *buf, int val)
{
	if ((val == -1) || (val == 0))
		sprintf(buf, "auto");
	else
		sprintf(buf, "%d%s Mbps", (val / 2), (val & 1) ? ".5" : "");
	return (buf);
}

/* handles both "rate" and "mrate", which makes the flow a bit complex */
static int
wl_rate_mrate(void *wl, cmd_t *cmd, char **argv)
{
	int error;
	int val;
	int band;
	int list[3];
	char aname[sizeof("a_mrate") + 1];
	char bgname[sizeof("bg_mrate") + 1];
	char *name;

	sprintf(aname, "a_%s", *argv);
	sprintf(bgname, "bg_%s", *argv);


	if ((error = wl_get(wl, WLC_GET_BAND, &band, sizeof (uint))) < 0)
		return error;

	if ((error = wl_get(wl, WLC_GET_BANDLIST, list, sizeof (list))) < 0)
		return error;

	if (!list[0]) 
		return -1;
	else if(list[0] > 2)
		list[0] = 2;

	if ((!strcmp(*argv, "rate"))) {
		/* it is "rate" */
		if (!*++argv) {
			/* it is "rate" get. handle it here */
			/* WLC_GET_RATE processing */
			if ((error = wl_get(wl, cmd->get, &val, sizeof (int))) < 0)
				return error;
			
			wl_printrate(val);
			return 0;
		}
		else
			--argv; /* move it back for later processing */
	}

	switch (band) {
	case WLC_BAND_AUTO : 
		if(list[0] > 1)
			return -1;
		else if (list[1] == WLC_BAND_A)
			name = (char *)aname;
		else if (list[1] == WLC_BAND_B)
			name = (char *)bgname;
		else
			return -1;			
	
		break;
		
	case WLC_BAND_A : 
		name = (char *)aname;
		break;
		
	case WLC_BAND_B :
		name = (char *)bgname;
		break;
	
	default :
		return -1;
		break;
	}

	if (!*++argv) {
		/* it is "mrate" get */
		if ((error = wl_iovar_getint(wl, name, &val) < 0))
			return error;
		
		if (val == 0)
			printf("auto\n");
		else
			wl_printrate(val);
			
		return 0;			
	} else {
		val = rate_string2int(*argv);
		return wl_iovar_setint(wl, name, val);
	}		
}

static int
wl_wepstatus(void *wl, cmd_t *cmd, char **argv)
{
	int val, error;
	char *name = "wsec";
	int wsec;

	if (!*++argv) {
		if ((error = wl_iovar_getint(wl, name, &val) < 0))
			return error;

		printf("%d\n", val);
		return 0;
	} else {
		val = atoi(*argv);
		if ((error = wl_iovar_getint(wl, name, &wsec) < 0))
			return error;

		if(val)
			wsec |= WEP_ENABLED;
		else
			wsec &= ~WEP_ENABLED;	
		
		return wl_iovar_setint(wl, name, wsec);
	}		
}

static int
wl_phy_rate(void *wl, cmd_t *cmd, char **argv)
{
	int val, error;
	int *pval;
	char *p;

	if (!*++argv) {
		pval = (int *) buf;
		strcpy(buf, cmd->name);
		if ((error = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
			return (error);
		val = *pval;
		wl_printrate(val);
		return (0);
	} else {
		val = rate_string2int(*argv);

		/* construct an iovar */
		strcpy(buf, cmd->name);
		p = &buf[strlen(buf) + 1];
		memcpy(p, &val, sizeof (uint));
		p += sizeof (uint);

		return (wl_set(wl, cmd->set, &buf, (p - buf)));
	}
}

static int
wl_assoc_info(void *wl, cmd_t *cmd, char **argv)
{
	uint i, req_ies_len=0, resp_ies_len=0;
	wl_assoc_info_t assoc_info;
	int ret;
	uint8 *pbuf;

	/* get the generic association information */
	strcpy(buf, cmd->name);
	if ((ret = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
		return ret;

	memcpy(&assoc_info, buf, sizeof(wl_assoc_info_t));

	printf("Assoc req:\n");
	printf("\tlen 0x%x\n", assoc_info.req_len);
	if (assoc_info.req_len) {
		printf("\tcapab  0x%x\n", ltoh16(assoc_info.req.capability));
		printf("\tlisten 0x%x\n", ltoh16(assoc_info.req.listen)); 
		req_ies_len = assoc_info.req_len - sizeof(struct dot11_assoc_req);
		if (assoc_info.flags & WLC_ASSOC_REQ_IS_REASSOC) {
			printf("\treassoc bssid %s\n", wl_ether_etoa(&assoc_info.reassoc_bssid.octet[0]));
			req_ies_len -= ETHER_ADDR_LEN;
		}
	}

	/* get the association req IE's if there are any */
	if (req_ies_len) {
		strcpy(buf, "assoc_req_ies");
		if ((ret = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
			return ret;

		printf("assoc req IEs:\n\t");
		for(i=1, pbuf=(uint8*)buf; i<=req_ies_len; i++) {
			printf("0x%02x ", *pbuf++);
			if (!(i%8))
				printf("\n\t");
		}
	}

	printf("\nAssoc resp:\n");
	printf("\tlen 0x%x\n", assoc_info.resp_len);
	if (assoc_info.resp_len) {
		printf("\tcapab  0x%x\n", ltoh16(assoc_info.resp.capability));
		printf("\tstatus 0x%x\n", ltoh16(assoc_info.resp.status)); 
		printf("\taid    0x%x\n", ltoh16(assoc_info.resp.aid)); 
		resp_ies_len = assoc_info.resp_len - sizeof(struct dot11_assoc_resp);
    }

	/* get the association resp IE's if there are any */
	if (resp_ies_len) {
		strcpy(buf, "assoc_resp_ies");
		if ((ret = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
			return ret;

		printf("assoc resp IEs:\n\t");
		for(i=1, pbuf=(uint8*)buf; i<=resp_ies_len; i++) {
			printf(" 0x%02x ", *pbuf++);
			if (!(i%8))
				printf("\n\t");
	
		}
	}
	printf("\n");
	
	return 0;
}

#ifdef BCMWPA2
static int
wl_pmkid_info(void *wl, cmd_t *cmd, char**argv)
{
	int i,j,ret;
	int *pval;
	pmkid_list_t *pmkid_info;

	if (!*++argv) {
		pval = (int *) buf;
		strcpy(buf, cmd->name);
		if ((ret = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
			return ret;
	
		pmkid_info = (pmkid_list_t *)buf;
		printf("\npmkid entries : %d\n", pmkid_info->npmkid);

		for(i=0; i<(int)pmkid_info->npmkid; i++) {
			printf("\tPMKID[%d]: %s =", i, wl_ether_etoa((char*)&pmkid_info->pmkid[i].BSSID.octet[0]));
			for (j=0; j < WPA2_PMKID_LEN; j++)
				printf("%02x ", pmkid_info->pmkid[i].PMKID[j]);
			printf("\n");
		}
	} 
	else {
#ifdef test_pmkid_info	
		char eaddr[6] = {0x0, 0x0, 0x1, 0x2, 0x3, 0x5};
		char eaddr1[6] = {0xa, 0xb, 0xc, 0xd, 0xe, 0xf};
		char id[WPA2_PMKID_LEN], id1[WPA2_PMKID_LEN];
		int i, len=(sizeof(uint32)+2*(sizeof(pmkid_t)));

		/* check that the set uses to "test" cmd */
		if (strcmp(*argv, "test")) {
			printf("\t wl pmkid_info only supports `test` a test specific set\n");
			return -1;
		}
		if ((pmkid_info = (pmkid_list_t *)malloc(len)) == NULL)	{
			printf("\tfailed to allocate buffer\n");
			return -1;
		}
		
		printf("\toverwriting pmkid table with test pattern\n"); 
		for(i=0; i<(int)sizeof(id); i++) {
			id[i] = i;
			id1[i] = (i*2);
		}

		/* "test" - creates two PMKID entries and sets the table to that */
		pmkid_info->npmkid = 2;
		memcpy(&pmkid_info->pmkid[0].BSSID.octet[0], &eaddr[0], ETHER_ADDR_LEN);
		memcpy(&pmkid_info->pmkid[0].PMKID[0], &id[0], WPA2_PMKID_LEN);
		memcpy(&pmkid_info->pmkid[1].BSSID.octet[0], &eaddr1[0], ETHER_ADDR_LEN);
		memcpy(&pmkid_info->pmkid[1].PMKID[0], &id1[0], WPA2_PMKID_LEN);

		ret = wl_var_setbuf(wl, cmd->name, pmkid_info, len);

		free(pmkid_info);
		
		return ret;
#else
		printf("\tset cmd ignored\n");
#endif
	}
		
	return 0;
}
#endif

static int
wl_rateset(void *wl, cmd_t *cmd, char **argv)
{
	wl_rateset_t rs, defrs;
	int error;
	uint i;

	error = 0;

	argv++;

	if (*argv == NULL) {
		/* get current rateset */
		if ((error = wl_get(wl, WLC_GET_CURR_RATESET, &rs, sizeof (wl_rateset_t))) < 0)
			return (error);

		dump_rateset(rs.rates, rs.count);
		printf("\n");
	} else {
		/* get default rateset */
		if ((error = wl_get(wl, WLC_GET_RATESET, &defrs, sizeof (wl_rateset_t))) < 0)
			return (error);

		if (!stricmp(*argv, "all")) {
			for (i = 0; i < defrs.count; i++)
				defrs.rates[i] |= 0x80;
			error = wl_set(wl, WLC_SET_RATESET, &defrs, sizeof (wl_rateset_t));
		}
		else if (!stricmp(*argv, "default")) {
			error = wl_set(wl, WLC_SET_RATESET, &defrs, sizeof (wl_rateset_t));
		}
		else {	/* arbitrary list */
			wl_parse_rateset(&defrs, argv);
			/* check for common error of no basic rates */
			for (i = 0; i < defrs.count; i++) {
				if (defrs.rates[i] & 0x80)
					break;
			}
			if (i < defrs.count) {
				error = wl_set(wl, WLC_SET_RATESET, &defrs, sizeof (wl_rateset_t));
			} else {
				cmd_usage(cmd);
				fprintf(stderr, "Bad Args: at least one rate must be marked Basic\n");
			}

		}

	}
	return (error);
}

static int
wl_sup_rateset(void *wl, cmd_t *cmd, char **argv)
{
	wl_rateset_t rs;
	bool got_basic;
	int error;
	uint i;
	
	error = 0;
	memset((char*)&rs, 0, sizeof (wl_rateset_t));

	argv++;

	if (*argv == NULL) {
		/* get rateset */
		if ((error = wl_get(wl, cmd->get, &rs, sizeof (wl_rateset_t))) < 0)
			return (error);

		dump_rateset(rs.rates, rs.count);
		printf("\n");
	} else {
		if (!stricmp(*argv, "-1") || !stricmp(*argv, "0")) {
			/* set an empty rateset */
			error = wl_set(wl, cmd->set, &rs, sizeof (wl_rateset_t));
		}
		else {	/* set the specified rateset */
			wl_parse_rateset(&rs, argv);
			/* check for common error of including a basic rate */
			got_basic = FALSE;
			for (i = 0; i < rs.count; i++) {
				if (rs.rates[i] & 0x80) {
					rs.rates[i] &= 0x7F;
					got_basic = TRUE;
				}
			}
			if (got_basic) {
				fprintf(stderr, "Warning: Basic rate attribute ignored for \"%s\" command\n", cmd->name);
			}
			error = wl_set(wl, cmd->set, &rs, sizeof (wl_rateset_t));
		}

	}
	return (error);
}

static int
wl_parse_rateset(wl_rateset_t* rs, char **argv)
{
	char* endp;
	char* arg;
	int r;
	int error = 0;
	
	memset(rs, 0, sizeof(*rs));
	
	while ((arg = *argv++) != NULL) {
		if (rs->count >= WL_NUMRATES) {
			fprintf(stderr, "parsing \"%s\", too many rates specified, max is %d rates\n", arg, WL_NUMRATES);
			error = -1;
			break;
		}

		/* convert the rate number to a 500kbps rate by multiplying by 2 */
		r = (int)(strtoul(arg, &endp, 0) * 2);
		if (endp == arg) {
			fprintf(stderr, "unable to convert the rate parameter \"%s\"\n", arg);
			error = -1;
			break;
		}
		
		/* parse a .5 specially */
		if (!strncmp(endp, ".5", 2)) {
			r += 1;
			endp += 2;
		}

		/* strip trailing space */
		while (isspace((int)endp[0]))
			endp++;

		/* check for a basic rate specifier */
		if (!stricmp(endp, "b") || !stricmp(endp, "(b)")) {
			r |= 0x80;
		} else if (endp[0] != '\0') {
			fprintf(stderr, "unable to convert trailing characters \"%s\" in the rate parameter \"%s\"\n",
			       endp, arg);
			error = -1;
			break;
		}
		rs->rates[rs->count++] = r;
	}

	return error;
}

static int
wl_channel(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	channel_info_t ci;

	if (!*++argv) {
		memset(&ci, 0, sizeof(ci));
		if ((ret = wl_get(wl, cmd->get, &ci, sizeof(channel_info_t))) < 0)
			return ret;
		if (ci.scan_channel) {
			printf("Scan in progress.\n");
			printf("current scan channel\t%d\n", ci.scan_channel);
		} else {
			printf("No scan in progress.\n");
		}
		printf("current mac channel\t%d\n", ci.hw_channel);
		printf("target channel\t%d\n", ci.target_channel);
		return 0;
	} else {
		ci.target_channel = atoi(*argv);
		ret =  wl_set(wl, cmd->set, &ci.target_channel, sizeof(int));
		return ret;
	}
}

int
wl_ether_atoe(const char *a, unsigned char *n)
{
	char *c = (char *) a;
	int i = 0;

	memset(n, 0, ETHER_ADDR_LEN);
	for (;;) {
		n[i++] = (unsigned char) strtoul(c, &c, 16);
		if (!*c++ || i == ETHER_ADDR_LEN)
			break;
	}
	return (i == ETHER_ADDR_LEN);
}

char *
wl_ether_etoa(const unsigned char *n)
{
	static char etoa_buf[ETHER_ADDR_LEN * 3];
	char *c = etoa_buf;
	int i;

	for (i = 0; i < ETHER_ADDR_LEN; i++) {
		if (i)
			*c++ = ':';
		c += sprintf(c, "%02X", n[i] & 0xff);
	}
	return etoa_buf;
}

int
wl_format_ssid(char* buf, uint8* ssid, int ssid_len)
{
	int i, c;
	char *p = buf;

	if (ssid_len > 32) ssid_len = 32;
	
	for (i = 0; i < ssid_len; i++) {
		c = (int)ssid[i];
		if (c == '\\') {
			*p++ = '\\';
			*p++ = '\\';
		} else if (isprint((uchar)c)) {
			*p++ = (char)c;
		} else {
			p += sprintf(p, "\\x%02X", c);
		}
	}
	*p = '\0';
	
	return p - buf;
}

static int
wl_plcphdr(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	int val;

	if (!*++argv) {
		if ((ret = wl_get(wl, cmd->get, &val, sizeof(int))) < 0)
			return ret;
		if (val == WLC_PLCP_AUTO)
			printf("long");
		else if (val == WLC_PLCP_SHORT)
			printf("auto");
		else if (val == WLC_PLCP_LONG)
			printf("debug");
		else
			printf("unknown");
		printf("\n");
		return 0;
	} else {
		if (!stricmp(*argv, "long"))
			val = WLC_PLCP_AUTO;
		else if (!stricmp(*argv, "auto"))
			val = WLC_PLCP_SHORT;
		else if (!stricmp(*argv, "debug"))
			val = WLC_PLCP_LONG;
		else
			return -1;
		return wl_set(wl, cmd->set, &val, sizeof(int));
	}
}

/* WLC_GET_RADIO and WLC_SET_RADIO in driver operate on radio_disabled which
 * is opposite of "wl radio [1|0]".  So invert for user.
 * In addition, display WL_RADIO_SW_DISABLE and WL_RADIO_HW_DISABLE bits.
 */
static int
wl_radio(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	uint val;

	if (!*++argv) {
		if (cmd->get < 0)
			return -1;			
		if ((ret = wl_get(wl, cmd->get, &val, sizeof(int))) < 0)
			return ret;
		printf("0x%04x\n", val);
		return 0;
	} else {
		if (cmd->set < 0)
			return -1;
		if (!stricmp(*argv, "on"))
			val = WL_RADIO_SW_DISABLE << 16;
		else if (!stricmp(*argv, "off"))
			val = WL_RADIO_SW_DISABLE << 16 | WL_RADIO_SW_DISABLE;
		else
			val = atoi(*argv);
		return wl_set(wl, cmd->set, &val, sizeof(int));
	}
}

static char *
ver2str(unsigned int vms, unsigned int vls)
{
	static char verstr[100];
	unsigned int maj, year, month, day, build;

	maj = (vms >> 16) & 0xFFFF;
	if (maj > 1000) {
		/* it is probably a date... */
		year = (vms >> 16) & 0xFFFF;
		month = vms & 0xFFFF;
		day = (vls >> 16) & 0xFFFF;
		build = vls & 0xFFFF;
		sprintf(verstr, "%d/%d/%d build %d", 
				month, day, year, build);
	} else {
		/* it is a tagged release. */
		sprintf(verstr, "%d.%d RC%d.%d", 
				(vms>>16)&0xFFFF, vms&0xFFFF, 
				(vls>>16)&0xFFFF, vls&0xFFFF);
	}
	return verstr;
}

static int
wl_version(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	char *p;

	printf("%s\n",
	       ver2str((EPI_MAJOR_VERSION << 16)| EPI_MINOR_VERSION,
		       (EPI_RC_NUMBER << 16) | EPI_INCREMENTAL_NUMBER));

	if ((ret = wl_get(wl, WLC_DUMP, buf, WLC_IOCTL_MAXLEN)) < 0)
		return ret;

	/* strip the first line off the dump buf output */
	p = strchr(buf, '\n');
	if (p) 
		*p = '\0';
	printf("%s\n", buf);
	return 0;
}

static int
wl_rateparam(void *wl, cmd_t *cmd, char **argv)
{
	int val[2];

	if (!*++argv)
		return -1;
	val[0] = atoi(*argv);
	if (!*++argv)
		return -1;
	val[1] = atoi(*argv);
	return wl_set(wl, cmd->set, val, 2 * sizeof(val));
}

/* wl scan
 * -s --ssid=ssid 
 * -t T --scan_type=T : [active|passive]
 * --bss_type=T : [infra|bss|adhoc|ibss]
 * -b --bssid=
 * -n --nprobes=
 * -a --active=
 * -p --passive=
 * -h --home=
 * -c --channels=
 */
static int
wl_scan(void *wl, cmd_t *cmd, char **argv)
{
	wl_scan_params_t* params;
	int params_size = WL_SCAN_PARAMS_FIXED_SIZE + WL_NUMCHANNELS * sizeof(uint16);
	int val = 0;
	char key[64];
	int keylen;
	char *p, *eq, *valstr, *endptr;
	char opt;
	bool positional_param;
	bool good_int;
	bool opt_end;
	int err = 0;
	
	params = (wl_scan_params_t*)malloc(params_size);
	if (params == NULL) {
		fprintf(stderr, "Error allocating %d bytes for scan params\n", params_size);
		return -1;
	}
	memset(params, 0, params_size);

	params->bss_type = DOT11_BSSTYPE_ANY;
	memcpy(&params->bssid, &ether_bcast, ETHER_ADDR_LEN);
	params->scan_type = -1;
	params->nprobes = -1;
	params->active_time = -1;
	params->passive_time = -1;
	params->home_time = -1;
	params->channel_num = 0;

	/* skip the command name */
	argv++;
	
	opt_end = FALSE;
	while ((p = *argv) != NULL) {
		argv++;
		positional_param = FALSE;
		memset(key, 0, sizeof(key));
		opt = '\0';
		valstr = NULL;
		good_int = FALSE;

		if (opt_end) {
			positional_param = TRUE;
			valstr = p;
		}
		else if (!strcmp(p, "--")) {
			opt_end = TRUE;
			continue;
		}
		else if (!strncmp(p, "--", 2)) {
			eq = strchr(p, '=');
			if (eq == NULL) {
				fprintf(stderr, "wl_scan: missing \"=\" in long param \"%s\"\n", p);
				err = -1;
				goto exit;
			}
			keylen = eq - (p + 2);
			if (keylen > 63) keylen = 63;
			memcpy(key, p + 2, keylen);

			valstr = eq + 1;
			if (*valstr == '\0') {
				fprintf(stderr, "wl_scan: missing value after \"=\" in long param \"%s\"\n", p);
				err = -1;
				goto exit;
			}
		}
		else if (!strncmp(p, "-", 1)) {
			opt = p[1];
			if (strlen(p) > 2) {
				fprintf(stderr, "wl_scan: only single char options, error on param \"%s\"\n", p);
				err = -1;
				goto exit;
			}
			if (*argv == NULL) {
				fprintf(stderr, "wl_scan: missing value parameter after \"%s\"\n", p);
				err = -1;
				goto exit;
			}
			valstr = *argv;
			argv++;
		} else {
			positional_param = TRUE;
			valstr = p;
		}

		/* parse valstr as int just in case */
		if (valstr) {
			val = (int)strtol(valstr, &endptr, 0);
			if (*endptr == '\0') {
				/* not all the value string was parsed by strtol */
				good_int = TRUE;
			}
		}
		
		if (opt == 's' || !strcmp(key, "ssid") || positional_param) {
			if (strlen(valstr) > 32) {
				fprintf(stderr, "SSID arg \"%s\" must be 32 chars or less\n", valstr);
				err = -1;
				goto exit;
			}
			strncpy(params->ssid.SSID, valstr, sizeof(params->ssid.SSID));
			params->ssid.SSID_len = strlen(valstr);
		}
		if (opt == 't' || !strcmp(key, "scan_type")) {
			if (!strcmp(valstr, "active")) {
				params->scan_type = DOT11_SCANTYPE_ACTIVE;
			} else if (!strcmp(valstr, "passive")) {
				params->scan_type = DOT11_SCANTYPE_PASSIVE;
			} else {
				fprintf(stderr, "scan_type value should be \"active\" or \"passive\", but got \"%s\"\n", valstr);
				err = -1;
				goto exit;
			}
		}
		if (!strcmp(key, "bss_type")) {
			if (!strcmp(valstr, "bss") || !strcmp(valstr, "infra")) {
				params->bss_type = DOT11_BSSTYPE_INFRASTRUCTURE;
			} else if (!strcmp(valstr, "ibss") || !strcmp(valstr, "adhoc")) {
				params->bss_type = DOT11_BSSTYPE_INDEPENDENT;
			} else if (!strcmp(valstr, "any")) {
				params->bss_type = DOT11_BSSTYPE_ANY;
			} else {
				fprintf(stderr, "bss_type value should be \"bss\", \"ibss\", or \"any\", but got \"%s\"\n", valstr);
				err = -1;
				goto exit;
			}
		}
		if (opt == 'b' || !strcmp(key, "bssid")) {
			if (!wl_ether_atoe(valstr, params->bssid.octet)) {
				fprintf(stderr, "could not parse \"%s\" as an ethternet MAC address\n", valstr);
				err = -1;
				goto exit;
			}
		}
		if (opt == 'n' || !strcmp(key, "nprobes")) {
			if (!good_int) {
				fprintf(stderr, "could not parse \"%s\" as an int for value nprobes\n", valstr);
				err = -1;
				goto exit;
			}
			params->nprobes = val;
		}
		if (opt == 'a' || !strcmp(key, "active")) {
			if (!good_int) {
				fprintf(stderr, "could not parse \"%s\" as an int for active dwell time\n", valstr);
				err = -1;
				goto exit;
			}
			params->active_time = val;
		}
		if (opt == 'p' || !strcmp(key, "passive")) {
			if (!good_int) {
				fprintf(stderr, "could not parse \"%s\" as an int for passive dwell time\n", valstr);
				err = -1;
				goto exit;
			}
			params->passive_time = val;
		}
		if (opt == 'h' || !strcmp(key, "home")) {
			if (!good_int) {
				fprintf(stderr, "could not parse \"%s\" as an int for home channel dwell time\n", valstr);
				err = -1;
				goto exit;
			}
			params->home_time = val;
		}
		if (opt == 'c' || !strcmp(key, "channels")) {
			params->channel_num = wl_parse_channel_list(valstr, params->channel_list, WL_NUMCHANNELS);
			if (params->channel_num == -1) {
				fprintf(stderr, "error parsing channel list arg\n");
				err = -1;
				goto exit;
			}
		}
	}

	params_size = WL_SCAN_PARAMS_FIXED_SIZE + params->channel_num * sizeof(uint16);
	err = wl_set(wl, cmd->set, params, params_size);

exit:	
	free(params);
	return err;
}

int
wl_parse_channel_list(char* list_str, uint16* channel_list, int channel_num)
{
	int num;
	int val;
	char* str;
	char* endptr;
	
	if (list_str == NULL)
		return -1;
	
	str = list_str;
	num = 0;
	while (*str != '\0') {
		val = (int)strtol(str, &endptr, 0);
		if (endptr == str) {
			fprintf(stderr, "could not parse channel number starting at substring \"%s\" in list:\n%s\n",
				str, list_str);
			return -1;
		}
		str = endptr + strspn(endptr, " ,");

		if (num == channel_num) {
			fprintf(stderr, "too many channels (more than %d) in channel list:\n%s\n",
				channel_num, list_str);
			return -1;
		}

		channel_list[num++] = (uint16)val;
	}

	return num;
}

/* channel info structure */
typedef struct {
	uint	chan;		/* channel number */
	uint	freq;		/* in Mhz */
} chan_info_t;

static chan_info_t chan_info[] = {
	/* B channels */
	{ 1,	2412},
	{ 2,	2417},
	{ 3,	2422},
	{ 4,	2427},
	{ 5,	2432},
	{ 6,	2437},
	{ 7,	2442},
	{ 8,	2447},
	{ 9,	2452},
	{ 10,	2457},
	{ 11,	2462},
	{ 12,	2467},
	{ 13,	2472},
	{ 14,	2484},

	/* A channels */
	/* 11a usa low */
	{ 36,	5180},
	{ 40,	5200},
	{ 44,	5220},
	{ 48,	5240},
	{ 52,	5260},
	{ 56,	5280},
	{ 60,	5300},
	{ 64,	5320},

	/* 11a Europe */
	{ 100,	5500},
	{ 104,	5520},
	{ 108,	5540},
	{ 112,	5560},
	{ 116,	5580},
	{ 120,	5600},
	{ 124,	5620},
	{ 128,	5640},
	{ 132,	5660},
	{ 136,	5680},
	{ 140,	5700},

	/* 11a usa high */
	{ 149,	5745},
	{ 153,	5765},
	{ 157,	5785},
	{ 161,	5805},

	/* 11a japan */
	{ 184,	4920},
	{ 188,	4940},
	{ 192,	4960},
	{ 196,	4980},
	{ 200,	5000},
	{ 204,	5020},
	{ 208,	5040},
	{ 212,	5060},
	{ 216,	5080}
};

uint
freq2channel(uint freq)
{
	int i;

	for (i = 0; i < (int)ARRAYSIZE(chan_info); i++) {
		if (chan_info[i].freq == freq)
			return (chan_info[i].chan);
	}
	return (0);
}

void
dump_rateset(uint8 *rates, uint count)
{
	uint i;
	uint r;
	bool b;

	printf("[ ");
	for (i = 0; i < count; i++) {
		r = rates[i] & 0x7f;
		b = rates[i] & 0x80;
		if (r == 0)
			break;
		printf("%d%s%s ", (r / 2), (r % 2)?".5":"", b?"(b)":"");
	}
	printf("]");
}

/* Helper routine to print the infrastructure mode while pretty printing the BSS list */
static const char *
capmode2str(uint16 capability)
{
	capability &= (DOT11_CAP_ESS | DOT11_CAP_IBSS);
	
	if (capability == DOT11_CAP_ESS)
		return "Managed";
	else if (capability == DOT11_CAP_IBSS)
		return "Ad Hoc";
	else
		return "<unknown>";
}

/* 
 * Traverse a string of 1-byte tag/1-byte length/variable-length value 
 * triples, returning a pointer to the substring whose first element 
 * matches tag
 */
static uint8 *
wlu_parse_tlvs(uint8 *buf, int buflen, uint key)
{
	uint8 *cp;
	int totlen;

	cp = buf;
	totlen = buflen;

	/* find tagged parameter */
	while (totlen >= 2) {
		uint tag;
		int len;

		tag = *cp;
		len = *(cp +1);
		
		/* validate remaining totlen */
		if ((tag == key) && (totlen >= (len + 2)))
			return (cp);

		cp += (len + 2);
		totlen -= (len + 2);
	}
	
	return NULL;
}

static int
wlu_bcmp(const void *b1, const void *b2, int len)
{
	return (memcmp(b1, b2, len));
}

/* Is this body of this tlvs entry a WPA entry? If */
/* not update the tlvs buffer pointer/length */
static bool 
wlu_is_wpa_ie(uint8 **wpaie, uint8 **tlvs, uint *tlvs_len)
{	
	uint8 *ie = *wpaie;

	/* If the contents match the WPA_OUI and type=1 */ 
	if ((ie[1] >= 6) && !wlu_bcmp(&ie[2], WPA_OUI "\x01", 4)) {
		return TRUE;
	}
	
	/* point to the next ie */
	ie += ie[1] + 2; 
	/* calculate the length of the rest of the buffer */
	*tlvs_len -= (int)ie - (int)*tlvs;	 
	/* update the pointer to the start of the buffer */
	*tlvs = ie;

	return FALSE;
}

static void
dump_bss_ie(uint8* cp, uint len)
{
	uint8 *wpaie;
	int i;
	int n;
	int offset;
	uint8 *parse = cp;
	uint parse_len = len;
	uint16 capabilities;
	int unicast_count = 0;
	uint8 oui[3];
	uint8 index=0;
	
	while ((wpaie = wlu_parse_tlvs(parse, parse_len, DOT11_MNG_WPA_ID)))
		if (wlu_is_wpa_ie(&wpaie, &parse, &parse_len))
			break; 
	if (wpaie == NULL)
		return;

	if ((wpaie[6] | (wpaie[7] << 8)) != WPA_VERSION) 
		return;	/* WPA version unsupported */

	printf("WPA:\n");

	/* Check for multicast suite */
	if (wpaie[1] >= 10) { 
		printf("\tmulticast cipher: ");
		if (!wlu_bcmp(&wpaie[8], WPA_OUI, 3)) {
			switch (wpaie[11]) {
			case WPA_CIPHER_NONE:
				printf("NONE\n");
				break;
			case WPA_CIPHER_WEP_40:
				printf("WEP64\n");
				break;
			case WPA_CIPHER_WEP_104:
				printf("WEP128\n");
				break;
			case WPA_CIPHER_TKIP:
				printf("TKIP\n");
				break;
			case WPA_CIPHER_AES_OCB:
				printf("AES-OCB\n");
				break;
			case WPA_CIPHER_AES_CCM:
				printf("AES-CCMP\n");
				break;
			default:	/* unknown WPA cipher */
				printf("WPA-Unknown(#%d)\n", wpaie[11]);
				break;
			}
		}
	}

	/* Check for unicast suite(s) */
	if (wpaie[1] >= 12) { 
		unicast_count = (wpaie[12] | (wpaie[13] << 8));
		printf("\tunicast ciphers(%d): ", unicast_count);
		for (i = 0 ; i < unicast_count; i++) {

			if (wpaie[1] < (12 + (i * 4) + 4))
				break;

			memcpy(oui, &wpaie[2 + 12 + (i * 4)], 3);
			index = wpaie[2 + 12 + (i * 4) + 3];
			
			if (!wlu_bcmp(oui, WPA_OUI, 3)) {
				switch (index) {
				case WPA_CIPHER_NONE:
					printf("NONE ");
					break;
				case WPA_CIPHER_WEP_40:
					printf("WEP64 ");
					break;
				case WPA_CIPHER_WEP_104:
					printf("WEP128 ");
					break;
				case WPA_CIPHER_TKIP:
					printf("TKIP ");
					break;
				case WPA_CIPHER_AES_OCB:
					printf("AES-OCB ");
					break;
				case WPA_CIPHER_AES_CCM:
					printf("AES-CCMP ");
					break;
				default:	
					printf("WPA-Unknown(#%d) ", index);
					break;
				}
			}
			else {
				printf("Unknown-%02X:%02X:%02X(#%d) ", oui[0], oui[1], oui[2], index);
			}
		}
		printf("\n");
	}
	/* Authentication Key Management*/
	if (wpaie[1] >= 16) { /* Fixed 8, Group 4 , 2 + min one Unicast 4 - 2 bytes of VerID and Len*/
	        offset = 8 + 4 + 2 + (unicast_count * 4) ;  
		n = wpaie[offset] + (wpaie[offset + 1] << 8 ); 
		printf("\tAKM Suites(%d): ", n);
		if (wpaie[1] < (offset + (n * 4))) {
			printf("parse seems weird offset:%d, len:%d\n", offset,wpaie[1]);
			return;
		}
		for (i = 0 ; i < n; i++) {

			memcpy(oui, &wpaie[offset + 2 + (i * 4)], 3);
			index = wpaie[offset + 2 + (i * 4) + 3];
			
			if (!wlu_bcmp(oui, WPA_OUI, 3)) {
				switch (index) {
				case RSN_AKM_NONE:
					printf("None ");
					break;
				case RSN_AKM_UNSPECIFIED:
					printf("WPA ");
					break;
				case RSN_AKM_PSK:
					printf("WPA-PSK ");
					break;
				default:	
					printf("WPA-Unknown(#%d)  ", index);
					break;
				}
			}
			else {
				printf("Unknown-%02X:%02X:%02X(#%d)  ", oui[0], oui[1], oui[2], index);
			}
		}
		printf("\n");
	} 
	/*Capabilities */
	if (wpaie[1] > 22) { 
		/* 
		** Fixed 8, Group 4 , 2 + min one Unicast 4 - 2 bytes of VerID and Len + 2 + 1 Auth Key
		** management
		*/
	        offset = 8 + 4 + 2 + ( (wpaie[12] | wpaie[13] << 8 ) * 4) ;  
		n = wpaie[offset] + (wpaie[offset + 1] << 8 ); 
		offset += 2 + (n * 4);
		if (wpaie[1] < (offset  - 2)) {
			printf("parse seems weird offset:%d, len:%d\n", offset,wpaie[1]);
			return;
		}
		if (wpaie[1] == (offset  - 2)) {
			printf("\tNo WPA Capabilities advertised\n");
			return;
		}
		capabilities = wpaie[offset] | (wpaie[offset + 1] << 8);
		printf("\tWPA Capabilities(0x%x): ", capabilities );
		if (capabilities & 0x0002) printf("No ");
		printf("Pairwise, "); 
		printf("%d ", 1<<((capabilities >> 2) & 0x0003));
		printf("PTK Replay Ctr(s)");
	}
	else
		printf("\tNo WPA Capabilities advertised\n");
}

/* Pretty print the BSS list */
static void
dump_networks(char *buf)
{
	wl_scan_results_t *list = (wl_scan_results_t*)buf;
	wl_bss_info_t *bi;
	uint i;

	if (list->count == 0)
		return;
	else if (list->version != WL_BSS_INFO_VERSION) {
		fprintf(stderr, "Sorry, your driver has bss_info_version %d "
			"but this program supports only version %d.\n",
			list->version, WL_BSS_INFO_VERSION);
		return;
	}

	bi = list->bss_info;
	for (i = 0; i < list->count; i++, bi = (wl_bss_info_t*)((int8*)bi + bi->length)) {
		dump_bss_info(bi);
	}
}

static void
dump_bss_info(wl_bss_info_t *bi)
{
	char ssidbuf[SSID_FMT_BUF_LEN];

	wl_format_ssid(ssidbuf, bi->SSID, bi->SSID_len);
		
	printf("SSID: \"%s\"", ssidbuf);
	printf("\n");

	printf("Mode: %s\t", capmode2str(bi->capability));
	printf("RSSI: %d dBm\t", bi->RSSI);
	printf("noise: %d dBm\t", bi->phy_noise);
	printf("Channel: %d", bi->channel);
	printf("\n");

	printf("BSSID: %s\t", wl_ether_etoa(bi->BSSID.octet));

	printf("Capability: ");
	if (bi->capability & DOT11_CAP_ESS) printf("ESS ");
	if (bi->capability & DOT11_CAP_IBSS) printf("IBSS ");
	if (bi->capability & DOT11_CAP_POLLABLE) printf("Pollable ");
	if (bi->capability & DOT11_CAP_POLL_RQ) printf("PollReq ");
	if (bi->capability & DOT11_CAP_PRIVACY) printf("WEP ");
	if (bi->capability & DOT11_CAP_SHORT) printf("ShortPre ");
	if (bi->capability & DOT11_CAP_PBCC) printf("PBCC ");
	if (bi->capability & DOT11_CAP_AGILITY) printf("Agility " );
	if (bi->capability & DOT11_CAP_SHORTSLOT) printf("ShortSlot ");
	if (bi->capability & DOT11_CAP_CCK_OFDM) printf("CCK-OFDM ");
	printf("\n");

	printf("Supported Rates: ");
	dump_rateset(bi->rateset.rates, bi->rateset.count);
	printf("\n");
	if (bi->ie_length)
		dump_bss_ie((uint8 *)(bi + 1), bi->ie_length);

	printf("\n\n");
}

static int
wl_dump_networks(void *wl, cmd_t *cmd, char **argv)
{
	int ret;

	ret = wl_get_scan(wl, WLC_SCAN_RESULTS, buf, WLC_IOCTL_MAXLEN);
	if (ret == 0)
		dump_networks(buf);

	return ret;
}

static int
wl_dump_chanlist(void *wl, cmd_t *cmd, char **argv)
{
	uint32 buf[WL_NUMCHANNELS + 1]; 
	wl_uint32_list_t *list;
	int ret;
	uint i;

	list = (wl_uint32_list_t *)buf;
	list->count = WL_NUMCHANNELS;
	ret = wl_get(wl, cmd->get, buf, sizeof(buf));
	if (ret < 0)
		return ret;
	
	for (i = 0; i < list->count; i++)
		printf("%d ", list->element[i]);
	printf("\n");
	return ret;
}

static int
wl_channels_in_country(void *wl, cmd_t *cmd, char **argv)
{
	wl_channels_in_country_t *cic;
	int ret;
	uint i, len;

	cic = (wl_channels_in_country_t *)buf;
	cic->buflen = sizeof(buf);
	cic->count = 0;

	/* country abbrev must follow */
	if (!*++argv) {
		fprintf(stderr, "missing country abbrev\n");
		return -1;
	}

	len = strlen(*argv);
	if ((len > 3) || (len < 2)) {
		fprintf(stderr, "invalid country abbrev: %s\n", *argv);
		return -1;
	}

	strcpy(cic->country_abbrev, *argv);

	/* band must follow */
	if (!*++argv) {
		fprintf(stderr, "missing band\n");
		return -1;
	}

	if (!stricmp(*argv, "a"))
		cic->band = WLC_BAND_A;
	else if (!stricmp(*argv, "b"))
		cic->band = WLC_BAND_B;
	else {
		fprintf(stderr, "unsupported band: %s\n", *argv);
		return -1;
	}

	ret = wl_get(wl, cmd->get, buf, cic->buflen);
	if (ret < 0)
		return ret;
	
	for (i = 0; i < cic->count; i++)
		printf("%d ", cic->channel[i]);
	printf("\n");

	return ret;
}

static int
wl_get_scan(void *wl, int opc, char *buf, uint buf_len)
{
	wl_scan_results_t *list = (wl_scan_results_t*)buf;
	int ret;

	list->buflen = buf_len;
	ret = wl_get(wl, opc, buf, buf_len);
	if (ret < 0)
		return ret;
	ret = 0;
	
	if (list->buflen == 0) {
		list->version = 0;
		list->count = 0;
	} else if (list->version != WL_BSS_INFO_VERSION) {
		fprintf(stderr, "Sorry, your driver has bss_info_version %d "
			"but this program supports only version %d.\n",
			list->version, WL_BSS_INFO_VERSION);
		list->buflen = 0;
		list->count = 0;
	}

	return ret;
}

static int
wl_spect(void *wl, cmd_t *cmd, char **argv)
{
	int ret, spect;
	char *endptr;

	if (!*++argv){
		if ((ret = wl_get(wl, cmd->get, &spect, sizeof(spect))) < 0)
			return ret;
		switch (spect){
		case SPECT_MNGMT_OFF:
			printf("Off\n");
			break;

		case SPECT_MNGMT_LOOSE:
			printf("Loose interpretation - may join non 11h AP.\n");
			break;

		case SPECT_MNGMT_STRICT:
			printf("Strict interpretation - may not join non 11h AP.\n");
			break;

		case SPECT_MNGMT_11D:
			printf("802.11D mode\n");
			break;
		}
		return(0);
	} else {
		spect = strtol(*argv, &endptr, 0);
		if (*endptr != '\0')
			return (-1);

		if (spect < SPECT_MNGMT_OFF || spect > SPECT_MNGMT_11D)
			return (-1);

		return wl_set(wl, cmd->set, &spect, sizeof(spect));
	}
}

static int
wl_status(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	struct ether_addr bssid;
	wlc_ssid_t ssid;
	char ssidbuf[SSID_FMT_BUF_LEN];
	wl_bss_info_t *bi;
	
	if ((ret = wl_get(wl, WLC_GET_BSSID, &bssid, ETHER_ADDR_LEN)) == 0) {
		/* The adapter is associated. */
		*(uint32*)buf = WLC_IOCTL_MAXLEN;
		if ((ret = wl_get(wl, WLC_GET_BSS_INFO, buf, WLC_IOCTL_MAXLEN)) < 0)
			return ret;

		bi = (wl_bss_info_t*)(buf + 4);
		if (bi->version == WL_BSS_INFO_VERSION)
			dump_bss_info(bi);
		else
			fprintf(stderr, "Sorry, your driver has bss_info_version %d "
				"but this program supports only version %d.\n",
				bi->version, WL_BSS_INFO_VERSION);
	} else {
		printf("Not associated. Last associated with ");

		if ((ret = wl_get(wl, WLC_GET_SSID, &ssid, sizeof(wlc_ssid_t))) < 0) {
			printf("\n");
			return ret;
		}
	
		wl_format_ssid(ssidbuf, ssid.SSID, ssid.SSID_len);
		printf("SSID: \"%s\"\n", ssidbuf);
	}

	return 0;
}

static int
wl_deauth_rc(void *wl, cmd_t *cmd, char **argv)
{
	scb_val_t scb_val;
	int ret;

	if (!*++argv) {
		fprintf(stderr, "STA MAC must be specified\n");
		ret = -1;

	} else if (!wl_ether_atoe(*argv, scb_val.ea.octet)) {
		fprintf(stderr,"Malformed STA MAC parameter\n");
		ret = -1;

	} else if (!*++argv) {
		/* No reason code furnished, so driver will use its default */
		ret = wl_set(wl, WLC_SCB_DEAUTHENTICATE, &scb_val.ea,
			     ETHER_ADDR_LEN);

	} else {
		scb_val.val = (uint32)strtoul(*argv, NULL, 0);
		ret = wl_set(wl, cmd->set, &scb_val, sizeof(scb_val));
	}
	return ret;
}

static int
wl_wpa_auth(void *wl, cmd_t *cmd, char **argv)
{
	int bsscfg_idx = 0;
	int consumed;
	uint32 wpa_auth;
	int ret = 0;
#ifndef DSLCPE		
	int i;
	static struct {
		uint32 val;
		char *name;
	} auth_mode[] =
		  /* Keep the numeric values in the staticly initialized
		     help string consistent.  Unfortunately, there isn't
		     an automatic way for that. */
		{{WPA_AUTH_NONE,	"WPA-NONE"},
		 {WPA_AUTH_UNSPECIFIED,	"WPA-802.1x"},
		 {WPA_AUTH_PSK,		"WPA-PSK"},
#ifdef	BCMWPA2
		 {WPA2_AUTH_UNSPECIFIED, "WPA2-802.1x"},
		 {WPA2_AUTH_PSK,	"WPA2-PSK"},
#endif
		 {WPA_AUTH_DISABLED,	"disabled"}};
#endif
	/* skip the command name */
	argv++;
	
	/* parse a bsscfg_idx option if present */
	if ((ret = wl_cfg_option(argv, cmd->name, &bsscfg_idx, &consumed)) != 0)
		return ret;

	argv += consumed;

	if (!*argv) {
		/* no arg, so this is a GET. */
#ifndef DSLCPE		
		char *mode_name;
#endif
		if (!consumed)
			ret = wl_iovar_getint(wl, "wpa_auth", &wpa_auth);
		else
			ret = wl_bssiovar_getint(wl, "wpa_auth", bsscfg_idx, &wpa_auth);
		
		if (ret < 0)
			return ret;
#ifndef DSLCPE
		mode_name = "Illegal";
		for (i = 0; i < (int)ARRAYSIZE(auth_mode); i++) {
			if (auth_mode[i].val == wpa_auth) {
				mode_name = auth_mode[i].name;
				break;
			}
		}
		printf("WPA auth mode %s (%d)\n", mode_name, wpa_auth);
#else
		printf("WPA auth mode %d\n", wpa_auth);
#endif		
	} else {
		/* there's an arg, so this is a SET. */
		ret = 1;
#ifndef DSLCPE		
		if (isdigit((int)*argv[0])) {
			/* param is a number; look for value in the list */
			wpa_auth = (uint32) strtoul(*argv, NULL, 0);
			for (i = 0; i < (int)ARRAYSIZE(auth_mode); i++) {
				if (auth_mode[i].val == wpa_auth) {
					ret = 0;
					break;
				}
			}
		} else {
			/* treat param as string to be matched in list */
			for (i = 0; i < (int)ARRAYSIZE(auth_mode); i++)
				if (!stricmp(auth_mode[i].name, *argv)) {
					wpa_auth = auth_mode[i].val;
					ret = 0;
					break;
				}
		}
#else
		if (isdigit((int)*argv[0])) {
			wpa_auth = (uint32) strtoul(*argv, NULL, 0);
			ret = 0;
		}
#endif		
		if (ret)
			fprintf(stderr, "%s is not a valid WPA auth mode\n", *argv);
		else {
			if (!consumed)
				ret = wl_iovar_setint(wl, "wpa_auth", wpa_auth);
			else
				ret = wl_bssiovar_setint(wl, "wpa_auth", bsscfg_idx, wpa_auth);
		}
	}
	return ret;
}

static int
wl_set_pmk(void *wl, cmd_t *cmd, char **argv)
{
	wsec_pmk_t psk;
	size_t key_len;

	if (!*++argv) {
		return -1;
	}
	key_len = strlen(*argv);
	if (key_len < WSEC_MIN_PSK_LEN || key_len > WSEC_MAX_PSK_LEN) {
		fprintf(stderr, "passphrase must be between %d and %d characters long\n",
		       WSEC_MIN_PSK_LEN, WSEC_MAX_PSK_LEN);
		return -1;
	}
	psk.key_len = (ushort) key_len;
	psk.flags = WSEC_PASSPHRASE;
	memcpy(psk.key, *argv, key_len);
	return wl_set(wl, cmd->set, &psk, sizeof(psk));
}

static int
wl_wsec(void *wl, cmd_t *cmd, char **argv)
{
	uint32 wsec;
	int bsscfg_idx = 0;
	int consumed;
	char *endptr;
	int error;

	argv++;

	/* parse a bsscfg_idx option if present */
	if ((error = wl_cfg_option(argv, "wsec", &bsscfg_idx, &consumed)) != 0)
		return error;

	argv += consumed;

	if (!*argv) {
		/* This is a GET */
		if (consumed == 0)
			error = wl_get(wl, WLC_GET_WSEC, &wsec, sizeof(uint32));
		else
			error = wl_bssiovar_getint(wl, "wsec", bsscfg_idx, &wsec);
		
		if (!error)
			wl_printint(wsec);
	} else {
		/* This is a SET */
		if (!stricmp(*argv, "off"))
			wsec = 0;
		else {
			wsec = strtol(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}
		}

		if (consumed == 0)
			error = wl_set(wl, WLC_SET_WSEC, &wsec, sizeof(uint32));
		else
			error = wl_bssiovar_setint(wl, "wsec", bsscfg_idx, wsec);
	}

	return error;
}

static int
parse_wep(char **argv, wl_wsec_key_t *key)
{
	char hex[] = "XX";
	unsigned char *data = key->data;
	char *keystr = *argv;

	switch (strlen(keystr)) {
	case 5:
	case 13:
	case 16:
		key->len = strlen(keystr);
		strcpy(data, keystr);
		break;
	case 12:
	case 28:
	case 34:
	case 66:
		/* strip leading 0x */
		if (!strnicmp(keystr, "0x", 2))
			keystr += 2;
		else
			return -1;
		/* fall through */
	case 10:
	case 26:
	case 32:
	case 64:
		key->len = strlen(keystr) / 2;
		while (*keystr) {
			strncpy(hex, keystr, 2);
			*data++ = (char) strtoul(hex, NULL, 16);
			keystr += 2;
		}
		break;
	default:
		return -1;
	}

	switch (key->len) {
	case 5:
		key->algo = CRYPTO_ALGO_WEP1;
		break;
	case 13:
		key->algo = CRYPTO_ALGO_WEP128;
		break;
	case 16:
		/* default to AES-CCM*/
		key->algo = CRYPTO_ALGO_AES_CCM;
		break;
	case 32:
		key->algo = CRYPTO_ALGO_TKIP;
		break;
	default:
		return -1;
	}

	/* Set as primary key by default */
	key->flags |= WL_PRIMARY_KEY;
			
	/* Get options */
	while (*++argv) {
		if (!strnicmp("ccm", *argv, 3) && key->len == 16)
			key->algo = CRYPTO_ALGO_AES_CCM;
		else if (!strnicmp("ocb", *argv, 3) && key->len == 16)
			key->algo = CRYPTO_ALGO_AES_OCB_MPDU;
		else if (!strnicmp("notx", *argv, 4))
			key->flags &= ~WL_PRIMARY_KEY;
		else if (!wl_ether_atoe(*argv, key->ea.octet))
			memset(&key->ea, 0, ETHER_ADDR_LEN);
	}
	
	return 0;
}

static int
wl_primary_key(void *wl, cmd_t *cmd, char **argv)
{
	int i, val, ret = 0;

	if (!*++argv) {
		i = 0;
		do {
			val = i;
			if (wl_get(wl, cmd->get, &val, sizeof(val)) < 0) {
				return -1;
			}
			if (val == TRUE) {
				printf("Key %d is primary\n", i);
				return 0;
			}
		} while (++i < DOT11_MAX_DEFAULT_KEYS);
		printf("No primary key set\n");

	} else {
		val = atoi(*argv);
		ret = wl_set(wl, cmd->set, &val, sizeof(val));
	}
	return ret;
}

static int
wl_addwep(void *wl, cmd_t *cmd, char **argv)
{
	wl_wsec_key_t key;
	int bsscfg_idx = 0;
	int consumed;
	int error;
	
	memset(&key, 0, sizeof(key));

	argv++;
	
	/* parse a bsscfg_idx option if present */
	if ((error = wl_cfg_option(argv, "addwep", &bsscfg_idx, &consumed)) != 0)
		return error;

	argv += consumed;

	/* GET operation not allowed */
	if (!*argv)
		return -1;

	key.index = atoi(*argv++);
	
	if (!*argv) {
		fprintf(stderr, "No key specified\n");
		return -1;
	}
	if (parse_wep(argv, &key))
		return -1;

	if (consumed == 0) {
		error = wl_set(wl, cmd->set, &key, sizeof(key));
	} else {
		error = wl_bssiovar_setbuf(wl, "wsec_key", bsscfg_idx,
			&key, sizeof(key), buf, WLC_IOCTL_MAXLEN);
	}

	return error;
}

static int
wl_rmwep(void *wl, cmd_t *cmd, char **argv)
{
	wl_wsec_key_t key;
	int bsscfg_idx = 0;
	int consumed;
	int error;
	
	memset(&key, 0, sizeof(key));

	argv++;
	
	/* parse a bsscfg_idx option if present */
	if ((error = wl_cfg_option(argv, "rmwep", &bsscfg_idx, &consumed)) != 0)
		return error;

	argv += consumed;

	/* GET operation not allowed */
	if (!*argv)
		return -1;

	key.index = atoi(*argv++);

	if (*argv) {
		if (!(wl_ether_atoe(*argv, key.ea.octet)))
			return -1;
	}

	if (consumed == 0) {
		error = wl_set(wl, cmd->set, &key, sizeof(key));
	} else {
		error = wl_var_setbuf(wl, "wsec_key", &key, sizeof(key));
	}

	return error;
}

static struct {
	uint value;
	char *string;
} wsec_test[] = {
	{WSEC_GEN_MIC_ERROR, 	"mic_error"},
	{WSEC_GEN_REPLAY, 	"replay"},
	{0,		NULL}
};


static int
wl_wsec_test(void *wl, cmd_t *cmd, char **argv)
{
	wl_wsec_key_t *key;
	int i, len;
	char *endptr, *buf;
	uint32 val, last_val;

	if (!*++argv)
		goto usage;

	val = strtol(*argv, &endptr, 0);
	if (endptr == *argv) {
		/* the value string was not parsed by strtol */
		for (i = 0; wsec_test[i].value; i++)
			if (stricmp(wsec_test[i].string, *argv) == 0) {
				val = wsec_test[i].value;
				break;
			}
		if (wsec_test[i].value == 0)
			goto usage;
	}
	++argv;

	switch (val) {
	case WSEC_GEN_REPLAY:
	case WSEC_GEN_MIC_ERROR:
		if (!*argv) {
			fprintf(stderr, "insufficient arguments\n");
			return -1;
		}
		len = sizeof(wl_wsec_key_t) + 4;
		buf = malloc(len);
		*(uint32 *)buf = val;
		key = (wl_wsec_key_t *)&buf[4];
		memset(key, 0, sizeof(wl_wsec_key_t));
		/* If it doesn't look like an ether addr, suppose it's a key index */
		if (!(wl_ether_atoe(*argv, key->ea.octet))) {
			memset(&key->ea, 0, ETHER_ADDR_LEN);
			key->index = atoi(*argv);
		}
		break;
	default:
		goto usage;
		break;
	}

	return wl_set(wl, cmd->set, buf, len);

usage:
	fprintf(stderr, "wsec test_type may be a number or name from the following set:");
	last_val = 0xffffffff;
	for (i = 0; (val = wsec_test[i].value); i++) {
		if (val != last_val)
			fprintf(stderr, "\n0x%04x %s", val, wsec_test[i].string);
		else
			fprintf(stderr, ", %s", wsec_test[i].string);
		last_val = val;
	}
	fprintf(stderr, "\n");

	return 0;
}

static int
wl_keys(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	uint i, j;
	union {
		int index;
		wl_wsec_key_t key;
	} u;
	int wep_is_on = 0;
	char *addr;
	int empty_first, empty_last;
	bool empty;
	
	if((ret = wl_iovar_getint(wl, "wsec", &wep_is_on)) < 0)
		fprintf(stderr, "Could not query wsec status.\n");

	empty_first = empty_last = -1;

	for (i = 0; ; i++) {	/* The upper limit is not known here. */
		u.index = i;
		ret = wl_get(wl, cmd->get, &u, sizeof(u));
		empty = (ETHER_ISNULLADDR(&u.key.ea) && u.key.algo == CRYPTO_ALGO_OFF);

		if (empty_first != -1 &&
		    (ret < 0 || !empty)) {
			if (empty_first == empty_last)
				printf("%3d: <empty>\n", empty_first);
			else
				printf("%3d - %3d: <empty>\n", empty_first, empty_last);
			empty_first = empty_last = -1;
		}

		if (ret < 0) {
			/* If at least one beyond the defaults could be read,
			   not knowing the limit must have been the error. */
			return (i >= DOT11_MAX_DEFAULT_KEYS) ? 0 : ret;
		}

		/* Key may not have been set yet */
		if (i < DOT11_MAX_DEFAULT_KEYS)
			addr = "(default)";
		else if (empty) {
			if (empty_first == -1) empty_first = i;
			empty_last = i;
			continue;
		} else
			addr = wl_ether_etoa(u.key.ea.octet);

		printf("%3d: %-17s Key %d: ", i, addr, u.key.index);

		switch(u.key.algo){
		case CRYPTO_ALGO_OFF:
			printf("Off");
			break;
		case CRYPTO_ALGO_WEP1:
			printf("WEP1");
			break;
		case CRYPTO_ALGO_WEP128:
			printf("WEP128");
			break;
		case CRYPTO_ALGO_TKIP:
			printf("TKIP");
			break;
		case CRYPTO_ALGO_AES_CCM:
			printf("AES CCM");
			break;
		case CRYPTO_ALGO_AES_OCB_MSDU:
			printf("AES OCB MSDU");
			break;
		case CRYPTO_ALGO_AES_OCB_MPDU:
			printf("AES OCB MPDU");
			break;
		}
		if (wep_is_on && u.key.flags & WL_PRIMARY_KEY)
			printf("*");
		printf("\t");

		if (u.key.len == 0)
			printf("No key present");
		else {
			if (u.key.flags & WL_SOFT_KEY)
				printf("soft ");
			printf("len %d, data 0x", u.key.len);
			for (j = 0; j < u.key.len; j++)
				printf("%02X", u.key.data[j]);

			for (j = 0; j < u.key.len; j++)
				if (!isprint(u.key.data[j]))
					break;
			if (j == u.key.len)
				printf(" (%.*s)", (int)u.key.len, u.key.data);
			
		}

		printf("\n");
	}
	
	return 0;
}

static int
wl_tsc(void *wl, cmd_t *cmd, char **argv)
{
	union {
		int32 index;
		uint8 tsc[EAPOL_WPA_KEY_RSC_LEN];
	} u;
	uint32 hi, lo;
	int index, ret;

	if (!*++argv)
		return -1;
	index = atoi(*argv);
	if (index < 0) {
		fprintf(stderr, "Key index %d out of range. Should be positive.\n", index);
		return -1;
	}
	u.index = index;
	if ((ret = wl_get(wl, cmd->get, &u, sizeof(u))) < 0)
		return ret;
	lo = u.tsc[0] | (u.tsc[1] << 8) | (u.tsc[2] << 16) | (u.tsc[3] << 24);
	hi = u.tsc[4] | (u.tsc[5] << 8) | (u.tsc[6] << 16) | (u.tsc[7] << 24);

	printf("Key %d TSC: 0x%04x:%08x\n", index, hi, lo);
	return 0;
}

static int
wl_get_current_power(void *wl, cmd_t *cmd, char **argv)
{
	int32 flags;
	int ret, i;
#if defined(DSLCPE) || defined(IL_BIGENDIAN)
	int32 eirp;
#else	
	unsigned char eirp;
#endif	
	tx_power_t power;
	bool valid;
	uint band_list[3];
	bool quiet = FALSE;

	argv++;
	if (*argv && !stricmp(*argv, "-q"))
		quiet = TRUE;

	strcpy(buf, "eirp");

	ret = wl_get(wl, WLC_GET_VAR, &buf, WLC_IOCTL_MAXLEN);
	if (ret) {
		return ret;
	}
#if defined(DSLCPE) || defined(IL_BIGENDIAN)
	eirp = *(int32*)buf;
#else	
	eirp = buf[0];
#endif

	memset(&power, 0, sizeof(power));
 	if ((ret = wl_get(wl, cmd->get, &power, sizeof(power))) < 0)
 		return ret;

	if ((ret = wl_get(wl, WLC_GET_UCFLAGS, &flags, sizeof(flags))) < 0)
 		valid = FALSE;
	else
 		valid = TRUE;

	/* Make the most of the info returned in band_list! 
	 * b/g and a
	 * b/g-uni
	 * a-uni
	 * NOTE: NO a and b/g case ...
	 */
	if ((ret = wl_get(wl, WLC_GET_BANDLIST, band_list, sizeof (band_list))) < 0)
		return (ret);

	/* Print for all */
	if (!quiet){
		printf("User Target:\t\t\t");
		printf(" %6.2f ", (float)power.txpwr_band_max[0]/(float)4);
		printf("\n");

		printf("Regulatory Local Max:\t\t %6.2f dB\n", (float)power.txpwr_local_max/(float)4);
		printf("Regulatory Local Constraint:\t %6.2f dB\n", (float)power.txpwr_local_constraint/(float)4);
		printf("Antgain used in Channel Max:\t  %s\n", eirp & 1 ? 
			"Yes, channel is EIRP" :
			"No, channel is Conducted");
		printf("Hardware Power Control:\t\t  ");
		if (valid)
			printf("%s\n", flags & 0x800000 ? "On" : "Off");
		else
			printf("Unknown - driver must be up\n");
		printf("Regulatory Channel Max:\t\t %6.2f dB\n", (float)power.txpwr_chan_reg_max/(float)4);
		printf("\n");

	}

	/* If B/G is present it's always the lower index */
	if (band_list[1] == WLC_BAND_B) {
		if (!quiet){
			printf("Srom antgain B/G:\t\t %6.2f dB\n", (float)power.txpwr_antgain[0]/(float)4);
			printf("Last B phy CCK est. power:\t %6.2f dB\n", (float)power.txpwr_est_Pout[0]/(float)4);
			printf("Last B phy OFDM est. power:\t %6.2f dB\n", (float)power.txpwr_est_Pout_gofdm/(float)4);

			printf("\n");
			printf("Min of User & Reg Limit:\n");
			printf(" CCK:\t ");
			for (i=0; i < 4; i++)
				printf("%6.2f ", (float)power.txpwr_limit[i]/(float)4);
			printf("\nOFDM:\t ");
			for (; i < NUM_PWRCTRL_RATES; i++)
				printf("%6.2f ", (float)power.txpwr_limit[i]/(float)4);
			printf("\n");

			printf("Srom limit B/G:\n");
			printf(" CCK:\t ");
			for (i=0; i < 4; i++)
				printf("%6.2f ",  (float)power.txpwr_bphy_cck_max[i]/(float)4);
			printf("\nOFDM:\t ");
			for (; i < NUM_PWRCTRL_RATES; i++)
				printf("%6.2f ",  (float)power.txpwr_bphy_cck_max[i]/(float)4);
			printf("\n");

			printf("Last B phy target power:\n");
			printf(" CCK:\t ");
			for (i=0; i < 4; i++)
				printf("%6.2f ", (float)power.txpwr_target[0][i]/(float)4);
			printf("\nOFDM:\t ");
			for (; i < NUM_PWRCTRL_RATES; i++)
				printf("%6.2f ", (float)power.txpwr_target[0][i]/(float)4);
			printf("\n");
		}
	}

	/* A band */
	if (band_list[1] == WLC_BAND_A  ||  (band_list[0] > 1 && band_list[2] == WLC_BAND_A)){
		if (!quiet){
			printf("\n");
			printf("Srom antgain A:\t\t\t %6.2f dB\n", (float)power.txpwr_antgain[1]/(float)4);
		}
		printf("Last A phy est. power:\t\t %6.2f dB\n", (float)power.txpwr_est_Pout[1]/(float)4);
		if (!quiet){
			printf("Srom limit A:\n");
			printf("OFDM:\t ");
			for (i=4; i < NUM_PWRCTRL_RATES; i++)
				printf("%6.2f ", (float)power.txpwr_aphy_max[i]/(float)4);
			printf("\n");

			printf("Last A phy target power:\n");
			printf("OFDM:\t ");
			for (i=4; i < NUM_PWRCTRL_RATES; i++)
				printf("%6.2f ", (float)power.txpwr_target[1][i]/(float)4);
			printf("\n");
		}
	}

	return ret;
}

#ifdef DSLCPE
static int
wl_get_instant_power(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	tx_inst_power_t *power;
	uint band_list[3];

	strcpy(buf, "txinstpwr");
	if ((ret = wl_get(wl, WLC_GET_VAR, &buf, WLC_IOCTL_MAXLEN)) < 0) {
		return ret;
	}

	power = (tx_inst_power_t *)buf;
	/* Make the most of the info returned in band_list!
	 * b/g and a
	 * b/g-uni
	 * a-uni
	 * NOTE: NO a and b/g case ...
	 */
	if ((ret = wl_get(wl, WLC_GET_BANDLIST, band_list, sizeof(band_list))) < 0)
		return (ret);

	/* If B/G is present it's always the lower index */
	if (band_list[1] == WLC_BAND_2G) {
			printf("Last B phy CCK est. power:\t %d.%d dB\n",
			       DIV_QUO(power->txpwr_est_Pout[0], 4),
			       DIV_REM(power->txpwr_est_Pout[0], 4));
			printf("Last B phy OFDM est. power:\t %d.%d dB\n",
			       DIV_QUO(power->txpwr_est_Pout_gofdm, 4),
			       DIV_REM(power->txpwr_est_Pout_gofdm, 4));

			printf("\n");
	}

	/* A band */
	if (band_list[1] == WLC_BAND_5G || (band_list[0] > 1 && band_list[2] == WLC_BAND_5G)) {
		printf("Last A phy est. power:\t\t %d.%d dB\n",
		       DIV_QUO(power->txpwr_est_Pout[1], 4),
		       DIV_REM(power->txpwr_est_Pout[1], 4));
	}

	return ret;
}
#endif

static int
wl_evm(void *wl, cmd_t *cmd, char **argv)
{
	int val[3];

	/* Get channel */
	if (!*++argv){
		fprintf(stderr, "Need to specify at least one parameter\n");
		return -1;
	}

	if (!stricmp(*argv, "off"))
		val[0] = 0;
	else
		val[0] = atoi(*argv);

	/* set optional parameters to default */
	val[1] = 4;	/* rate in 500Kb units */
	val[2] = 0;	/* This is ignored */

	/* Get optional rate and convert to 500Kb units */
	if (*++argv)
		val[1] = rate_string2int(*argv);

	return wl_set(wl, cmd->set, val, sizeof(val));
}

static int
wl_join(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	uint i;
	wlc_ssid_t ssid;
	wl_wsec_key_t key;
	int wsec = 0, auth = 0, infra = -1;
	int wpa_auth = -1;

	/* specified ssid */
	if (!*++argv)
		return -1;

	if (strlen(*argv) > sizeof(ssid.SSID))
		return -1;

	ssid.SSID_len = strlen(*argv);
	strncpy(ssid.SSID, *argv, ssid.SSID_len);
	
	/* get current wsec */
	if (wl_iovar_getint(wl, "wsec", &wsec) < 0)
		wsec = 0;

	while (*++argv) {
		/* specified wep key */
		memset(&key, 0, sizeof(wl_wsec_key_t));
		if (!stricmp(*argv, "wepkey") || !stricmp(*argv, "wep") || !stricmp(*argv, "key")) {
			if (!*++argv)
				return -1;
			key.index = 0;
			if (parse_wep(argv, &key))
				return -1;
			if ((ret = wl_set(wl, WLC_SET_KEY, &key, sizeof(wl_wsec_key_t))) < 0)
				return ret;
			wsec |= WEP_ENABLED;
		}
		/* specified infrastructure mode */
		else if (!stricmp(*argv, "imode") || !stricmp(*argv, "infra") || !stricmp(*argv, "mode")) {
			if (!*++argv)
				return -1;
			if (!stricmp(*argv, "ibss") || !stricmp(*argv, "adhoc") || !stricmp(*argv, "ad-hoc"))
				infra = 0;
			else if (!stricmp(*argv, "bss") || !stricmp(*argv, "managed") || !strnicmp(*argv, "infra", 5))
				infra = 1;
			else
				return -1;
		}
		/* specified authentication mode */
		else if (!stricmp(*argv, "amode") || !strnicmp(*argv, "auth", 4)) {
			if (!*++argv)
				return -1;
			if (!stricmp(*argv, "open"))
				auth = 0;
			else if (!stricmp(*argv, "shared"))
				auth = 1;
			else if (!stricmp(*argv, "wpanone"))
				wpa_auth = WPA_AUTH_NONE;
			else if (!stricmp(*argv, "wpa"))
				wpa_auth = WPA_AUTH_UNSPECIFIED;
			else if (!stricmp(*argv, "wpapsk"))
				 wpa_auth = WPA_AUTH_PSK;
#ifdef BCMWPA2
			else if (!stricmp(*argv, "wpa2"))
				wpa_auth = WPA2_AUTH_UNSPECIFIED;
			else if (!stricmp(*argv, "wpa2psk"))
				wpa_auth = WPA2_AUTH_PSK;
#endif
			else
				return -1;
		}
	}

	if (infra == -1 && wl_get_scan(wl, WLC_SCAN_RESULTS, buf, WLC_IOCTL_MAXLEN) == 0) {
		wl_scan_results_t *list = (wl_scan_results_t*)buf;
		wl_bss_info_t *bi = list->bss_info;
		infra = 1; /* assume infrastructure */

		/* get the default value for infra */
		for (i = 0; i < list->count; i++, bi = (wl_bss_info_t*)((int8*)bi + bi->length)) {
			if (bi->SSID_len == ssid.SSID_len &&
			    !strncmp(bi->SSID, ssid.SSID, ssid.SSID_len)) {
				infra = (bi->capability & DOT11_CAP_ESS)? 1 : 0;
				break;
			}
		}
	}
	
	/* set infrastructure mode */
	if ((ret = wl_set(wl, WLC_SET_INFRA, &infra, sizeof(int))) < 0)
		return ret;
	
	/* set authentication mode */
	if ((ret = wl_set(wl, WLC_SET_AUTH, &auth, sizeof(int))) < 0)
		return ret;

	/* set wsec mode */
	if ((ret = wl_iovar_setint(wl, "wsec", wsec)) < 0)
		return ret;
	
	/* set WPA_auth mode if one was given */
	if (wpa_auth != -1 && (ret = wl_set(wl, WLC_SET_WPA_AUTH, &wpa_auth, sizeof(wpa_auth))) < 0)
		return ret;

	/* set ssid */
	return wl_set(wl, WLC_SET_SSID, &ssid, sizeof(wlc_ssid_t));
}

/* Set or Get the "ssid" iovar, with an optional config index argument:
 *	wl ssid [-C N]|[--cfg=N] ssid
 * 
 * Option:
 *	-C N
 *	--cfg=N
 *	--config=N
 *	--configuration=N
 *		specify the config index N
 * If cfg index not given on a set, the WLC_SET_SSID ioctl will be used
 */
static int
wl_ssid(void *wl, cmd_t *cmd, char **argv)
{
	char ssidbuf[SSID_FMT_BUF_LEN];
	wlc_ssid_t ssid;
	int bsscfg_idx = 0;
	int consumed;
	int error;
	
	argv++;

	/* parse a bsscfg_idx option if present */
	if ((error = wl_cfg_option(argv, "ssid", &bsscfg_idx, &consumed)) != 0)
		return error;

	argv += consumed;

	if (*argv == NULL) {
		if (consumed == 0) {
			/* no config index, use WLC_GET_SSID on the interface */
			error = wl_get(wl, WLC_GET_SSID, &ssid, sizeof(ssid));
		} else {
			/* use "ssid" iovar since a config option was given */
			error = wl_bssiovar_get(wl, "ssid", bsscfg_idx, &ssid, sizeof(ssid));
		}
		if (error < 0)
			return error;

		wl_format_ssid(ssidbuf, ssid.SSID, ssid.SSID_len);
		printf("Current SSID: \"%s\"\n", ssidbuf);
	} else {
		if (strlen(argv[0]) > 32) {
 			fprintf(stderr, "SSID arg \"%s\" must be 32 chars or less\n", argv[0]);
			return -1;
		}
		strcpy(ssid.SSID, argv[0]);
		ssid.SSID_len = strlen(ssid.SSID);

		wl_format_ssid(ssidbuf, ssid.SSID, ssid.SSID_len);
		printf("Setting SSID \"%s\"\n", ssidbuf);

		if (consumed == 0) {
			/* no config index given, use WLC_SET_SSID */
			error = wl_set(wl, WLC_SET_SSID, &ssid, sizeof(wlc_ssid_t));
		} else {
			/* use "ssid" iovar since a config option was given */
			error = wl_bssiovar_set(wl, "ssid", bsscfg_idx, &ssid, sizeof(wlc_ssid_t));
		}
	}
	return error;
}


static int
wl_tssi(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	int val;

	if (cmd->get < 0)
		return -1;			
	if ((ret = wl_get(wl, cmd->get, &val, sizeof(int))) < 0)
		return ret;

	printf("CCK %d OFDM %d\n", (val & 0xff), (val >> 8) & 0xff);
	return 0;
}

/* Quarter dBm units to mW
 * Table starts at QDBM_OFFSET, so the first entry is mW for qdBm=153
 * Table is offset so the last entry is largest mW value that fits in
 * a uint16.
 */

#define QDBM_OFFSET 153
#define QDBM_TABLE_LEN 40

/* Smallest mW value that will round up to the first table entry, QDBM_OFFSET.
 * Value is ( mW(QDBM_OFFSET - 1) + mW(QDBM_OFFSET) ) / 2
 */
#define QDBM_TABLE_LOW_BOUND 6493

/* Largest mW value that will round down to the last table entry,
 * QDBM_OFFSET + QDBM_TABLE_LEN-1.
 * Value is ( mW(QDBM_OFFSET + QDBM_TABLE_LEN - 1) + mW(QDBM_OFFSET + QDBM_TABLE_LEN) ) / 2.
 */
#define QDBM_TABLE_HIGH_BOUND 64938

static const uint16 nqdBm_to_mW_map[QDBM_TABLE_LEN] = {
/* qdBm:        +0		+1		+2		+3		+4		+5		+6		+7	*/
/* 153: */      6683,	7079,	7499,	7943,	8414,	8913,	9441,	10000,
/* 161: */      10593,	11220,	11885,	12589,	13335,	14125,	14962,	15849,
/* 169: */      16788,	17783,	18836,	19953,	21135,	22387,	23714,	25119,
/* 177: */      26607,	28184,	29854,	31623,	33497,	35481,	37584,	39811,
/* 185: */      42170,	44668,	47315,	50119,	53088,	56234,	59566,	63096
};

static uint16
wl_qdbm_to_mw(uint8 qdbm)
{
	uint factor = 1;
	int idx = qdbm - QDBM_OFFSET;
	
	if (idx > QDBM_TABLE_LEN) {
		/* clamp to max uint16 mW value */
		return 0xFFFF;
	}
	
	/* scale the qdBm index up to the range of the table 0-40
	 * where an offset of 40 qdBm equals a factor of 10 mW.
	 */
	while (idx < 0) {
		idx += 40;
		factor *= 10;
	}
	
	/* return the mW value scaled down to the correct factor of 10,
	 * adding in factor/2 to get proper rounding. */
	return ((nqdBm_to_mW_map[idx] + factor/2) / factor);
}

static uint8
wl_mw_to_qdbm(uint16 mw)
{
	uint8 qdbm;
	int offset;
	uint mw_uint = mw;
	uint boundary;
	
	/* handle boundary case */
	if (mw_uint <= 1)
		return 0;
	
	offset = QDBM_OFFSET;
	
	/* move mw into the range of the table */
	while (mw_uint < QDBM_TABLE_LOW_BOUND) {
		mw_uint *= 10;
		offset -= 40;
	}

	for (qdbm = 0; qdbm < QDBM_TABLE_LEN-1; qdbm++) {
		boundary = nqdBm_to_mW_map[qdbm] + (nqdBm_to_mW_map[qdbm+1] - nqdBm_to_mW_map[qdbm])/2;
		if (mw_uint < boundary) break;
	}

	qdbm += (uint8)offset;

	return(qdbm);
}

#define UNIT_MW		1
#define UNIT_QDBM	2
#define UNIT_DBM	3
static int
wl_txpwr1(void *wl, cmd_t *cmd, char **argv)
{
	int ret, val, new_val = 0, unit;
	char buf[20];
	uint32 *int_ptr;
	bool override = FALSE;

	strcpy(buf, "qtxpower");

	if (!*++argv) {
		if (cmd->get < 0)
			return -1;			
		if ((ret = wl_get(wl, WLC_GET_VAR, &buf, strlen(buf) + 5)) < 0)
			return ret;

		int_ptr = (uint32 *)&buf[0];
		val = *int_ptr;
		override = (bool)(val & WL_TXPWR_OVERRIDE);
		val &= ~WL_TXPWR_OVERRIDE;
		printf("TxPower is %d qdbm, %6.2f dbm, %d mW  Override is %s\n",
			val, (float)val/(float)4, wl_qdbm_to_mw((uint8)(MIN(val, 0xff))),
			override ? "On" : "Off");
		return 0;
	} else {
		/* for set */
		unit = UNIT_DBM;	/* default units */

		/* override can be used in combo with any unit */
		if (!strcmp(*argv, "-o")) {
			override = TRUE;
			if (!*++argv)
				return(-1);
		}

		if (!strcmp(*argv, "-d")) {
			unit = UNIT_DBM;
			argv++;
		}
		else if (!strcmp(*argv, "-q")) {
			unit = UNIT_QDBM;
			argv++;
		}
		else if (!strcmp(*argv, "-m")) {
			unit = UNIT_MW;
			argv++;
		}

		/* override can be used in combo with any unit */
		if (!strcmp(*argv, "-o")) {
			override = TRUE;
			argv++;
		}

		if (!*argv)
			return(-1);

		val = atoi(*argv);

		if (val == -1){
			val = 127;		/* Max val of 127 qdbm */
			unit = UNIT_QDBM;
		}

		if (val <= 0){
			return -1;
		}

		switch (unit){
		case UNIT_MW:
			new_val = wl_mw_to_qdbm((uint16)MIN(val, 0xffff));
			break;
		case UNIT_DBM:
			new_val = val * 4;
			break;
		case UNIT_QDBM:
			new_val = val;
			break;
		}
		int_ptr = (uint32 *)&buf[strlen(buf) + 1];
		if (override)
			new_val |= WL_TXPWR_OVERRIDE;
		*int_ptr = new_val;
		if ((ret = wl_set(wl, WLC_SET_VAR, &buf, strlen(buf)+5)) < 0)
			return ret;
		return 0;
	}
}

static int
wl_txpwr(void *wl, cmd_t *cmd, char **argv)
{
	int error;
	int val;
	char *endptr;
	int override;
	char *name = "qtxpower";
	uint16 mw;
	
	if (!*++argv) {
		if ((error = wl_iovar_getint(wl, name, &val) < 0))
			return error;

		/* Report power in mw with WL_TXPWR_OVERRIDE
		 * bit indicating the status 
		 */
		printf("pwr in mw %d\n", val);
		override = val & WL_TXPWR_OVERRIDE;
		val &= ~WL_TXPWR_OVERRIDE;
		printf("pwr in mw after override adj %d\n", val);
		mw = wl_qdbm_to_mw((uint8)(MIN(val, 0xff)));
		printf("%d\n", mw | override);
		return 0;
	} else {
		if (!strcmp(*argv, "-u")) {
			override = 0;
			argv++;
		} else
			override = WL_TXPWR_OVERRIDE;

		val = strtol(*argv, &endptr, 0);
		if (*endptr != '\0') {
			/* not all the value string was parsed by strtol */
			return -1;
		}

		val = wl_mw_to_qdbm((uint16)MIN(val, 0xffff));

		/* wl command input power will override current power set if told so */
		val |= override;

		return wl_iovar_setint(wl, name, val);
	}
}

static int
wl_get_txpwr_limit(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	uint8 val_qdbm;
	uint16 val_mw;
	tx_power_t power;

	ret = wl_get(wl, cmd->get, &power, sizeof(power));
	if (ret < 0)
		return ret;

	val_qdbm = MIN(power.txpwr_band_max[0], power.txpwr_local_max);
	val_mw = wl_qdbm_to_mw((uint8)(MIN(val_qdbm, 0xff)));

	printf("%d mW (%6.2f dBm)\n", val_mw, (float)val_qdbm/(float)4);

	return ret;
}

static int
wl_atten(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	atten_t atten;
	char *endptr;

	if (!*++argv) {
		if (cmd->get < 0)
			return -1;

		if ((ret = wl_get(wl, cmd->get, &atten, sizeof(atten_t))) < 0)
			return ret;

		printf("tx %s bb/radio/ctl1 %d/%d/%d\n",
		       (atten.auto_ctrl ? "auto" : ""),
			atten.bb, atten.radio, atten.txctl1);

		return 0;
	} else {
		if (cmd->set < 0)
			return -1;

#ifdef DSLCPE
		if (!stricmp(*argv, "auto")) {
			atten.auto_ctrl = WL_ATTEN_PCL_ON;
		}
		else if (!stricmp(*argv, "manual")) {
			atten.auto_ctrl = WL_ATTEN_PCL_OFF;
		}
		else {
			atten.auto_ctrl = WL_ATTEN_APP_INPUT_PCL_OFF;
#else
		if (!stricmp(*argv, "auto"))
			atten.auto_ctrl = 1;
		else {
			atten.auto_ctrl = 0;
#endif
			atten.bb = (uint16)strtoul(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}

			if (!*++argv)
				return -1;

			atten.radio = (uint16)strtoul(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}

			if (!*++argv)
				return -1;

			atten.txctl1 = (uint16)strtoul(*argv, &endptr, 0);
			if (*endptr != '\0') {
				/* not all the value string was parsed by strtol */
				return -1;
			}

		}

		return wl_set(wl, cmd->set, &atten, sizeof(atten_t));
	}
}

static int
wl_maclist(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	struct maclist *maclist = (struct maclist *) buf;
	struct ether_addr *ea;
	uint i, max = (WLC_IOCTL_MAXLEN - sizeof(int)) / ETHER_ADDR_LEN;

	if (!*++argv) {
		if (cmd->get < 0)
			return -1;
		maclist->count = max;
		if ((ret = wl_get(wl, cmd->get, maclist, WLC_IOCTL_MAXLEN)) < 0)
			return ret;
		for (i = 0, ea = maclist->ea; i < maclist->count && i < max; i++, ea++)
			printf("%s %s\n", cmd->name, wl_ether_etoa(ea->octet));
		return 0;
	} else {
		if (cmd->set < 0)
			return -1;
		/* Clear list */
		maclist->count = 0;
		if (!stricmp(*argv, "none") || !stricmp(*argv, "clear"))
			return wl_set(wl, cmd->set, maclist, sizeof(int));
		/* Get old list */
		maclist->count = max;
		if ((ret = wl_get(wl, cmd->get, maclist, WLC_IOCTL_MAXLEN)) < 0)
			return ret;
		/* Append to old list */
		ea = &maclist->ea[maclist->count];
		while (*argv && maclist->count < max) {
			if (!wl_ether_atoe(*argv, ea->octet)){
				printf("Problem parsing MAC address \"%s\".\n", *argv);
				return -1;
			}
			maclist->count++;
			ea++;
			argv++;
		}
		/* Set new list */
		return wl_set(wl, cmd->set, maclist, WLC_IOCTL_MAXLEN);
	}
}

static int
wl_maclist_1(void *wl, cmd_t *cmd, char **argv)
{
	struct maclist *maclist;
	struct ether_addr *ea;
	uint i;
	int ret;

	strcpy(buf, argv[0]);

	if ((ret = wl_get(wl, cmd->get, buf, WLC_IOCTL_MAXLEN)) < 0)
		return ret;

	maclist = (struct maclist *)buf;
	
	for (i = 0, ea = maclist->ea; i < maclist->count; i++, ea++)
		printf("%s %s\n", cmd->name, wl_ether_etoa(ea->octet));
	return 0;
}

static int
wl_out(void *wl, cmd_t *cmd, char **argv)
{
	return wl_set(wl, WLC_OUT, NULL, 0);
}

static int
wl_band(void *wl, cmd_t *cmd, char **argv)
{
	uint band;
	int error;

	error = 0;

	argv++;

	if (*argv == NULL) {	/* get current band */
		if ((error = wl_get(wl, WLC_GET_BAND, &band, sizeof (uint))) < 0)
			return (error);

		if (band == WLC_BAND_AUTO)
			printf("auto\n");
		else if (band == WLC_BAND_A)
			printf("a\n");
		else if (band == WLC_BAND_B)
			printf("b\n");
		else
			printf("unrecognized band value %d\n", band);
	} else {		/* set the band */
		if (!stricmp(*argv, "auto"))
			band = WLC_BAND_AUTO;
		else if (!stricmp(*argv, "a"))
			band = WLC_BAND_A;
		else if (!stricmp(*argv, "b"))
			band = WLC_BAND_B;
		else {
			printf("unsupported band: %s\n", *argv);
			return (-1);
		}

		error = wl_set(wl, WLC_SET_BAND, &band, sizeof (uint));

	}
	return (error);
}

static int
wl_bandlist(void *wl, cmd_t *cmd, char **argv)
{
	uint list[3];
	int error;
	uint i;

	error = 0;

	argv++;

	if ((error = wl_get(wl, WLC_GET_BANDLIST, list, sizeof (list))) < 0)
		return (error);

	/* list[0] is count, followed by 'count' bands */

	if (list[0] > 2)
		list[0] = 2;

	for (i = 1; i <= list[0]; i++)
		if (list[i] == WLC_BAND_A)
			printf("a ");
		else if (list[i] == WLC_BAND_B)
			printf("b ");
		else
			printf("? ");
	printf("\n");

	return (0);
}

static int
wl_phylist(void *wl, cmd_t *cmd, char **argv)
{
	char buf[128];
	int error;
	char *cp;

	error = 0;

	argv++;

	if ((error = wl_get(wl, WLC_GET_PHYLIST, buf, sizeof (buf))) < 0)
		return (error);

	cp = buf;

	for (; *cp; cp++)
		printf("%c ", *cp);
	printf("\n");

	return (0);
}

#ifdef linux
#define UPGRADE_BUFSIZE	512
#else
#define UPGRADE_BUFSIZE	1024
#endif

static int
wl_upgrade(void *wl, cmd_t *cmd, char **argv)
{
#if	defined(_CFE_)
	return CFE_ERR_UNSUPPORTED;
#elif	defined(_HNDRTE_)
	return 0;
#else
	FILE *fp;
	int ret = 0;
	struct {
		uint32 offset;
		char buf[UPGRADE_BUFSIZE];
	} block;
	uint len;

	if (!*++argv)
		return -1;

	if (!(fp = fopen(*argv, "rb"))) {
		fprintf(stderr, "%s: No such file or directory\n", *argv);
		return -2;
	}

	printf("Programming %s...", *argv);
	fflush(stdout);
	block.offset = 0;
	while ((len = fread(block.buf, 1, sizeof(block.buf), fp))) {
		if ((ret = wl_set(wl, cmd->set, &block, 4 + len)) < 0)
			break;
		block.offset += len;
		printf(".");
		fflush(stdout);
	}

	if (ferror(fp)) {
		ret = ferror(fp);
		printf("\nerror reading %s\n", *argv);
	} else {
		long status = WLC_UPGRADE_PENDING;
		int retries;

		printf("\nCommitting image to flash...\n");
		while (status == WLC_UPGRADE_PENDING) {
			retries = 10;
retry:
			if ((ret = wl_get(wl, WLC_UPGRADE_STATUS, &status, sizeof (status))) < 0) {
				/* the first attempt to get status will likely fail due to dev reset */
				if (retries--)
					goto retry;
				break;
			}
		}
		if (status == WLC_UPGRADE_SUCCESS)
			printf("\nDone\n\nSuccessfully downloaded %d bytes\n", block.offset);
		else
			fprintf(stderr, "\n*** UPGRADE FAILED! *** (status %ld)\n", status);
	}

	fclose(fp);
	return ret;
#endif
}

#include <bcmnvram.h>

static int
wl_otpw(void *wl, cmd_t *cmd, char **argv)
{
#if	defined(_CFE_)
	return CFE_ERR_UNSUPPORTED;
#elif	defined(_HNDRTE_)
	return 0;
#else
	FILE *fp;
	int ret = 0;
	struct nvram_header *nvr;
	char *p, *msg, buf[1024 - 128];
	uint len;

	if (!*++argv)
		return -1;

	if (!(fp = fopen(*argv, "rb"))) {
		fprintf(stderr, "%s: No such file or directory\n", *argv);
		return -2;
	}

	len = fread(buf, 1, sizeof(buf) - 1, fp);
	if ((ret = ferror(fp))) {
		printf("\nerror %d reading %s\n", ret, *argv);
		ret = -3;
		goto out;
	}
	if (!feof(fp)) {
		printf("\nFile %s too large\n", *argv);
		ret = -4;
		goto out;
	}

	/* Got the bits, do they look like the output of nvserial? */
	nvr = (struct nvram_header *)buf;
	if (nvr->magic == NVRAM_MAGIC) {
		len  = nvr->len - sizeof(struct nvram_header);
		p = (char *)(nvr + 1);
		msg = "nvserial";
	} else {
		p = buf;
		msg = "raw";
	}

	printf("Writing %d bytes from %s file %s to otp ...\n", len, msg, *argv);
	fflush(stdout);

	if (len & 1) {
		buf[len++] = '\0';
	}
	if ((ret = wl_set(wl, cmd->set, p, len)) < 0) {
		printf("\nError %d writing %s to otp\n", ret, *argv);
	}

out:
	fclose(fp);
	return ret;
#endif
}

static int
wl_get_pktcnt(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	get_pktcnt_t pktcnt;

	memset(&pktcnt, 0, sizeof(pktcnt));
	if ((ret = wl_get(wl, cmd->get, &pktcnt, sizeof(pktcnt))) < 0)
		return ret;

	printf("Receive: good packet %d, bad packet %d\n", pktcnt.rx_good_pkt, pktcnt.rx_bad_pkt);
	printf("Transmit: good packet %d, bad packet %d\n", pktcnt.tx_good_pkt, pktcnt.tx_bad_pkt);
    
	return ret;
}


static int
wl_interfere(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	int val;
	char *endptr;
	int mode;

	if (!*++argv) {
		if (cmd->get < 0)
			return -1;			
		if ((ret = wl_get(wl, cmd->get, &mode, sizeof(mode))) < 0)
			return ret;
		switch(mode & 0x7f){
		case INTERFERE_NONE:
			printf("All interference mitigation is disabled.\n");
			break;
		case NON_WLAN:
			printf("Non-wireless LAN Interference mitigation is enabled.\n");
			break;
		case WLAN_MANUAL:
			printf("Wireless LAN Interference mitigation is enabled.\n");
			break;
		case WLAN_AUTO:
			printf("Auto Wireless LAN Interference mitigation is enabled and ");
			if (mode & AUTO_ACTIVE)
				printf("active\n");
			else
				printf("not active\n");

			break;
		}
		return 0;
	} else {
		mode = INTERFERE_NONE;
		val = strtol(*argv, &endptr, 0);
		if (*endptr != '\0')
			return(-1);

		switch(val){
			case 0:
				mode = INTERFERE_NONE;
				break;
			case 1:
				mode = NON_WLAN;
				break;
			case 2:
				mode = WLAN_MANUAL;
				break;
			case 3:
				mode = WLAN_AUTO;
				break;
			default:
				return(-1);
		}

		return wl_set(wl, cmd->set, &mode, sizeof(mode));
	}
}

static cntry_name_t *
wlc_cntry_name_to_country(char *long_name)
{
	cntry_name_t *cntry;
	for (cntry = cntry_names; cntry->name && stricmp(long_name, cntry->name);cntry++);
	return (!cntry->name ? NULL : cntry);
}

static cntry_name_t *
wlc_cntry_abbrev_to_country(char *abbrev)
{
	cntry_name_t *cntry;
	if (!*abbrev || strlen(abbrev) > 3 || strlen(abbrev) < 2)
		return(NULL);
	for (cntry = cntry_names; cntry->name && strnicmp(abbrev, cntry->abbrev, strlen(abbrev));cntry++);
	return (!cntry->name ? NULL : cntry);
}

int
wl_country(void *wl, cmd_t *cmd, char **argv)
{
	cntry_name_t *cntry;
	char* name = "<unknown>";
	char* abbrev;
	int ret;
	uint i;

	buf[0] = 0;
	if (!*++argv) {
		if ((ret = wl_get(wl, cmd->get, &buf, sizeof (buf))))
			return ret;
		if (strlen(buf) == 0) {
			printf("No country set\n");
			return 0;
		}
		if ((cntry = wlc_cntry_abbrev_to_country(buf)) != NULL){
			name = cntry->name;
		}
		printf("%s/%s\n", buf, name);
		return 0;
	}

	if (!stricmp(*argv, "list")){
		wl_country_list_t *cl = (wl_country_list_t *)buf;
		cntry_name_t *cntry;

		cl->buflen = sizeof(buf);
		cl->count = 0;

		/* band may follow */
		if (*++argv) {
			cl->band_set = TRUE;
			if (!stricmp(*argv, "a"))
				cl->band = WLC_BAND_A;
			else if (!stricmp(*argv, "b") || !stricmp(*argv, "g"))
				cl->band = WLC_BAND_B;
			else {
				printf("unsupported band: %s\n", *argv);
				return -1;
			}
		} else {
			cl->band_set = FALSE;
		}
	
		ret = wl_get(wl, WLC_GET_COUNTRY_LIST, buf, cl->buflen);
		if (ret < 0)
			return ret;
		
		printf("Supported countries: abbreviation and long name\n");
		for (i = 0; i < cl->count; i++) {
			char *abbrev = &cl->country_abbrev[i*WLC_CNTRY_BUF_SZ];
			cntry = wlc_cntry_abbrev_to_country(abbrev);
			printf("%s\t%s\n", abbrev, cntry ? cntry->name : abbrev);
		}
		return ret;
	}

	/* if matching long name, find abbrev name, otherwise assume it's abbrev name */
	if ((cntry = wlc_cntry_name_to_country(*argv)) != NULL)
		abbrev = cntry->abbrev;
	else
		abbrev = *argv;
	
	if (strlen(abbrev) > 3) {
		printf("Parameter \"%s\" was unrecognized as a country name, and is\n"
		       "too long for a max 3 character country abbreviation.\n",
		       abbrev);
		return 1;
	}

	printf("Setting country code using abbreviation: \"%s\"\n", abbrev);
	strcpy(buf, abbrev);

	if (wl_set(wl, cmd->set, buf, strlen(buf) + 1) < 0) {
		printf("Setting country \"%s\" failed !\n", abbrev);
		return 1;
	}
	return(0);
}

#define ACI_SPIN	"spin"
#define ACI_ENTER	"enter"
#define ACI_EXIT	"exit"
#define ACI_GLITCH	"glitch"

#ifdef DSLCPE
static int
wl_aci_args(void *wl, cmd_t *cmd, char **argv)
{
	wl_aci_args_t aci_args;
	int val;
	char *endptr;
	int *data;

	memset(&aci_args, 0, sizeof(wl_aci_args_t));
	if (!*++argv) {
		if (wl_get(wl, cmd->get, &aci_args, sizeof(wl_aci_args_t)) < 0){
			printf("Cannot get ACI args\n");
			printf("aci_args.enter_aci_thresh = %d\n", aci_args.enter_aci_thresh);
			return 1;
		}
		printf("Glitch count to enter ACI scan mode: %d\n", aci_args.enter_aci_thresh);
		printf("Glitch count to exit ACI mode: %d\n", aci_args.exit_aci_thresh);
		printf("Usecs to spin between rssi samples : %d\n", aci_args.usec_spin);
		printf("Interval (in seconds) for ACI scanning in presence of high glitch count: %d\n", 
				aci_args.glitch_delay);
	} else {
		if (wl_get(wl, cmd->get, &aci_args, sizeof(wl_aci_args_t)) < 0){
			printf("Cannot get ACI args for setting\n");
			return (-1);
		}

		while (*argv){
			/* Parse keyword */
			data = (void *)NULL;
			if (!strncmp(*argv, ACI_ENTER, strlen("en")))
				data = &aci_args.enter_aci_thresh;
			if (!strncmp(*argv, ACI_EXIT, strlen("ex")))
				data = &aci_args.exit_aci_thresh;
			if (!strncmp(*argv, ACI_SPIN, strlen(*argv)))
				data = &aci_args.usec_spin;
			if (!strncmp(*argv, ACI_GLITCH, strlen(*argv)))
				data = &aci_args.glitch_delay;
			if (!data){
				printf("Bad parameter specification\n");
				return(-1);
			}
			/* Parse value */
			if (*(argv+1)){
				val = strtol(*(argv+1), &endptr, 0);
				if (*endptr != '\0') {
					/* not all the value string was parsed by strtol */
					return (-1);
				}
				if (val < 0){
					printf("Invalid value for %s\n", *argv);
					return(-1);
				}
				*data = val;
			} else {
				printf("Need to specify a value\n");
				return(-1);
			}
			argv += 2;
		}

		if (wl_set(wl, cmd->set, &aci_args, sizeof(wl_aci_args_t)) < 0){
			printf("Cannot set ACI args\n");
			return (-1);
		}
	}
	return(0);
}
#endif

/*
 *  RADAR detection parameter control
 */ 
static int
wl_radar_args(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	wl_radar_args_t radar_args;
	wl_radar_args_t *radar_args_ptr;

	if (!*++argv) {
		/* Get */
		if ((ret = wl_var_getbuf(wl, cmd->name, NULL, 0, (void**)&radar_args_ptr)) < 0)
			return ret;

		printf("npulses %d ncontig %d min_pw %d max_pw %d, thresh0 0x%x, thresh1 0x%x\n", 
			radar_args_ptr->npulses, radar_args_ptr->ncontig, radar_args_ptr->min_pw, 
			radar_args_ptr->max_pw, radar_args_ptr->thresh0, radar_args_ptr->thresh1);
	} else {
		/* Set */
		char *endptr;

		/* Order is npulses,ncontig,min_pw,max_pw, thresh0, thresh1 */
		radar_args.npulses = atoi(*argv);
		
		if (!*++argv) {
			return -1;
		}		
		radar_args.ncontig = atoi(*argv);
		
		if (!*++argv) {
			return -1;
		}		
		radar_args.min_pw = atoi(*argv);
	
		if (!*++argv) {
			return -1;
		}		
		radar_args.max_pw = atoi(*argv);

		if (!*++argv) {
			return -1;
		}		
		radar_args.thresh0 = (uint16)strtol(*argv, &endptr, 0);
		if (*endptr != '\0') {
			/* not all the value string was parsed by strtol */
			return -1;
		}

		if (!*++argv) {
			return -1;
		}		

		radar_args.thresh1 = (uint16)strtol(*argv, &endptr, 0);
		if (*endptr != '\0') {
			/* not all the value string was parsed by strtol */
			return -1;
		}

		return wl_var_setbuf(wl, cmd->name, &radar_args, sizeof(wl_radar_args_t));
	}
	return ret;
}

static int
wl_dfs_status(void *wl, cmd_t *cmd, char **argv)
{
	int ret;
	wl_dfs_status_t *dfs_status_ptr;
	char *dfs_cacstate_str[WL_DFS_CACSTATES] = {
		"IDLE", 
		"PRE-ISM Channel Availability Check(CAC)", 
		"In-Service Monitoring(ISM)", 
		"Channel Switching Announcement(CSA)", 
		"POST-ISM Channel Availability Check", 
		"PRE-ISM Ouf Of Channels(OOC)", 
		"POST-ISM Out Of Channels(OOC)"
	};
	
	if ((ret = wl_var_getbuf(wl, cmd->name, NULL, 0, (void**)&dfs_status_ptr)) < 0)
		return ret;

	if (dfs_status_ptr->state >= WL_DFS_CACSTATES) {
		printf("Unknown dfs state %d.\n", dfs_status_ptr->state);
		return -1;
	}
	
	printf("state %s time elapsed %dms radar channel cleared by dfs ", 
		dfs_cacstate_str[dfs_status_ptr->state], dfs_status_ptr->duration);
		
	if (dfs_status_ptr->channel_cleared) {
		printf("%d\n", dfs_status_ptr->channel_cleared);
	}
	else {
		printf("none\n");
	}

	return ret;
}

static int
wl_wds_wpa_role_old(void *wl, cmd_t *cmd, char **argv)
{
	uint remote[2];
	uint *sup = remote;
	if (wl_get(wl, WLC_WDS_GET_REMOTE_HWADDR, remote, sizeof(remote)) < 0){
		printf("Unable to get remote endpoint's hwaddr\n");
		return -1;
	}
	if (wl_get(wl, cmd->get, remote, sizeof(remote)) < 0){
		printf("Unable to get local endpoint's WPA role\n");
		return -1;
	}
	printf("Local endpoing's WPA role: %s\n", *sup ? "supplicant" : "authenticator");
	return 0;
}
static int
wl_wds_wpa_role(void *wl, cmd_t *cmd, char **argv)
{
	uchar var[256];
	uchar *mac;
	uchar *sup;
	int len;
	if (strlen("wds_wpa_role") + 1 + ETHER_ADDR_LEN + 1 > sizeof(var))
		return -1;
	/* build var required by WLC_GET|SET_VAR */
	len = sprintf(var, "%s", "wds_wpa_role") + 1;
	mac = var + len;
	if (wl_get(wl, WLC_WDS_GET_REMOTE_HWADDR, mac, ETHER_ADDR_LEN) < 0){
		printf("Unable to get remote endpoint's hwaddr\n");
		return -1;
	}
	len += ETHER_ADDR_LEN + 1;
	if (argv[1]) {
		sup = mac + ETHER_ADDR_LEN;
		switch ((*sup = atoi(argv[1]))) {
		case WL_WDS_WPA_ROLE_AUTH:
		case WL_WDS_WPA_ROLE_SUP:
		case WL_WDS_WPA_ROLE_AUTO:
			if (wl_set(wl, cmd->set, var, len) < 0)
				printf("Unable to set local endpoint's WPA role\n");
			break;
		default:
			printf("Invalid WPA role %s. %u:authenticator, %u:supplicant, %u:auto\n",
				argv[1], WL_WDS_WPA_ROLE_AUTH, WL_WDS_WPA_ROLE_SUP, WL_WDS_WPA_ROLE_AUTO);
			break;
		}
	}
	else if (wl_get(wl, cmd->get, var, len) < 0){
		printf("Unable to get local endpoint's WPA role\n");
		return -1;
	}
	else {
		sup = var;
		printf("Local endpoint's WPA role: %s\n", *sup ? "supplicant" : "authenticator");
	}
	return 0;
}

static int
wl_measure_req(void *wl, cmd_t *cmd, char **argv)
{
	uint32 val;
	struct ether_addr ea;

	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}

	if (!stricmp(*argv, "tpc"))
		val = WLC_MEASURE_TPC;
	else if (!stricmp(*argv, "basic"))
		val = WLC_MEASURE_CHANNEL_BASIC;
	else if (!stricmp(*argv, "cca"))
		val = WLC_MEASURE_CHANNEL_CCA;
	else if (!stricmp(*argv, "rpi"))
		val = WLC_MEASURE_CHANNEL_RPI;
	else {
		printf("error: unknown measurement type %s\n", *argv);
		return -1;
	}
	argv++;
	
	if (!*argv) {
		printf("error: missing target address\n");
		return -1;
	}
		
	if (!wl_ether_atoe(*argv, ea.octet)) {
		printf("error: could not parse MAC address %s\n", *argv);
		return -1;
	}

	memcpy(&buf[0], &val, sizeof(uint32));
	memcpy(&buf[4], ea.octet, ETHER_ADDR_LEN);

	return wl_set(wl, cmd->set, buf, sizeof(uint32) + ETHER_ADDR_LEN);
}

static int
wl_send_quiet(void *wl, cmd_t *cmd, char **argv)
{
	dot11_quiet_t quiet;

	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}
	/* Order is count, duration, offset */
	quiet.count = atoi(*argv);
	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}
	quiet.duration = atoi(*argv);
	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}
	quiet.offset = atoi(*argv);
	quiet.period = 0;

	return(wl_set(wl, cmd->set, &quiet, sizeof(quiet)));
}

static int
wl_send_csa(void *wl, cmd_t *cmd, char **argv)
{
	dot11_channel_switch_t csa;

	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}
	/* Order is mode, count channel*/
	csa.mode = atoi(*argv) ? 1 : 0;
	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}
	csa.count = atoi(*argv);
	if (!*++argv) {
		printf("error: missing arguments\n");
		return -1;
	}
	csa.channel = atoi(*argv);

	return(wl_set(wl, cmd->set, &csa, sizeof(csa)));
}

static int
wl_var_setint(void *wl, cmd_t *cmd, char **argv)
{
	int32 val;
	int len;
	char *varname;
	char *endptr;
	char *p;
	
	if (!*argv) {
		printf("set: missing arguments\n");
		return -1;
	}
	
	varname = *argv++;

	if (!*argv) {
		printf("set: missing value argument for set of \"%s\"\n", varname);
		return -1;
	}
	
	val = strtol(*argv, &endptr, 0);
	if (*endptr != '\0') {
		/* not all the value string was parsed by strtol */
		printf("set: error parsing value \"%s\" as an integer for set of \"%s\"\n", *argv, varname);
		return -1;
	}

	strcpy(buf, varname);
	p = buf;
	while (*p != '\0') {
		*p = tolower(*p);
		p++;
	}

	/* skip the null */
	p++;
	
	memcpy(p, &val, sizeof (uint));
	len = (p - buf) +  sizeof (uint);

	return (wl_set(wl, WLC_SET_VAR, &buf, len));
}

static int
wl_var_get(void *wl, cmd_t *cmd, char **argv)
{
	char *varname;
	char *p;
	
	if (!*argv) {
		printf("get: missing arguments\n");
		return -1;
	}
	
	varname = *argv++;

	if (*argv) {
		printf("get: error, extra arg \"%s\"\n", *argv);
		return -1;
	}
	
	strcpy(buf, varname);
	p = buf;
	while (*p != '\0') {
		*p = tolower(*p);
		p++;
	}
	return (wl_get(wl, WLC_GET_VAR, &buf, WLC_IOCTL_MAXLEN));
}

static int
wl_var_getint(void *wl, cmd_t *cmd, char **argv)
{
	int err;
	int32 val;

	if ((err = wl_var_get(wl, cmd, argv)))
		return (err);

	val = *(int32*)buf;
	
	if (val < 10)
		printf("%d\n", val);
	else
		printf("%d (0x%x)\n", val, val);

	return (0);
}

static int
wl_var_getandprintstr(void *wl, cmd_t *cmd, char **argv)
{
	int err;

	if ((err = wl_var_get(wl, cmd, argv)))
		return (err);

	printf("%s\n", buf);
	return (0);
}

void 
wl_printlasterror(void *wl)
{
	char *cmd[2] = {"bcmerrorstr"};

	fprintf(stderr, "%s: ", av0);
	if (wl_var_getandprintstr(wl, NULL, cmd) != 0) {
		fprintf(stderr, "\nError getting the last error\n");
	}
	return;
}

/* just issue a wl_var_setint() or a wl_var_getint() if there is a 2nd arg */
static int
wl_varint(void *wl, cmd_t *cmd, char *argv[])
{
	if (argv[1])
		return (wl_var_setint(wl, cmd, argv));
	else
		return (wl_var_getint(wl, cmd, argv));
}

static int
wl_var_getbuf(void *wl, char *iovar, void *param, int param_len, void **bufptr)
{
	int len;
	
	memset(buf, 0, WLC_IOCTL_MAXLEN);
	strcpy(buf, iovar);

	/* include the null */
	len = strlen(iovar) + 1;
	
	if (param_len)
		memcpy(&buf[len], param, param_len);

	*bufptr = buf;
		
	return wl_get(wl, WLC_GET_VAR, &buf, WLC_IOCTL_MAXLEN);
}

static int
wl_var_setbuf(void *wl, char *iovar, void *param, int param_len)
{
	int len;

	memset(buf, 0, WLC_IOCTL_MAXLEN);
	strcpy(buf, iovar);

	/* include the null */
	len = strlen(iovar) + 1;
	
	if (param_len)
		memcpy(&buf[len], param, param_len);

	len += param_len;

	return wl_set(wl, WLC_SET_VAR, &buf, len);
}

static int
wl_var_void(void *wl, cmd_t *cmd, char **argv)
{
	if (cmd->set < 0)
		return -1;

	return wl_var_setbuf (wl, cmd->name, NULL, 0);
}	

/* 
 * format an iovar buffer
 */
static uint
wl_iovar_mkbuf(char *name, char *data, uint datalen, char *buf, uint buflen, int *perr)
{
	uint len;

	len = strlen(name) + 1;
	
	/* check for overflow */
	if ((len + datalen) > buflen) {
		*perr = BCME_BUFTOOSHORT;
		return 0;
	}
	
	strcpy(buf, name);

	/* append data onto the end of the name string */
	if (datalen > 0)
		memcpy(&buf[len], data, datalen);

	len += datalen;
	
	*perr = 0;
	return len;
}

static int
wl_iovar_getint(void *wl, char *name, int *var)
{
	int len;
	char ibuf[WLC_IOCTL_SMLEN];
	int error;
	
	len = wl_iovar_mkbuf(name, NULL, 0, ibuf, sizeof(ibuf), &error);
	if (error) 
		return error;
		
	if ((error = wl_get(wl, WLC_GET_VAR, &ibuf, sizeof(ibuf))) < 0)
		return error;

	memcpy(var, ibuf, sizeof(int));

	return 0; 
}

static int
wl_iovar_setint(void *wl, char *name, int var)
{
	int len;
	char ibuf[WLC_IOCTL_SMLEN];
	int error;
		
	len = wl_iovar_mkbuf(name, (char *)&var, sizeof(var), ibuf, sizeof(buf), &error);
	if (error) 
		return error;
		
	if ((error = wl_set(wl, WLC_SET_VAR, &ibuf, len)) < 0)
		return error;
	
	return 0;
}

/* 
 * format a bsscfg indexed iovar buffer
 */
static int
wl_bssiovar_mkbuf(char *iovar, int bssidx, void *param, int paramlen, void *bufptr, int buflen, int *perr)
{
	char *prefix = "bsscfg:";
	int8* p;
	uint prefixlen;
	uint namelen;
	uint iolen;

	prefixlen = strlen(prefix);	/* length of bsscfg prefix */
	namelen = strlen(iovar) + 1;	/* length of iovar name + null */
	iolen = prefixlen + namelen + sizeof(int) + paramlen;

	/* check for overflow */
	if (buflen < 0 || iolen > (uint)buflen) {
		*perr = BCME_BUFTOOSHORT;
		return 0;
	}

	p = (int8*)bufptr;

	/* copy prefix, no null */
	memcpy(p, prefix, prefixlen);	
	p += prefixlen;

	/* copy iovar name including null */
	memcpy(p, iovar, namelen);	
	p += namelen;

	/* bss config index as first param */
	memcpy(p, &bssidx, sizeof(int32));
	p += sizeof(int32);
	
	/* parameter buffer follows */
	if (paramlen)
		memcpy(p, param, paramlen);

	*perr = 0;
	return iolen;
}

/* 
 * set named & bss indexed driver iovar providing both parameter and i/o buffers
 */
static int
wl_bssiovar_setbuf(void* wl, char *iovar, int bssidx, void *param, int paramlen, void *bufptr, int buflen)
{
	int err;
	int iolen;
	
	iolen = wl_bssiovar_mkbuf(iovar, bssidx, param, paramlen, bufptr, buflen, &err);
	if (err)
		return err;
	
	return wl_set(wl, WLC_SET_VAR, bufptr, iolen);
}

/* 
 * get named & bss indexed driver iovar providing both parameter and i/o buffers
 */
static int
wl_bssiovar_getbuf(void* wl, char *iovar, int bssidx, void *param, int paramlen, void *bufptr, int buflen)
{
	int err;
	int iolen;
	
	iolen = wl_bssiovar_mkbuf(iovar, bssidx, param, paramlen, bufptr, buflen, &err);
	if (err)
		return err;
	
	return wl_get(wl, WLC_GET_VAR, bufptr, buflen);
}

/* 
 * get named & bss indexed driver variable to buffer value
 */
static int
wl_bssiovar_get(void *wl, char *iovar, int bssidx, void *outbuf, int len)
{
	char smbuf[WLC_IOCTL_SMLEN];
	int err;

	/* use the return buffer if it is bigger than what we have on the stack */
	if (len > (int)sizeof(smbuf)) {
		err = wl_bssiovar_getbuf(wl, iovar, bssidx, NULL, 0, outbuf, len);
	} else {
		memset(smbuf, 0, sizeof(smbuf));
		err = wl_bssiovar_getbuf(wl, iovar, bssidx, NULL, 0, smbuf, sizeof(smbuf));
		if (err == 0)
			memcpy(outbuf, smbuf, len);
	}

	return err;
}

/* 
 * set named & bss indexed driver variable to buffer value
 */
static int
wl_bssiovar_set(void *wl, char *iovar, int bssidx, void *param, int paramlen)
{
	char smbuf[WLC_IOCTL_SMLEN];

	memset(smbuf, 0, sizeof(smbuf));

	return wl_bssiovar_setbuf(wl, iovar, bssidx, param, paramlen, smbuf, sizeof(smbuf));
}

/* 
 * get named & bsscfg indexed driver variable as an int value
 */
static int
wl_bssiovar_getint(void *wl, char *iovar, int bssidx, int *pval)
{
	return wl_bssiovar_get(wl, iovar, bssidx, pval, sizeof(int));
}

/* 
 * set named & bsscfg indexed driver variable to int value
 */
static int
wl_bssiovar_setint(void *wl, char *iovar, int bssidx, int val)
{
	return wl_bssiovar_set(wl, iovar, bssidx, &val, sizeof(int));
}

static int
wl_nvdump(void *wl, cmd_t *cmd, char **argv)
{
	int err;
	char *p;
	
	/* skip the "nvdump" command name if present */
	if (!strcmp("nvdump", *argv))
		argv++;

	if (*argv) {
		printf("nvdump error: extra arg \"%s\"\n", *argv);
		return -1;
	}
	
	err = wl_get(wl, WLC_NVRAM_DUMP, &buf, WLC_IOCTL_MAXLEN);
	if (err) {
		return err;
	}

	p = buf;
	while (*p != '\0') {
		printf("%s\n", p);
		p += strlen(p) + 1;
	}

	return err;
}

static int
wl_nvget(void *wl, cmd_t *cmd, char **argv)
{
	int err;
	char *varname;
	
	/* skip the "nvget" command name if present */
	if (!strcmp("nvget", *argv))
		argv++;

	if (!*argv) {
		printf("nvget: missing arguments\n");
		return -1;
	}
	
	varname = *argv++;

	if (*argv) {
		printf("nvget error: extra arg \"%s\"\n", *argv);
		return -1;
	}
	
	strcpy(buf, varname);
	
	err = wl_get(wl, WLC_NVRAM_GET, &buf, WLC_IOCTL_MAXLEN);
	if (err) {
		return err;
	}

	printf("%s\n", buf);

	return err;
}

static int
wl_nvset(void *wl, cmd_t *cmd, char **argv)
{
	char *varname;

	/* skip the "nvset" command name if present */
	if (!strcmp("nvset", *argv))
		argv++;

	if (!*argv) {
		printf("nvset: missing arguments\n");
		return -1;
	}

	varname = *argv++;

	if (*argv) {
		fprintf(stderr, "nvset error: extra arg \"%s\"; format is name=value (no spaces around '=')\n", *argv);
		return -1;
	}

	if (!strchr(varname, '=')) {
		fprintf(stderr, "nvset error: no '=' in \"%s\", format is name=value (no spaces around '=')\n", *argv);
		return -1;
	}

	strcpy(buf, varname);

	return (wl_set(wl, WLC_NVRAM_SET, &buf, strlen(buf) + 1));
}

static int
wl_chan_info(void *wl, cmd_t *cmd, char **argv)
{
	uint bitmap;
	uint channel;
	int buflen, err, first, last, minutes;
	char *param;
	bool all;

	if (!*++argv){
		first = 0;
		last = 255;
		all = TRUE;
	} else {
		last = first = atoi(*argv);
		if (last <= 0){
			printf(" Usage: %s [channel | All ]\n", cmd->name);
			return -1;
		}
		all = FALSE;
	}

	for (; first <= last; first++){
		channel = first & WL_CHANSPEC_CHAN_MASK;
		if (channel < 14)
			channel |= WL_CHANSPEC_BAND_B;
		else
			channel |= WL_CHANSPEC_BAND_A;

		strcpy(buf, "per_chan_info");
		buflen = strlen(buf) + 1;
		param = (char *)(buf + buflen);
		memcpy(param, (char*)&channel, sizeof(channel)); 

		if ((err = wl_get(wl, cmd->get, buf, sizeof(buf)))) 
			return err;

		bitmap = *(uint *)buf;
		minutes = (bitmap >> 24) & 0xff;

		if (!(bitmap & WL_CHAN_VALID_HW)){
			if (!all)
				printf("Invalid Channel\n");
			continue;
		}

		if (!(bitmap & WL_CHAN_VALID_SW)){
			if (!all)
				printf("Not supported in current locale\n");
			continue;
		}

		printf("Channel %d\t", channel & WL_CHANSPEC_CHAN_MASK);

		if (bitmap & WL_CHAN_BAND_A)
			printf("A Band");
		else
			printf("B Band");
			
		if (bitmap & WL_CHAN_RADAR){
			printf(", RADAR Sensitive");
		}
		if (bitmap & WL_CHAN_RADAR_PASSIVE){
			printf(", Passive");
		}
		if (bitmap & WL_CHAN_INACTIVE){
			printf(", Temporarily Out of Service for %d minutes", minutes);
		}
		printf("\n");
	}

	return(0);
}

static int
wl_sta_info(void *wl, cmd_t *cmd, char **argv)
{
	sta_info_t *sta;
	struct ether_addr ea;
	char *param;
	int buflen, err;

	/* convert the ea string into an ea struct */
	if (!*++argv || !wl_ether_atoe(*argv, (char*)&ea)) {
		printf(" ERROR: no valid ether addr provided\n");
		return -1;
	} 

	strcpy(buf, "sta_info");
	buflen = strlen(buf) + 1;
	param = (char *)(buf + buflen);
	memcpy(param, (char*)&ea, ETHER_ADDR_LEN); 

	if ((err = wl_get(wl, cmd->get, buf, sizeof(buf)))) 
		return err;

	/* display the sta info */
	sta = (sta_info_t *)buf;
	printf(" STA %s:\n", *argv);
	printf("\t rateset ");
	dump_rateset(sta->rateset.rates, sta->rateset.count);
	printf("\n\t idle %d seconds\n", sta->idle);
	printf("\t in network %d seconds\n", sta->in);
	printf("\t state:%s%s%s\n",
	       (sta->flags & WL_STA_AUTHE) ? " AUTHENTICATED" : "",
	       (sta->flags & WL_STA_ASSOC) ? " ASSOCIATED" : "",
	       (sta->flags & WL_STA_AUTHO) ? " AUTHORIZED" : "");
	printf("\t flags 0x%x:%s%s%s%s%s%s%s%s\n", sta->flags,
	       (sta->flags & WL_STA_BRCM) ? " BRCM" : "",
	       (sta->flags & WL_STA_ABCAP) ? " ABCAP" : "",
	       (sta->flags & WL_STA_WME) ? " WME" : "",
	       (sta->flags & WL_STA_PS) ? " PS" : "",
	       (sta->flags & WL_STA_APSD_BE) ? " APSD_BE" : "",
	       (sta->flags & WL_STA_APSD_BK) ? " APSD_BK" : "",
	       (sta->flags & WL_STA_APSD_VI) ? " APSD_VI" : "",
	       (sta->flags & WL_STA_APSD_VO) ? " APSD_VO" : "");

	return(0);
}

static int
wl_revinfo(void *wl, cmd_t *cmd, char **argv)
{
	int err;
	wlc_rev_info_t revinfo;

	err = wl_get(wl, WLC_GET_REVINFO, &revinfo, sizeof(revinfo));
	if (err) {
		return err;
	}

	printf("vendorid 0x%x\n", revinfo.vendorid);
	printf("deviceid 0x%x\n", revinfo.deviceid);
	printf("radiorev 0x%x\n", revinfo.radiorev);
	printf("chipnum 0x%x\n", revinfo.chipnum);
	printf("chiprev 0x%x\n", revinfo.chiprev);
	printf("corerev 0x%x\n", revinfo.corerev);
	printf("boardid 0x%x\n", revinfo.boardid);
	printf("boardvendor 0x%x\n", revinfo.boardvendor);
	printf("boardrev 0x%x\n", revinfo.boardrev);
	printf("driverrev 0x%x\n", revinfo.driverrev);
	printf("ucoderev 0x%x\n", revinfo.ucoderev);
	printf("bus 0x%x\n", revinfo.bus);

	return 0;
}

static int
wl_rm_request(void *wl, cmd_t *cmd, char **argv)
{
	miniopt_t to;
	char* fn_name = "wl_rm_request";
	wl_rm_req_t *rm_ptr;
	wl_rm_req_t rm;
	wl_rm_req_elt_t req;
	int buflen = 0;
	int err, opt_err;
	int type;
	bool in_measure = FALSE;
	
	memset(buf, 0, WLC_IOCTL_MAXLEN);
	memset(&rm, 0, WL_RM_REQ_FIXED_LEN);
	memset(&req, 0, sizeof(wl_rm_req_elt_t));
	
	strcpy(buf, "rm_req");
	buflen = strlen(buf) + 1;

	rm_ptr = (wl_rm_req_t*)(buf + buflen);
	buflen += WL_RM_REQ_FIXED_LEN;

	/* toss the command name */
	argv++;
	
	miniopt_init(&to, fn_name, "p");
	while ((opt_err = miniopt(&to, argv)) != -1) {
		if (opt_err == 1) {
			err = -1;
			goto exit;
		}
		argv += to.consumed;
		
		if (to.opt == 't') {
			if (!to.good_int) {
				fprintf(stderr, "%s: could not parse \"%s\" as an int for the token\n",
					fn_name, to.valstr);
				err = -1;
				goto exit;
			}
				
			if (!in_measure)
				rm.token = to.val;
			else
				req.token = to.val;
		}
		if (to.opt == 'c') {
			if (!to.good_int) {
				fprintf(stderr, "%s: could not parse \"%s\" as an int for channel\n",
					fn_name, to.valstr);
				err = -1;
				goto exit;
			}
			req.chanspec = to.val;
		}
		if (to.opt == 'd') {
			if (!to.good_int) {
				fprintf(stderr, "%s: could not parse \"%s\" as an int for duration\n",
					fn_name, to.valstr);
				err = -1;
				goto exit;
			}
			req.dur = to.val;
		}

		if (to.opt == 'p') {
			req.flags = WL_RM_FLAG_PARALLEL;
		}

		if (to.positional) {
			if (!strcmp(to.valstr, "basic")) {
				type = WL_RM_TYPE_BASIC;
			} else if (!strcmp(to.valstr, "cca")) {
				type = WL_RM_TYPE_CCA;
			} else if (!strcmp(to.valstr, "rpi")) {
				type = WL_RM_TYPE_RPI;
			} else {
				fprintf(stderr, "%s: could not parse \"%s\" as a measurement type\n",
					fn_name, to.valstr);
				err = -1;
				goto exit;
			}
			/* complete the previous measurement */
			if (in_measure) {
				memcpy(buf + buflen, &req, sizeof(wl_rm_req_elt_t));
				buflen += sizeof(wl_rm_req_elt_t);
				rm.count++;
				/* measure to measure default param update */
				req.token++;	/* each measure gets a new token */
				req.flags = 0;	/* measure flags are cleared between measures */
			}
			in_measure = TRUE;
			req.type = (int8)type;
		}
	}

	/* complete the last measurement */
	if (in_measure) {
		memcpy(buf + buflen, &req, sizeof(wl_rm_req_elt_t));
		buflen += sizeof(wl_rm_req_elt_t);
		rm.count++;
	}
	
	if (rm.count == 0) {
		fprintf(stderr, "%s: no measurement requests specified\n",
			fn_name);
		err = -1;
		goto exit;
	}

	memcpy(rm_ptr, &rm, WL_RM_REQ_FIXED_LEN);

	err = wl_set(wl, WLC_SET_VAR, &buf, buflen);

 exit:
	return err;
}

static int
wl_rm_report(void *wl, cmd_t *cmd, char **argv)
{
	wl_rm_rep_t *rep_set;
	wl_rm_rep_elt_t rep;
	char extra[128];
	char* p;
	char* name;
	uint8* data;
	int err, bin;
	uint32 val;
	uint16 channel, band;
	int len;
	
	strcpy(buf, "rm_rep");

	err = wl_get(wl, WLC_GET_VAR, &buf, WLC_IOCTL_MAXLEN);
	if (err) {
		return err;
	}

	rep_set = (wl_rm_rep_t *)buf;

	printf("Measurement Report: token %d, length %d\n", rep_set->token, rep_set->len);

	len = rep_set->len;
	data = (uint8*)rep_set->rep;
	for ( ; len > 0; (len -= rep.len), (data += rep.len)) {
		if (len >= WL_RM_REP_ELT_FIXED_LEN)
			memcpy(&rep, data, WL_RM_REP_ELT_FIXED_LEN);
		else
			break;
		
		data += WL_RM_REP_ELT_FIXED_LEN;
		len -= WL_RM_REP_ELT_FIXED_LEN;

		if (rep.type == WL_RM_TYPE_BASIC)
			name = "Basic";
		else if (rep.type == WL_RM_TYPE_CCA)
			name = "CCA";
		else if (rep.type == WL_RM_TYPE_RPI)
			name = "RPI";
		else 
			name = NULL;
		
		if (name)
			printf("\nReport   : %s\n", name);
		else
			printf("\nReport   : %d <unknown>\n", rep.type);

		p = extra;
		if (rep.flags & WL_RM_FLAG_PARALLEL) {
			if (p != extra) p += sprintf(p, " | ");
			p += sprintf(p, "Parallel");
		}
		if (rep.flags & WL_RM_FLAG_LATE) {
			if (p != extra) p += sprintf(p, " | ");
			p += sprintf(p, "Late");
		}
		if (rep.flags & WL_RM_FLAG_INCAPABLE) {
			if (p != extra) p += sprintf(p, " | ");
			p += sprintf(p, "Incapable");
		}
		if (rep.flags & WL_RM_FLAG_REFUSED) {
			if (p != extra) p += sprintf(p, " | ");
			p += sprintf(p, "Refused");
		}

		if (p != extra) {
			printf("flags    : 0x%02x (%s)\n", rep.flags, extra);
		} else {
			printf("flags    : 0x%02x\n", rep.flags);
		}
		printf("token    : %4d\n", rep.token);

		if (rep.flags & (WL_RM_FLAG_LATE |
				 WL_RM_FLAG_INCAPABLE |
				 WL_RM_FLAG_REFUSED)) {
			continue;
		}
		
		channel = (rep.chanspec & WL_CHANSPEC_CHAN_MASK);
		band = (rep.chanspec & WL_CHANSPEC_BAND_MASK);
		
		printf("channel  : %4d %s\n", channel, 
		       (band == WL_CHANSPEC_BAND_A)? "(a)":"(b)");
		printf("start tsf: 0x%x:%08x\n", rep.tsf_h, rep.tsf_l);
		printf("duration : %4d TU\n", rep.dur);

		if (len < (int)rep.len) {
			printf("Error: partial report element, %d report bytes "
			       "remain, element claims %d\n",
			       len, rep.len);
			break;
		}
		
		if (rep.type == WL_RM_TYPE_BASIC) {
			if (rep.len >= 4) {
				memcpy(&val, data, sizeof(uint32));
				printf("Basic bits: 0x%08x\n", val);
			}
		} else if (rep.type == WL_RM_TYPE_CCA) {
			if (rep.len >= 4) {
				memcpy(&val, data, sizeof(uint32));
				printf("Carrier Fraction: %d / 255\n", val);
			}
		} else if (rep.type == WL_RM_TYPE_RPI) {
			if (rep.len >= sizeof(wl_rm_rpi_rep_t)) {
				wl_rm_rpi_rep_t rpi_rep;
				int8 min = -128;
				int8 max;
				
				memcpy(&rpi_rep, data, sizeof(wl_rm_rpi_rep_t));
				
				for (bin = 0; bin < 8; bin++) {
					max = rpi_rep.rpi_max[bin];
					if (bin == 0)
						printf("       Power <= %3d: ",
						       max);
					else if (bin < 7)
						printf(" %3d < Power <= %3d: ",
						       min, max);
					else
						printf(" %3d < Power       : ",
						       min);
					min = max;
					printf("%3d\n", rpi_rep.rpi[bin]);
				}
			}
		}
	}

	return err;
}
static int
wl_join_pref(void *wl, cmd_t *cmd, char **argv)
{
	uint8* data;
	int err;
	int len;
	int i;
	
	strcpy(buf, "join_pref");

	/* set */
	if (argv[1]) {
		len = strlen(buf);
		data = (uint8 *)argv[1];
		for (i = len + 1, len += 1 + strlen(data) / 2;
		    (i < len) && (i < (int)sizeof (buf)); i ++) {
			uint8 hex[] = "XX";
			hex[0] = *data++;
			hex[1] = *data++;
			buf[i] = (uint8)strtoul(hex, NULL, 16);
		}
		err = wl_set(wl, WLC_SET_VAR, buf, i);
	}
	/* get */
	else if (!(err = wl_get(wl, WLC_GET_VAR, buf, WLC_IOCTL_MAXLEN))) {
		len = *(int *)buf;
		data = buf + sizeof(int);
		for (i = 0; i < len; i ++)
			printf("%02x", data[i]);
		printf("\n");
	}
	return err;
}
static int
wl_assoc_pref(void *wl, cmd_t *cmd, char **argv)
{
	uint assoc_pref;
	int err;
	
	/* set */
	if (argv[1]) {
		if (!strcmp(argv[1], "auto") || !strcmp(argv[1], "0"))
			assoc_pref = WLC_BAND_AUTO;
		else if (!strcmp(argv[1], "a") || !strcmp(argv[1], "1"))
			assoc_pref = WLC_BAND_A;
		else if (!strcmp(argv[1], "b") || !strcmp(argv[1], "g") || !strcmp(argv[1], "2"))
			assoc_pref = WLC_BAND_B;
		else
			return -1;
		err = wl_set(wl, cmd->set, &assoc_pref, sizeof(assoc_pref));
	}
	/* get */
	else if (!(err = wl_get(wl, cmd->get, &assoc_pref, sizeof(assoc_pref)))) {
		switch (assoc_pref) {
		case WLC_BAND_AUTO:
			printf("auto\n");
			break;
		case WLC_BAND_A:
			printf("a\n");
			break;
		case WLC_BAND_B:
			printf("b/g\n");
			break;
		}
	}
	return err;
}

/*
 * Get or Set WME Access Class (AC) parameters
 *	wl wme_ac ap|sta [be|bk|vi|vo [ecwmax|ecwmin|txop|aifsn|acm <value>] ...]
 * Without args past ap|sta, print current values
 */
static int
wl_wme_ac_req(void *wl, cmd_t *cmd, char **argv)
{
	int err, buflen;
	wme_acparam_t acparam_cur[AC_COUNT], acparam_new[AC_COUNT], *acp;
	char *ac_str, *param, *val;
	bool acm;
	int aci, aifsn, ecwmin, ecwmax, txop;
	char ac_names[AC_COUNT][6] = {"AC_BE", "AC_BK", "AC_VI", "AC_VO"};
	int ap_mode = 0;
	int sta_param = 0;

	argv++;

	if ((err = wl_get(wl, WLC_GET_AP, &ap_mode, sizeof(ap_mode))))
		return err;

	if ((param = *argv++) == NULL)
		return USAGE_ERROR;

	if (!strcmp(param, "ap"))
		sta_param = 0;
	else if (!strcmp(param, "sta"))
		sta_param = 1;
	else
		return USAGE_ERROR;

	if (!ap_mode && !sta_param) {
		fprintf(stderr, "Can't work with AP parameters on STA\n");
		return USAGE_ERROR;
	}

	/*
	 * On call to wl_get, buf contains the NUL-terminated
	 * string "wme_ac_sta" or "wme_ac_ap".
	 * On return, gotten data starts at beginning of buf.
	 *
	 * On call to wl_set, buf contains the NUL-terminated
	 * string "wme_ac_sta" or "wme_ac_ap", followed by the data.
	 * Only a return value is returned.
	 *
	 */
	if (sta_param)
		strcpy(buf, "wme_ac_sta");
	else
		strcpy(buf, "wme_ac_ap");

	buflen = WLC_IOCTL_MAXLEN;

	/*
	 * Get current acparams, using buf as an input buffer.
	 * Return data is array of 4 ACs of wme params.
	 */

	if ((err = wl_get(wl, cmd->get, &buf, buflen)) < 0)
		return err;

	memcpy(&acparam_cur, buf, sizeof(acparam_cur));

	if ((ac_str = *argv++) == NULL) {
		printf("AC Parameters %s\n",
		       ap_mode ? (sta_param ? "advertised for STA" : "for AP") : "for STA");

		for (aci = 0; aci < AC_COUNT; aci++) {
			acp = &acparam_cur[aci];
			if (((acp->ACI & WME_ACI_MASK) >> WME_ACI_SHIFT) != aci)
				printf("Warning: AC params out of order\n");
			acm = (acp->ACI & WME_ACM_MASK) ? 1 : 0;
			aifsn = acp->ACI & WME_AIFSN_MASK;
			ecwmin = acp->ECW & WME_ECWMIN_MASK;
			ecwmax = (acp->ECW & WME_ECWMAX_MASK) >> WME_ECWMAX_SHIFT;
			txop = acp->TXOP;
			printf("%s: raw: ACI 0x%x ECW 0x%x TXOP 0x%x\n",
			       ac_names[aci],
			       acp->ACI, acp->ECW, acp->TXOP);
			printf("       dec: aci %d acm %d aifsn %d "
			       "ecwmin %d ecwmax %d txop 0x%x\n",
			       aci, acm, aifsn, ecwmin, ecwmax, txop);
			/* CWmin = 2^(ECWmin) - 1 */
			/* CWmax = 2^(ECWmax) - 1 */
			/* TXOP = number of 32 us units */
			printf("       eff: CWmin %d CWmax %d TXop %dusec\n",
			       WME_ECW2CW(ecwmin), WME_ECW2CW(ecwmax), WME_TXOP2USEC(txop));
		}

		err = 0;
	} else {
		if (strcmp(ac_str, "be") == 0)
			aci = AC_BE;
		else if (strcmp(ac_str, "bk") == 0)
			aci = AC_BK;
		else if (strcmp(ac_str, "vi") == 0)
			aci = AC_VI;
		else if (strcmp(ac_str, "vo") == 0)
			aci = AC_VO;
		else
			return USAGE_ERROR;

		/* Preload new values with current values */
		memcpy(&acparam_new, &acparam_cur, sizeof(acparam_new));

		acp = &acparam_new[aci];

		while ((param = *argv++) != NULL) {
			if ((val = *argv++) == NULL)
				return USAGE_ERROR;

			if (!strcmp(param, "acm")) {
				if (!stricmp(val, "on") || !stricmp(val, "1"))
					acp->ACI |= WME_ACM_MASK;
				else if (!stricmp(val, "off") || !stricmp(val, "0"))
					acp->ACI &= ~WME_ACM_MASK;
				else {
					fprintf(stderr, "acm value must be 1|0\n");
					return USAGE_ERROR;
				}
			} else if (!strcmp(param, "aifsn")) {
				aifsn = (int)strtol(val, NULL, 0);
				if (aifsn >= WME_AIFSN_MIN && aifsn <= WME_AIFSN_MAX)
					acp->ACI =
					        (acp->ACI & ~WME_AIFSN_MASK) |
					        (aifsn & WME_AIFSN_MASK);
				else {
					fprintf(stderr, "aifsn %d out of range (%d-%d)\n",
					        aifsn, WME_AIFSN_MIN, WME_AIFSN_MAX);
					return USAGE_ERROR;
				}
			} else if (!strcmp(param, "ecwmax")) {
				ecwmax = (int)strtol(val, NULL, 0);
				if (ecwmax >= WME_ECW_MIN && ecwmax <= WME_ECW_MAX)
					acp->ECW =
					        ((ecwmax << WME_ECWMAX_SHIFT) & WME_ECWMAX_MASK) |
					        (acp->ECW & WME_ECWMIN_MASK);
				else {
					fprintf(stderr, "ecwmax %d out of range (%d-%d)\n",
					        ecwmax, WME_ECW_MIN, WME_ECW_MAX);
					return USAGE_ERROR;
				}
			} else if (!strcmp(param, "ecwmin")) {
				ecwmin = (int)strtol(val, NULL, 0);
				if (ecwmin >= WME_ECW_MIN && ecwmin <= WME_ECW_MAX)
					acp->ECW =
					        ((acp->ECW & WME_ECWMAX_MASK) |
					         (ecwmin & WME_ECWMIN_MASK));
				else {
					fprintf(stderr, "ecwmin %d out of range (%d-%d)\n",
					        ecwmin, WME_ECW_MIN, WME_ECW_MAX);
					return USAGE_ERROR;
				}
			} else if (!strcmp(param, "txop")) {
				txop = (int)strtol(val, NULL, 0);
				if (txop >= WME_TXOP_MIN && txop <= WME_TXOP_MAX)
					acp->TXOP = txop;
				else {
					fprintf(stderr, "txop %d out of range (%d-%d)\n",
					        txop, WME_TXOP_MIN, WME_TXOP_MAX);
					return USAGE_ERROR;
				}
			} else {
				fprintf(stderr, "unexpected param %s\n", param);
				return USAGE_ERROR;
			}
		}

		/*
		 * Now use buf as an output buffer.
		 * Put WME acparams after "wme_ac\0" in buf.
		 * NOTE: only one of the four ACs can be set at a time.
		 */
		if (sta_param)
			strcpy(buf, "wme_ac_sta");
		else
			strcpy(buf, "wme_ac_ap");

		memcpy(buf + strlen(buf) + 1, acp, sizeof(wme_acparam_t));

		err = wl_set(wl, cmd->set, &buf, buflen);
	}

	return err;
}

/*
 * Get or Set WME APSD control parameters
 *	wl wme_apsd_sta <max_sp_len> <be> <bk> <vi> <vo>
 *	  <max_sp_len> is 0 (all), 2, 4, or 6
 *        <be>, <bk>, <vi>, <vo> are each 0 or 1 for APSD enable
 *  with no args, print current values
 */
static int
wl_wme_apsd_sta(void *wl, cmd_t *cmd, char **argv)
{
	int err;
	int buflen;
	char *param;
	int ap_mode;
	int qosinfo;
	int msp, max_sp_len, be, bk, vi, vo;

	if ((err = wl_get(wl, WLC_GET_AP, &ap_mode, sizeof (ap_mode))))
		return err;

	if (ap_mode) {
		printf("%s: STA only\n", cmd->name);
		return -1;
	}

	/* Display current params if no args, else set params */

	memset(buf, 0, WLC_IOCTL_MAXLEN);
	strcpy(buf, "wme_qosinfo");
	buflen = WLC_IOCTL_MAXLEN;

	param = *++argv;

	if (param == NULL) {
		if ((err = wl_get(wl, cmd->get, &buf, buflen)) < 0)
			return err;

		memcpy(&qosinfo, buf, sizeof (qosinfo));

		msp = (qosinfo & WME_QI_STA_MAXSPLEN_MASK) >> WME_QI_STA_MAXSPLEN_SHIFT;
		be = (qosinfo & WME_QI_STA_APSD_BE_MASK) >> WME_QI_STA_APSD_BE_SHIFT;
		bk = (qosinfo & WME_QI_STA_APSD_BK_MASK) >> WME_QI_STA_APSD_BK_SHIFT;
		vi = (qosinfo & WME_QI_STA_APSD_VI_MASK) >> WME_QI_STA_APSD_VI_SHIFT;
		vo = (qosinfo & WME_QI_STA_APSD_VO_MASK) >> WME_QI_STA_APSD_VO_SHIFT;

		max_sp_len = msp * 2;

		printf("Max SP Length = %d, APSD: BE=%d BK=%d VI=%d VO=%d\n",
		       max_sp_len, be, bk, vi, vo);
	} else {
		max_sp_len = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		be = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		bk = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		vi = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		vo = (int)strtol(param, 0, 0);

		if (((be | bk | vi | vo) & ~1) | (max_sp_len & ~6)) {
			printf("%s: Invalid parameter\n", cmd->name);
			return -1;
		}

		msp = max_sp_len / 2;

		qosinfo = (msp << WME_QI_STA_MAXSPLEN_SHIFT) & WME_QI_STA_MAXSPLEN_MASK;
		qosinfo |= (be << WME_QI_STA_APSD_BE_SHIFT) & WME_QI_STA_APSD_BE_MASK;
		qosinfo |= (bk << WME_QI_STA_APSD_BK_SHIFT) & WME_QI_STA_APSD_BK_MASK;
		qosinfo |= (vi << WME_QI_STA_APSD_VI_SHIFT) & WME_QI_STA_APSD_VI_MASK;
		qosinfo |= (vo << WME_QI_STA_APSD_VO_SHIFT) & WME_QI_STA_APSD_VO_MASK;

		memcpy(&buf[strlen(buf) + 1], &qosinfo, sizeof (qosinfo));

		err = wl_set(wl, cmd->set, &buf, buflen);
	}

	return err;
}

/*
 * Get or Set WME discard policy
 *	wl wme_dp <be> <bk> <vi> <vo>
 *        <be>, <bk>, <vi>, <vo> are each 0/1 for discard newest/oldest first
 *  with no args, print current values
 */
static int
wl_wme_dp(void *wl, cmd_t *cmd, char **argv)
{
	int err;
	int buflen;
	char *param;
	int dp;
	int be, bk, vi, vo;

	/* Display current params if no args, else set params */

	memset(buf, 0, WLC_IOCTL_MAXLEN);
	strcpy(buf, "wme_dp");
	buflen = WLC_IOCTL_MAXLEN;

	param = *++argv;

	if (param == NULL) {
		if ((err = wl_get(wl, cmd->get, &buf, buflen)) < 0)
			return err;

		memcpy(&dp, buf, sizeof (dp));

		be = (dp >> AC_BE) & 1;
		bk = (dp >> AC_BK) & 1;
		vi = (dp >> AC_VI) & 1;
		vo = (dp >> AC_VO) & 1;

		printf("Discard oldest first: BE=%d BK=%d VI=%d VO=%d\n", be, bk, vi, vo);
	} else {
		be = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		bk = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		vi = (int)strtol(param, 0, 0);
		if ((param = *++argv) == NULL)
			return -1;
		vo = (int)strtol(param, 0, 0);

		if ((be | bk | vi | vo) & ~1) {
			printf("%s: Invalid parameter\n", cmd->name);
			return -1;
		}

		dp = (be << AC_BE) | (bk << AC_BK) | (vi << AC_VI) | (vo << AC_VO);

		memcpy(&buf[strlen(buf) + 1], &dp, sizeof (dp));

		err = wl_set(wl, cmd->set, &buf, buflen);
	}

	return err;
}

static int
wl_add_ie(void *wl, cmd_t *cmd, char **argv)
{
	return (wl_vndr_ie(wl, "add", argv));
}

static int
wl_del_ie(void *wl, cmd_t *cmd, char **argv)
{
	return (wl_vndr_ie(wl, "del", argv));
}

static int
wl_vndr_ie(void *wl, char *command, char **argv)
{
	vndr_ie_setbuf_t *ie_setbuf;
	uint32 pktflag;
	int ielen, datalen, buflen, iecount;
	int err = 0;
	int ap_mode = 0;
	
	if (!argv[1] || !argv[2] || !argv[3]) {
		fprintf(stderr, "Too few arguments\n");
		return -1;
	} 

	if ((err = wl_get(wl, WLC_GET_AP, &ap_mode, sizeof (ap_mode)))) {
		return -1;
	}

        if (!ap_mode) {
		fprintf(stderr, "This command can only be executed on the AP\n");
		return -1;
	}

	pktflag = atoi (argv[1]);
	if ((pktflag & 
	    ~(VNDR_IE_BEACON_FLAG |
	     VNDR_IE_PRBRSP_FLAG |
	     VNDR_IE_ASSOCRSP_FLAG |
	     VNDR_IE_AUTHRSP_FLAG))) {
		fprintf(stderr, "Invalid packet flag 0x%x (%d)\n", pktflag, pktflag);
		return -1;
	}

	ielen = atoi (argv[2]);
	if (ielen > VNDR_IE_MAX_LEN) {
		fprintf(stderr, "IE length should be <= %d\n", VNDR_IE_MAX_LEN);
		return -1;
	}
	else if (ielen < VNDR_IE_MIN_LEN) {
		fprintf(stderr, "IE length should be >= %d\n", VNDR_IE_MIN_LEN);
		return -1;
	}

	if (strlen (argv[3]) != OUI_STR_SIZE) {
		fprintf(stderr, "Invalid OUI\n");
		return -1;
	}

	datalen = ielen - VNDR_IE_MIN_LEN;

	if (datalen > 0) {
		if (!argv[4]) {
			fprintf(stderr, "Data bytes should be specified for IE of length %d", ielen);
			return -1;
		}
		else {   
			/* Ensure each data byte is 2 characters long */
			if ((int)strlen (argv[4]) < (datalen * 2)) {
				fprintf(stderr, "Please specify all the data bytes for this IE\n");
				return -1;
			}
		}
	}

	if (!datalen && (argv[4] != NULL))
		fprintf(stderr, "Ignoring data bytes for IE of length %d", ielen);

	buflen = sizeof(vndr_ie_setbuf_t) + datalen - 1;
	ie_setbuf = (vndr_ie_setbuf_t *) malloc(buflen);

	if (! ie_setbuf) {
		fprintf(stderr, "memory alloc failure\n");
		return -1;
	}

	/* Copy the vndr_ie SET command ("add"/"del") to the buffer */
	strcpy (ie_setbuf->cmd, command);

	
	/* Buffer contains only 1 IE */
	iecount = 1;
	memcpy (&ie_setbuf->vndr_ie_buffer.iecount, &iecount, sizeof(int));

	/* 
	 * The packet flag bit field indicates the packets that will
	 * contain this IE
	 */
	memcpy (&ie_setbuf->vndr_ie_buffer.vndr_ie_list[0].pktflag, &pktflag, sizeof(uint32));

	/* Now, add the IE to the buffer */
	ie_setbuf->vndr_ie_buffer.vndr_ie_list[0].vndr_ie_data.len = (uchar) ielen;

	if ((err = get_oui_bytes ((uchar *)argv[3], 
			&ie_setbuf->vndr_ie_buffer.vndr_ie_list[0].vndr_ie_data.oui[0]))) {
		free (ie_setbuf);
		fprintf(stderr, "Error parsing OUI arg\n");
		return err;
	}

	if (datalen) {
		if ((err = get_ie_data ((uchar *)argv[4], 
				&ie_setbuf->vndr_ie_buffer.vndr_ie_list[0].vndr_ie_data.data[0], datalen))) {
			free (ie_setbuf);
			fprintf(stderr, "Error parsing data arg\n");
			return err;
		}
	}

	err = wl_var_setbuf (wl, "vndr_ie", ie_setbuf, buflen);

	free(ie_setbuf);

	return (err);
}

static int
wl_list_ie(void *wl, cmd_t *cmd, char **argv)
{
	uchar *iebuf;
	char *data;
	int tot_ie, pktflag, iecount, count, datalen, col;
	vndr_ie_buf_t *ie_getbuf;
	vndr_ie_info_t *ie_info;
	vndr_ie_t *ie;
	int err = 0;

	err = wl_var_getbuf (wl, "vndr_ie", NULL, 0, (void**)&ie_getbuf);
	if (err == 0) {
		memcpy (&tot_ie, &ie_getbuf->iecount, sizeof(int));
		printf("Total IEs %d\n", tot_ie);

		iebuf = (uchar *)&ie_getbuf->vndr_ie_list[0];

		for (iecount = 0; iecount < tot_ie; iecount++) {
			ie_info = (vndr_ie_info_t *) iebuf;
			memcpy (&pktflag, &ie_info->pktflag, sizeof(uint32));
			iebuf += sizeof(uint32);

			printf("\n");

			ie = &ie_info->vndr_ie_data;
			printf("IE index = %d\n", iecount);
			printf("-----------------\n");
			printf("Pkt Flg = 0x%x\n", pktflag);
			printf("Length  = %d\n", ie->len);
			printf("OUI     = %02x:%02x:%02x\n", 
			         	ie->oui[0], ie->oui[1], ie->oui[2]);
			printf("Data:\n");
	
			data = &ie->data[0];
			datalen = ie->len - VNDR_IE_MIN_LEN;	
			for (count = 0; (count < datalen); ) {
				for (col = 0; (col < MAX_DATA_COLS) && (count < datalen); col++, count++) {
					printf("%02x ", *data++);
				}
				printf("\n");
			}

			iebuf += ie->len + VNDR_IE_HDR_LEN;	
		}
	}
	else {
		fprintf(stderr, "Error %d getting IOVar\n", err);
	}

	return err;
}

static int
wl_rand(void *wl, cmd_t *cmd, char **argv)
{
	char *randbuf;
	uint16 randnum;
	int err;

	if ((err = wl_var_getbuf (wl, cmd->name, NULL, 0, (void**)&randbuf)))
		return (err);

	memcpy(&randnum, randbuf, sizeof (uint16));
	printf("%d\n", randnum);

	return (0);
}

#define	PRVAL(name)	pbuf += sprintf(pbuf, "%s %d ", #name, cnt.name)
#define	PRNL()		pbuf += sprintf(pbuf, "\n")

static int
wl_counters(void *wl, cmd_t *cmd, char **argv)
{
	char *statsbuf;
	wl_cnt_t cnt;
	int err;
	uint i;
	char *pbuf = buf;

	if ((err = wl_var_getbuf (wl, cmd->name, NULL, 0, (void**)&statsbuf)))
		return (err);

	memcpy(&cnt, statsbuf, sizeof(cnt));

	if (cnt.version != WL_CNT_T_VERSION) {
		printf("\tIncorrect version of counters struct: expected %d; got %d\n",
			   WL_CNT_T_VERSION, cnt.version);
		return -1;
	}

	/* summary stat counter line */
	PRVAL(txframe); PRVAL(txbyte); PRVAL(txretrans); PRVAL(txerror); 
	PRVAL(rxframe); PRVAL(rxbyte); PRVAL(rxerror); PRNL();

	PRVAL(txprshort); PRVAL(txdmawar); PRVAL(txnobuf); PRVAL(txnoassoc); 
	PRVAL(txchit); PRVAL(txcmiss); PRNL();

	PRVAL(reset); PRVAL(txserr); PRVAL(txphyerr); PRVAL(txphycrs); 
	PRVAL(txfail); PRNL();

	pbuf += sprintf(pbuf, "d11_txfrag %d d11_txmulti %d d11_txretry %d d11_txretrie %d\n",
		cnt.txfrag, cnt.txmulti, cnt.txretry, cnt.txretrie);

	pbuf += sprintf(pbuf, "d11_txrts %d d11_txnocts %d d11_txnoack %d d11_txfrmsnt %d\n",
		cnt.txrts, cnt.txnocts, cnt.txnoack, cnt.txfrmsnt);

	PRVAL(rxcrc); PRVAL(rxnobuf); PRVAL(rxnondata); PRVAL(rxbadds); 
	PRVAL(rxbadcm); PRVAL(rxdup); PRVAL(rxfragerr); PRNL();

	PRVAL(rxrunt); PRVAL(rxgiant); PRVAL(rxnoscb); PRVAL(rxbadproto); 
	PRVAL(rxbadsrcmac); PRNL();

	pbuf += sprintf(pbuf, "d11_rxfrag %d d11_rxmulti %d d11_rxundec %d\n",
		cnt.rxfrag, cnt.rxmulti, cnt.rxundec);

	PRVAL(rxctl); PRVAL(rxbadda); PRVAL(rxfilter); PRNL();

	pbuf += sprintf(pbuf, "rxuflo: ");
	for (i = 0; i < NFIFO; i++)
		pbuf += sprintf(pbuf, "%d ", cnt.rxuflo[i]);
	pbuf += sprintf(pbuf, "\n");
	PRVAL(txallfrm); PRVAL(txrtsfrm); PRVAL(txctsfrm); PRVAL(txackfrm); PRNL();
	PRVAL(txdnlfrm); PRVAL(txbcnfrm); PRVAL(txtplunfl); PRVAL(txphyerr); PRNL();
	pbuf += sprintf(pbuf, "txfunfl: ");
	for (i = 0; i < NFIFO; i++)
		pbuf += sprintf(pbuf, "%d ", cnt.txfunfl[i]);
	pbuf += sprintf(pbuf, "\n");

	/* WPA2 counters */
	PRNL();
	PRVAL(tkipmicfaill); PRVAL(tkipicverr); PRVAL(tkipcntrmsr); PRNL(); 
	PRVAL(tkipreplay); PRVAL(ccmpfmterr); PRVAL(ccmpreplay); PRNL();
	PRVAL(ccmpundec); PRVAL(fourwayfail); PRVAL(wepundec); PRNL();
	PRVAL(wepicverr); PRVAL(decsuccess); PRVAL(rxundec); PRNL(); 

	PRNL();
	PRVAL(rxfrmtoolong); PRVAL(rxfrmtooshrt); PRVAL(rxinvmachdr); PRVAL(rxbadfcs); PRNL();
	PRVAL(rxbadplcp); PRVAL(rxcrsglitch); PRVAL(rxstrt); PRVAL(rxdfrmucastmbss); PRNL();
	PRVAL(rxmfrmucastmbss); PRVAL(rxcfrmucast); PRVAL(rxrtsucast); PRVAL(rxctsucast); PRNL();
	PRVAL(rxackucast); PRVAL(rxdfrmocast); PRVAL(rxmfrmocast); PRVAL(rxcfrmocast); PRNL();
	PRVAL(rxrtsocast); PRVAL(rxctsocast); PRVAL(rxdfrmmcast); PRVAL(rxmfrmmcast); PRNL();
	PRVAL(rxcfrmmcast); PRVAL(rxbeaconmbss); PRVAL(rxdfrmucastobss); PRVAL(rxbeaconobss); PRNL();
	PRVAL(rxrsptmout); PRVAL(bcntxcancl); PRVAL(rxf0ovfl); PRVAL(rxf1ovfl); PRNL();
	PRVAL(rxf2ovfl); PRVAL(txsfovfl); PRVAL(pmqovfl); PRNL();
	PRVAL(rxcgprqfrm); PRVAL(rxcgprsqovfl); PRVAL(txcgprsfail); PRVAL(txcgprssuc); PRNL();
	PRVAL(prs_timeout); PRVAL(rxnack); PRVAL(frmscons); PRVAL(txnack); PRVAL(txglitch_nack); PRNL();
	PRVAL(txburst); PRVAL(rxburst); PRVAL(txphyerror); PRNL();

	printf(buf);
	printf("\n");
	return (0);
}

static int
wl_wme_counters(void *wl, cmd_t *cmd, char **argv)
{
	char *statsbuf;
	wl_wme_cnt_t cnt;
	int err;
	void *ptr;
	char *pbuf = buf;
	char ac_names[AC_COUNT][6] = {"AC_BE", "AC_BK", "AC_VI", "AC_VO"};
	uint ac;
	int ap_mode = 0;

	if ((err = wl_var_getbuf (wl, cmd->name, NULL, 0, &ptr)))
		return (err);

	statsbuf = (char *)ptr;
	memcpy(&cnt, statsbuf, sizeof(cnt));

	if (cnt.version != WL_WME_CNT_VERSION) {
		printf("\tIncorrect version of counters struct: expected %d; got %d\n",
			WL_CNT_T_VERSION, cnt.version);
		return -1;
	}

	if ((err = wl_get(wl, WLC_GET_AP, &ap_mode, sizeof(ap_mode)))) {
		return -1;
	}

	/* summary stat counter line */
	for (ac = AC_BE; ac < AC_COUNT; ac++) {
		pbuf += sprintf(pbuf, "%s: tx frames: %d bytes: %d failed frames: %d "
		                "failed bytes: %d\n",
		                ac_names[ac], cnt.tx[ac].packets, cnt.tx[ac].bytes,
		                cnt.tx_failed[ac].packets, cnt.tx_failed[ac].bytes);
		pbuf += sprintf(pbuf, "       rx frames: %d bytes: %d failed frames: %d "
		                "failed bytes: %d\n", cnt.rx[ac].packets, cnt.rx[ac].bytes,
		                cnt.rx_failed[ac].packets, cnt.rx_failed[ac].bytes);

		if (ap_mode)
			pbuf += sprintf(pbuf, "       foward frames: %d bytes: %d \n",
			                cnt.forward[ac].packets, cnt.forward[ac].bytes);
	}

	printf(buf);
	printf("\n");
	return (0);
}

static int
wl_diag(void *wl, cmd_t *cmd, char **argv)
{
	uint testindex;
	int buflen, err;
	char *param;
	uint32 testresult;

	if (!*++argv){
		printf(" Usage: %s testindex[1-4]\n", cmd->name);
		return -1;
	}
		
	testindex = atoi(*argv);
	
	strcpy(buf, "diag");
	buflen = strlen(buf) + 1;
	param = (char *)(buf + buflen);
	memcpy(param, (char*)&testindex, sizeof(testindex));
	
	if ((err = wl_get(wl, cmd->get, buf, sizeof(buf)))) 
		return err;

	testresult = *(uint32 *)buf;
	if (testresult != 0 ) {
		printf("\ndiag test %d failed(error code %d)\n", testindex, testresult);
	} else 
		printf("\ndiag test %d passed\n", testindex);

	return(0);
}

void
miniopt_init(miniopt_t *t, char* name, char* flags)
{
	static char *null_flags = "";
	
	memset(t, 0, sizeof (miniopt_t));
	t->name = name;
	if (flags == NULL)
		t->flags = null_flags;
	else
	t->flags = flags;
}

int 
miniopt(miniopt_t *t, char **argv)
{
	int keylen;
	char *p, *eq, *valstr, *endptr;
	int err = 0;

	t->consumed = 0;
	t->positional = FALSE;
	memset(t->key, 0, MINIOPT_MAXKEY);
	t->opt = '\0';
	t->valstr = NULL;
	t->good_int = FALSE;

	if (*argv == NULL) {
		err = -1;
		goto exit;
	}
	
	p = *argv++;
	t->consumed++;

	if (!t->opt_end && !strcmp(p, "--")) {
		t->opt_end = TRUE;
		if (*argv == NULL) {
			err = -1;
			goto exit;
		}
		p = *argv++;
		t->consumed++;
	}

	if (t->opt_end) {
		t->positional = TRUE;
		valstr = p;
	}
	else if (!strncmp(p, "--", 2)) {
		eq = strchr(p, '=');
		if (eq == NULL) {
			fprintf(stderr, "%s: missing \"=\" in long param \"%s\"\n", t->name, p);
			err = 1;
			goto exit;
		}
		keylen = eq - (p + 2);
		if (keylen > 63) keylen = 63;
		memcpy(t->key, p + 2, keylen);

		valstr = eq + 1;
		if (*valstr == '\0') {
			fprintf(stderr, "%s: missing value after \"=\" in long param \"%s\"\n", t->name , p);
			err = 1;
			goto exit;
		}
	}
	else if (!strncmp(p, "-", 1)) {
		t->opt = p[1];
		if (strlen(p) > 2) {
			fprintf(stderr, "%s: only single char options, error on param \"%s\"\n", t->name, p);
			err = 1;
			goto exit;
		}
		if (strchr(t->flags, t->opt)) {
			/* this is a flag option, no value expected */
			valstr = NULL;
		} else {
			if (*argv == NULL) {
				fprintf(stderr, "%s: missing value parameter after \"%s\"\n", t->name, p);
				err = 1;
				goto exit;
			}
			valstr = *argv;
			argv++;
			t->consumed++;
		}
	} else {
		t->positional = TRUE;
		valstr = p;
	}

	/* parse valstr as int just in case */
	if (valstr) {
		t->val = (int)strtol(valstr, &endptr, 0);
		if (*endptr == '\0') {
				/* not all the value string was parsed by strtol */
			t->good_int = TRUE;
		}
	}

	t->valstr = valstr;

 exit:
	if (err == 1)
		t->opt = '?';
	
	return err;
}

static int
get_oui_bytes (uchar *oui_str, uchar *oui)
{
	int index;
	uchar val;
	uchar *src, *dest;
	char hexstr[3];

	src = oui_str;
	dest = oui;

	for (index = 0; index < MAX_OUI_SIZE; index++) {
		hexstr[0] = src[0];
		hexstr[1] = src[1];
		hexstr[2] = '\0';

		val = (uchar) strtoul(hexstr, NULL, 16);

		*dest++ = val;
		src += 2;

		if ((index < (MAX_OUI_SIZE - 1)) && (*src++ != ':'))
			return -1;
	}

	return 0;
}

static int 
get_ie_data (uchar *data_str, uchar *ie_data, int len)
{
	uchar *src, *dest;
	uchar val;
	int index;
	char hexstr[3];

	src = data_str;
	dest = ie_data;

	for (index = 0; index < len; index++) {
		hexstr[0] = src[0];
		hexstr[1] = src[1];
		hexstr[2] = '\0';

		val = (uchar) strtoul(hexstr, NULL, 16);

		*dest++ = val;
		src += 2;
	}

	return 0;
}

static int
hexstrtobitvec(const uchar *cp, uchar *bitvec, int veclen)
{
	uchar value = 0;
	int nibble;		/* index of current hex-format nibble to process */
	int even;		/* 1 if even number of nibbles, 0 if odd number */
	int i = 0;

	if (cp[0] == '0' && cp[1] == 'x')
		cp += 2;

	memset(bitvec, '\0', veclen);
	nibble = strlen(cp);
	if (!nibble)
		return -1;
	even = ((nibble % 2) == 0);

	/* convert from right to left (lsb is rightmost byte) */
	--nibble;
	while (nibble >= 0 && i < veclen && (isxdigit(cp[nibble]) && 
           (value = isdigit(cp[nibble]) ? cp[nibble]-'0' : 
			(islower(cp[nibble]) ? toupper(cp[nibble]) : cp[nibble])-'A'+10) < 16)) {
		if (even == ((nibble+1) % 2)) {
			bitvec[i] += value*16;
			++i;
		} else
			bitvec[i] = value;
		--nibble;
	}

	return ((nibble == -1 && i <= veclen) ? 0 : -1);
}

static int
wl_bitvec128(void *wl, cmd_t *cmd, char **argv)
{
	char *vbuf;
	int err;
	uchar bitvec[16];
	bool skipzeros;
	int i;

	err = 0;
	skipzeros = TRUE;

	/* set */
	if (argv[1]) {
		memset(bitvec, '\0', sizeof(bitvec));
		if (!(err = hexstrtobitvec(argv[1], bitvec, sizeof(bitvec))))
			err = wl_var_setbuf(wl, cmd->name, bitvec, sizeof(bitvec));
	}
	/* get */
	else {
		memset(buf, '\0', sizeof(buf));
		if (!(err = wl_var_getbuf(wl, cmd->name, NULL, 0, (void **)&vbuf))) {
			printf("0x");
			for (i = (sizeof (bitvec) - 1); i >= 0; i--) {
				if (vbuf[i] || (i == 0))
					skipzeros = FALSE;
				if (skipzeros)
					continue;
				printf("%02x", vbuf[i] & 0xff);
			}
			printf("\n");
		}
	} 

	return (err);
}

static int
wl_auto_channel_sel(void *wl, cmd_t *cmd, char **argv)
{
	/*
	* The following condition(s) must be met when Auto Channel Selection 
	* is enabled.
	*  - the I/F is up (change radio channel requires it is up?)
	*  - the AP must not be associated (setting SSID to empty should 
	*    make sure it for us)
	*/
	int chosen = 0;
	wl_uint32_list_t request;
	int ret = 0;
	
	if (!*++argv) {
		ret = wl_get(wl, cmd->get, &chosen, sizeof(chosen));
		if (ret >= 0 && chosen != 0) {
			printf("%d\n", chosen);	
			return 0;	
		}		
	} else {				
		if (atoi(*argv) == 1) {
			request.count = 0;			
			ret = wl_set(wl, cmd->set, &request, sizeof(request));			
		} else if (atoi(*argv) == 2) {
			ret = wl_get(wl, cmd->get, &chosen, sizeof(chosen));
			if(ret >= 0 && chosen != 0)
				ret = wl_set(wl, WLC_SET_CHANNEL, &chosen, sizeof(chosen));
		} else {
			ret = BCME_BADARG;	
		}				
	}
	if (ret < 0) {
		cmd_usage(cmd);	
	}
	return ret;
}


cntry_name_t cntry_names[] = {

{"AFGHANISTAN",		 "AF"},
{"ALBANIA",		 "AL"},
{"ALGERIA",		 "DZ"},
{"AMERICAN SAMOA", 	 "AS"},
{"ANDORRA",		 "AD"},
{"ANGOLA",		 "AO"},
{"ANGUILLA",		 "AI"},
{"ANTARCTICA",		 "AQ"},
{"ANTIGUA AND BARBUDA",	 "AG"},
{"ARGENTINA",		 "AR"},
{"ARMENIA",		 "AM"},
{"ARUBA",		 "AW"},
{"ASCENSION ISLAND",		"AC"},
{"ASHMORE AND CARTIER ISLANDS",		"Z1"},
{"AUSTRALIA",		 "AU"},
{"AUSTRIA",		 "AT"},
{"AZERBAIJAN",		 "AZ"},
{"BAHAMAS",		 "BS"},
{"BAHRAIN",		 "BH"},
{"BAKER ISLAND",		"Z2"},
{"BANGLADESH",		 "BD"},
{"BARBADOS",		 "BB"},
{"BELARUS",		 "BY"},
{"BELGIUM",		 "BE"},
{"BELIZE",		 "BZ"},
{"BENIN",		 "BJ"},
{"BERMUDA",		 "BM"},
{"BHUTAN",		 "BT"},
{"BOLIVIA",		 "BO"},
{"BOSNIA AND HERZEGOVINA",		"BA"},
{"BOTSWANA",		 "BW"},
{"BOUVET ISLAND",	 "BV"},
{"BRAZIL",		 "BR"},
{"BRITISH INDIAN OCEAN TERRITORY", 	"IO"},
{"BRUNEI DARUSSALAM",	 "BN"},
{"BULGARIA",		 "BG"},
{"BURKINA FASO",	 "BF"},
{"BURUNDI",		 "BI"},
{"CAMBODIA",		 "KH"},
{"CAMEROON",		 "CM"},
{"CANADA",		 "CA"},
{"CAPE VERDE",		 "CV"},
{"CAYMAN ISLANDS",	 "KY"},
{"CENTRAL AFRICAN REPUBLIC",		"CF"},
{"CHAD",		 "TD"},
{"CHANNEL ISLANDS",		"Z1"},
{"CHILE",		 "CL"},
{"CHINA",		 "CN"},
{"CHRISTMAS ISLAND",	 "CX"},
{"CLIPPERTON ISLAND",		"CP"},
{"COCOS (KEELING) ISLANDS",		"CC"},
{"COLOMBIA",		 "CO"},
{"COMOROS",		 "KM"},
{"CONGO",		 "CG"},
{"CONGO, THE DEMOCRATIC REPUBLIC OF THE", "CD"},
{"COOK ISLANDS",	 "CK"},
{"CORAL SEA ISLANDS",		"Z1"},
{"COSTA RICA",		 "CR"},
{"COTE D'IVOIRE",	 "CI"},
{"CROATIA",		 "HR"},
{"CUBA",		 "CU"},
{"CYPRUS",		 "CY"},
{"CZECH REPUBLIC",	 "CZ"},
{"DENMARK",		 "DK"},
{"DJIBOUTI",		 "DJ"},
{"DOMINICA",		 "DM"},
{"DOMINICAN REPUBLIC", 	 "DO"},
{"ECUADOR",		 "EC"},
{"EGYPT",		 "EG"},
{"EL SALVADOR",		 "SV"},
{"EQUATORIAL GUINEA",	 "GQ"},
{"ERITREA",		 "ER"},
{"ESTONIA",		 "EE"},
{"ETHIOPIA",		 "ET"},
{"FALKLAND ISLANDS (MALVINAS)",	"FK"},
{"FAROE ISLANDS",	 "FO"},
{"FIJI",		 "FJ"},
{"FINLAND",		 "FI"},
{"FRANCE",		 "FR"},
{"FRENCH GUIANA",	 "GF"},
{"FRENCH POLYNESIA",	 "PF"},
{"FRENCH SOUTHERN TERRITORIES",	 "TF"},
{"GABON",		 "GA"},
{"GAMBIA",		 "GM"},
{"GEORGIA",		 "GE"},
{"GERMANY",		 "DE"},
{"GHANA",		 "GH"},
{"GIBRALTAR",		 "GI"},
{"GREECE",		 "GR"},
{"GREENLAND",		 "GL"},
{"GRENADA",		 "GD"},
{"GUADELOUPE",		 "GP"},
{"GUAM",		 "GU"},
{"GUANTANAMO BAY",		"Z1"},
{"GUATEMALA",		 "GT"},
{"GUERNSEY",		"GG"},
{"GUINEA",		 "GN"},
{"GUINEA-BISSAU",	 "GW"},
{"GUYANA",		 "GY"},
{"HAITI",		 "HT"},
{"HEARD ISLAND AND MCDONALD ISLANDS",	"HM"},
{"HOLY SEE (VATICAN CITY STATE)", 	"VA"},
{"HONDURAS",		 "HN"},
{"HONG KONG",		 "HK"},
{"HOWLAND ISLAND",		"Z2"},
{"HUNGARY",		 "HU"},
{"ICELAND",		 "IS"},
{"INDIA",		 "IN"},
{"INDONESIA",		 "ID"},
{"IRAN, ISLAMIC REPUBLIC OF",		"IR"},
{"IRAQ",		 "IQ"},
{"IRELAND",		 "IE"},
{"ISRAEL",		 "IL"},
{"ITALY",		 "IT"},
{"JAMAICA",		 "JM"},
{"JAN MAYEN",		"Z1"},
{"JAPAN",		 "JP"},
{"JAPAN_1",		"J1"},
{"JAPAN_2",		"J2"},
{"JAPAN_3",		"J3"},
{"JAPAN_4",		"J4"},
{"JAPAN_5",		"J5"},
{"JAPAN_6",		"J6"},
{"JAPAN_7",		"J7"},
{"JAPAN_8",		"J8"},
{"JARVIS ISLAND",		"Z2"},
{"JERSEY",		"JE"},
{"JOHNSTON ATOLL",		"Z2"},
{"JORDAN",		 "JO"},
{"KAZAKHSTAN",		 "KZ"},
{"KENYA",		 "KE"},
{"KINGMAN REEF",		"Z2"},
{"KIRIBATI",		 "KI"},
{"KOREA, DEMOCRATIC PEOPLE'S REPUBLIC OF", "KP"},
{"KOREA, REPUBLIC OF",	 "KR"},
{"KUWAIT",		 "KW"},
{"KYRGYZSTAN",		 "KG"},
{"LAO PEOPLE'S DEMOCRATIC REPUBLIC", 	"LA"},
{"LATVIA",		 "LV"},
{"LEBANON",		 "LB"},
{"LESOTHO",		 "LS"},
{"LIBERIA",		 "LR"},
{"LIBYAN ARAB JAMAHIRIYA",		"LY"},
{"LIECHTENSTEIN",	 "LI"},
{"LITHUANIA",		 "LT"},
{"LUXEMBOURG",		 "LU"},
{"MACAO",		 "MO"},
{"MACEDONIA, THE FORMER YUGOSLAV REPUBLIC OF",	 "MK"},
{"MADAGASCAR",		 "MG"},
{"MALAWI",		 "MW"},
{"MALAYSIA",		 "MY"},
{"MALDIVES",		 "MV"},
{"MALI",		 "ML"},
{"MALTA",		 "MT"},
{"MAN, ISLE OF",		"IM"},
{"MARSHALL ISLANDS",	 "MH"},
{"MARTINIQUE",		 "MQ"},
{"MAURITANIA",		 "MR"},
{"MAURITIUS",		 "MU"},
{"MAYOTTE",		 "YT"},
{"MEXICO",		 "MX"},
{"MICRONESIA, FEDERATED STATES OF", 	"FM"},
{"MIDWAY ISLANDS",		"Z2"},
{"MOLDOVA, REPUBLIC OF", "MD"},
{"MONACO",		 "MC"},
{"MONGOLIA",		 "MN"},
{"MONTSERRAT",		 "MS"},
{"MOROCCO",		 "MA"},
{"MOZAMBIQUE",		 "MZ"},
{"MYANMAR",		 "MM"},
{"NAMIBIA",		 "NA"},
{"NAURU",		 "NR"},
{"NEPAL",		 "NP"},
{"NETHERLANDS",		 "NL"},
{"NETHERLANDS ANTILLES", "AN"},
{"NEW CALEDONIA",	 "NC"},
{"NEW ZEALAND",		 "NZ"},
{"NICARAGUA",		 "NI"},
{"NIGER",		 "NE"},
{"NIGERIA",		 "NG"},
{"NIUE",		 "NU"},
{"NORFOLK ISLAND",	 "NF"},
{"NORTHERN MARIANA ISLANDS",		"MP"},
{"NORWAY",		 "NO"},
{"OMAN",		 "OM"},
{"PAKISTAN",		 "PK"},
{"PALAU",		 "PW"},
{"PALESTINIAN TERRITORY, OCCUPIED", 	"PS"},
{"PALMYRA ATOLL",		"Z2"},
{"PANAMA",		 "PA"},
{"PAPUA NEW GUINEA",	 "PG"},
{"PARAGUAY",		 "PY"},
{"PERU",		 "PE"},
{"PHILIPPINES",		 "PH"},
{"PITCAIRN",		 "PN"},
{"POLAND",		 "PL"},
{"PORTUGAL",		 "PT"},
{"PUERTO RICO",		 "PR"},
{"QATAR",		 "QA"},
{"REUNION",		 "RE"},
{"ROMANIA",		 "RO"},
{"ROTA ISLAND",		"Z1"},
{"RUSSIAN FEDERATION",	 "RU"},
{"RWANDA",		 "RW"},
{"SAIPAN",		"Z1"},
{"SAINT HELENA",	 "SH"},
{"SAINT KITTS AND NEVIS",		"KN"},
{"SAINT LUCIA",		 "LC"},
{"SAINT PIERRE AND MIQUELON",	 	"PM"},
{"SAINT VINCENT AND THE GRENADINES", 	"VC"},
{"SAMOA",		 "WS"},
{"SAN MARINO",		 "SM"},
{"SAO TOME AND PRINCIPE",		"ST"},
{"SAUDI ARABIA",	 "SA"},
{"SENEGAL",		 "SN"},
{"SEYCHELLES",		 "SC"},
{"SIERRA LEONE",	 "SL"},
{"SINGAPORE",		 "SG"},
{"SLOVAKIA",		 "SK"},
{"SLOVENIA",		 "SI"},
{"SOLOMON ISLANDS",	 "SB"},
{"SOMALIA",		 "SO"},
{"SOUTH AFRICA",	 "ZA"},
{"SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS", "GS"},
{"SPAIN",		 "ES"},
{"SRI LANKA",		 "LK"},
{"SUDAN",		 "SD"},
{"SURINAME",		 "SR"},
{"SVALBARD AND JAN MAYEN",		"SJ"},
{"SWAZILAND",		 "SZ"},
{"SWEDEN",		 "SE"},
{"SWITZERLAND",		 "CH"},
{"SYRIAN ARAB REPUBLIC", "SY"},
{"TAIWAN, PROVINCE OF CHINA", 		"TW"},
{"TAJIKISTAN",		 "TJ"},
{"TANZANIA, UNITED REPUBLIC OF",	"TZ"},
{"THAILAND",		 "TH"},
{"TIMOR-LESTE (EAST TIMOR)",		"TL"},
{"TOGO",		 "TG"},
{"TOKELAU",		 "TK"},
{"TONGA",		 "TO"},
{"TRINIDAD AND TOBAGO",	 "TT"},
{"TRISTAN DA CUNHA",		"TA"},
{"TUNISIA",		 "TN"},
{"TURKEY",		 "TR"},
{"TURKMENISTAN",	 "TM"},
{"TURKS AND CAICOS ISLANDS",		"TC"},
{"TUVALU",		 "TV"},
{"UGANDA",		 "UG"},
{"UKRAINE",		 "UA"},
{"UNITED ARAB EMIRATES", "AE"},
{"UNITED KINGDOM",	 "GB"},
{"UNITED STATES",	 "US"},
{"UNITED STATES MINOR OUTLYING ISLANDS",		"UM"},
{"URUGUAY",		 "UY"},
{"UZBEKISTAN",		 "UZ"},
{"VANUATU",		 "VU"},
{"VENEZUELA",		 "VE"},
{"VIET NAM",		 "VN"},
{"VIRGIN ISLANDS, BRITISH", "VG"},
{"VIRGIN ISLANDS, U.S.", "VI"},
{"WALLIS AND FUTUNA",	 "WF"},
{"WESTERN SAHARA", 	 "EH"},
{"YEMEN",		 "YE"},
{"YUGOSLAVIA",		 "YU"},
{"ZAMBIA",		 "ZM"},
{"ZIMBABWE",		 "ZW"},
{"RADAR CHANNELS",	"RDR"},
{"ALL CHANNELS",	"ALL"},
{NULL, 			 NULL}
};
